/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

Copyright (c) 1995 International Computer Science Institute
All Rights Reserved.

Created by Wieland Holfelder (whd@icsi.berkeley.edu)
Tenet Group
International Computer Science Institute

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research and non-profit purposes,
without fee, and without a written agreement is hereby granted,
provided that the above copyright notice, this paragraph and the
following five paragraphs appear in all copies.

Permission to incorporate this software into commercial products may
be obtained from the International Computer Science Institute,
1947 Center Street, Suite 600, Berkeley, CA 94704.

IN NO EVENT SHALL THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE BE
LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE INTERNATIONAL
COMPUTER SCIENCE INSTITUTE  HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE SPECIFICALLY DISCLAIMS
ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE INTERNATIONAL
COMPUTER SCIENCE INSTITUTE  HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE,
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

The International Computer Science Institute will appreciate
receiving any enhancements and modifications of this software
made by others, and reserves the right to include them in future
releases. The Institute also recommends that this software not
be redistributed by third parties, but be obtained by all directly
from the ICSI server, so that all users of the software may
receive bug notifications and other relevant announcements.

The German Academic Exchange Service has supported the development
of this software. Their support is gratefully acknowledged.

* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static const char rcsid[]="$Id: MBoneMedia.cc,v 1.3 1996/08/12 07:57:58 whd Exp $";

#include "MBoneMedia.h"

extern const char version[];

MBoneMedia::MBoneMedia(String		_filename,
		       int		_read_only,
		       String		_ipaddr,
		       int		_port,
		       u_int16_t		_confid,
		       u_int8_t		_ttl,
		       int		_ctrl_offset,
		       int		_media_buffer_size,
		       int		_session_buffer_size,
		       int		_id_buffer_size,
		       int		_number):

  media_timer(this, &media_timeout),
  ctrl_timer(this, &ctrl_timeout),
  no_data_timer(this, &no_data_timeout),
  media_data_socket(this, &data_handle),
  media_session_socket(this, &session_handle),
  media_file()

{
  vcr_file_rc frc;
  int rc;
  #ifdef DEBUG1
  cerr << "MBoneMedia()\n";
  #endif

  Media_buffer_size=0;
  Media_buffer=NULL;
  Id_buffer_size=0;
  Id_buffer=NULL;
  Session_buffer_size=0;
  Session_buffer=NULL;

  media_filename=_filename;
  media_ipaddr=_ipaddr;
  media_port=_port;
  media_confid=_confid;
  media_ttl=_ttl;
  nb_of_sockets=1;
  if (_ctrl_offset) nb_of_sockets=2;
  media_number=_number;
  recording_enabled=0;

  media_state=INVALID;
  media_fmt=RTP;
  mute=0;

  data_queue.clear();
  session_queue.clear();
  src_set.clear();
  sync_src=next_sync_src=0;

  /* initialize buffer for media packets: */
  init_media_buffer(&media_buffer, _media_buffer_size);

  /* initialize buffer for session data: */
  init_session_buffer(&session_buffer, _session_buffer_size);

  /* initialize buffer for id data: */
  init_id_buffer(&id_buffer, _id_buffer_size);

  media_packet_pending=0;
  event_hdlr=(MediaEvtHdlr) NULL;
  refresh_media_msg=10000;
  send_ctrl_period=0;
  ctrl_packet_size=0;
  avg_ctrl_packet_size=0;

  /* send MEDIA_MAGIC with every session msg. */
  MEDIA_MAGIC="MBone VCR";
  char x[20],y[20],z[20];
  sscanf((const char *) version, "%s %s %s", x, y, z);
  media_msg=(const char *) z;

  if ((frc=media_file.open(media_filename, _read_only))<=VCR_FILE_OK) {
    if ((rc=media_data_socket.open(media_ipaddr,media_port,media_ttl,1,1))>0) {
      if ((nb_of_sockets==1) ||
	  ((nb_of_sockets==2) &&
	   ((rc=media_session_socket.open(media_ipaddr,media_port+_ctrl_offset,media_ttl,1,1))>0))) {
	media_file.goto_start();
	media_state=IDLE;
	media_data_socket.deactivate_read_hdlr();
	if (nb_of_sockets==2) media_session_socket.deactivate_read_hdlr();
	media_timer.cancel();
	media_timer.msched(500);
      } else {
        #ifdef DEBUG1
	cerr << "MBoneMedia::MBoneMedia() media_session_socket.open() failed!\n";
        #endif
	error_msg="Cannot open session socket.";
	media_data_socket.close();
	media_file.close();
	media_state=INVALID;
      }
    } else {
      #ifdef DEBUG1
      cerr << "MBoneMedia::MBoneMedia() media_data_socket.open() failed!\n";
      #endif
      error_msg="Cannot open data socket.";
      media_file.close();
      media_state=INVALID;
    }
  } else {
    #ifdef DEBUG1
    cerr << "MBoneMedia::MBoneMedia() Cannot open file \"" << media_filename << "\".\n";
    #endif
    char num[10];
    sprintf(num,"%d",frc);
    error_msg="Cannot open media_files ("+media_filename+".idx/.rec) Error: "+num;
    media_state=INVALID;
  };
};

MBoneMedia::~MBoneMedia()
{
  VCRPacket p;
  #ifdef DEBUG1
  cerr << "~MBoneMedia()\n";
  #endif

  while (data_queue.length()>0) {
    p=data_queue.deq();
    if (p.data) free(p.data);
  };
  src_set.clear();

  free_media_buffer(&media_buffer);
  free_session_buffer(&session_buffer);
  free_id_buffer(&id_buffer);

  call_event_hdlr(MEDIA_GONE);
};

media_rcs MBoneMedia::media_change_state(media_states new_media_state,
					 u_int32_t time,
					 int start_rec) {
  VCR_Hdr vcr_hdr;
  vcr_file_rc frc;
  u_int32_t old_length;
  VCRDisplay& display=VCRDisplay::instance();
  VCRTcl& tcl=VCRTcl::instance();

  switch (new_media_state) {
  case PLAYING:
    while (data_queue.length()>0) {
      if (data_queue.front().data) free(data_queue.front().data);
      data_queue.del_front();
    };
    while (session_queue.length()>0) {
      if (session_queue.front().data) free(session_queue.front().data);
      session_queue.del_front();
    };
    start_playing();
    media_msg="playing";
    refresh_media_msg=1000;
    media_state=new_media_state;
    break;
  case RECORDING:
    if (!is_writable()) return MEDIA_IS_READ_ONLY;
    start_recording();
    recording_started=start_rec;
    timed_out=0;
    media_msg="recording";
    refresh_media_msg=1000;
    last_ts=0;
    rec_timeout=tcl.int_var("session(rec_timeout)");
    if (!get_mute()) {
      if (nb_of_sockets==2) media_session_socket.activate_read_hdlr();
      media_data_socket.activate_read_hdlr();
    };
    break;
  case FF:
    media_msg="forwarding";
    refresh_media_msg=250;
    media_state=new_media_state;
    break;
  case REW:
    media_msg="rewinding";
    refresh_media_msg=250;
    media_state=new_media_state;
    break;
  case IDLE:
    media_msg="stopped";
    refresh_media_msg=5000;
    tcl.clear_led(media_number);

    switch (media_state) {
    case PLAYING:
      media_timer.cancel();
      stop_playing();
      while (data_queue.length()>0) {
	if (data_queue.front().data) free(data_queue.front().data);
	data_queue.del_front();
      };
      while (session_queue.length()>0) {
	if (session_queue.front().data) free(session_queue.front().data);
	session_queue.del_front();
      };
      media_timer.cancel();
      media_timer.msched(refresh_media_msg);
      break;
    case RECORDING:
      stop_recording();
      no_data_timer.cancel();
      media_data_socket.deactivate_read_hdlr();
      if (nb_of_sockets==2) media_session_socket.deactivate_read_hdlr();
      if (is_recording()) {
	media_file.flush();
	old_length=media_file.length();
	media_file.goto_end();
	u_int32_t time=display.get_time();
	if ((media_file.timestamp()<time) ||
	    ((media_file.timestamp()<old_length) && old_length<time)) {
#ifdef DEBUG1
	  cerr << "MBoneMedia::media_change_state(): write TIME_CHANNEL info" << endl;
#endif
	  vcr_hdr.ch=TIME_CHANNEL;
	  vcr_hdr.idx=0;
	  vcr_hdr.l=0;
	  vcr_hdr.ts=time;
	  if ((frc=media_file.write(&vcr_hdr,NULL))!=VCR_FILE_OK) {
	    cerr << "MBoneMedia::media_change_state() media_file.write() failed ("<<frc<<").\n";
	    call_event_hdlr(MEDIA_ERROR);
	    recording_started=0;
	    return MEDIA_ERROR;
	  };
	  media_file.flush();
	};
      };
      recording_started=0;
      break;
    default:
      break;
    };
    media_state=new_media_state;
    tcl.clear_led(media_number);
    break;
  default:
    return MEDIA_INVALID_STATE;
  };
  /* send msg and set new timer */
  media_timeout();
  return MEDIA_OK;

};

media_rcs MBoneMedia::start_recording() {
  return MEDIA_OK;
};

media_rcs MBoneMedia::stop_recording() {
  return MEDIA_OK;
};

media_rcs MBoneMedia::start_playing() {
  return MEDIA_OK;
};

media_rcs MBoneMedia::stop_playing() {
  return MEDIA_OK;
};

media_rcs MBoneMedia::set_index(u_int8_t idx, VCR_Index* vcr_idx_p) {
  VCRDisplay& display=VCRDisplay::instance();

  if (media_state == RECORDING) {
    return media_return(media_file.set_index(idx, vcr_idx_p));
  } else {
    if (get_length()<display.get_time()) {
      return MEDIA_IDX_TOO_LATE;
    } else {
      return media_return(media_file.write_index(idx, vcr_idx_p));
    };
  }
};


void MBoneMedia::data_handle(eventtype event) {
  int length;
  struct sockaddr_in sin;
  if (event!=MCS_READ) return;
  #ifdef DEBUG2
  cerr << "MBoneMedia::data_handle()\n";
  #endif
  length=media_data_socket.mcs_recv(&media_buffer,media_buffer_size(),&sin);
  if (media_state!=RECORDING) return;
  /* if this is the very first data packet, inform the other media */
  /* to start recording as well */
  if (!is_recording()) call_event_hdlr(MEDIA_RECORDING_STARTED);

  timed_out=0;

  if (rec_timeout) {
    no_data_timer.cancel();
    no_data_timer.ssched(rec_timeout);
  };

  media_data_handle(&sin, length);
};

void MBoneMedia::media_data_handle(struct sockaddr_in *sinp, int length) {
  cerr << "MBoneMedia::media_data_handle() should be overloaded...\n";
};

void MBoneMedia::session_handle(eventtype event) {
  int length;
  struct sockaddr_in sin;
  if (event!=MCS_READ) return;
  #ifdef DEBUG2
  cerr << "MBoneMedia::session_handle()\n";
  #endif
  length=media_session_socket.mcs_recv(&media_buffer,media_buffer_size(),&sin);
  if (media_state!=RECORDING) return;

  media_session_handle(&sin, length);
};

void MBoneMedia::media_session_handle(struct sockaddr_in *sinp, int length) {
  cerr << "MBoneMedia::media_session_handle() should be overloaded...\n";
};

void MBoneMedia::media_timeout() {
  char msg[200];
  u_int32_t time;
  VCRDisplay& display=VCRDisplay::instance();

  #ifdef DEBUG2
  cerr << "MBoneMedia::media_timeout()\n";
  #endif
  time=display.get_time();
  sprintf(msg,"%s%s %s (%.2d:%.2d:%.2d)%s", (get_mute()?"<":""),
	  (const char *) MEDIA_MAGIC, (const char *) media_msg,
	  ms2hour(time), ms2mod_min(time), ms2mod_sec(time),(get_mute()?">":""));
  send_msg(SEND_ID, msg);
  media_timer.cancel();
  media_timer.msched(refresh_media_msg);
};

void MBoneMedia::ctrl_timeout() {

  #ifdef DEBUG2
  cerr << "MBoneMedia::ctrl_timeout()\n";
  #endif
  ctrl_timer.cancel();
  ctrl_timer.ssched(send_ctrl_period);
};

void MBoneMedia::no_data_timeout() {

  #ifdef DEBUG2
  cerr << "MBoneMedia::no_data_timeout()\n";
  #endif
  timed_out=1;
  no_data_timer.cancel();
  call_event_hdlr(MEDIA_NO_DATA_TIMEOUT);
};

media_rcs MBoneMedia::send_msg(msg_codes code, String msg) {
  cerr << "MBoneMedia::send_msg(" << int(code) << ", " << msg << ")\n";
  return MEDIA_OK;
};

media_rcs MBoneMedia::send_ctrl() {
  cerr << "MBoneMedia::send_ctrl()\n";
  return MEDIA_OK;
};
