/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

Copyright (c) 1995 International Computer Science Institute
All Rights Reserved.

Created by Wieland Holfelder (whd@icsi.berkeley.edu)
Tenet Group
International Computer Science Institute

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research and non-profit purposes,
without fee, and without a written agreement is hereby granted,
provided that the above copyright notice, this paragraph and the
following five paragraphs appear in all copies.

Permission to incorporate this software into commercial products may
be obtained from the International Computer Science Institute,
1947 Center Street, Suite 600, Berkeley, CA 94704.

IN NO EVENT SHALL THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE BE
LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE INTERNATIONAL
COMPUTER SCIENCE INSTITUTE  HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE SPECIFICALLY DISCLAIMS
ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE INTERNATIONAL
COMPUTER SCIENCE INSTITUTE  HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE,
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

The International Computer Science Institute will appreciate
receiving any enhancements and modifications of this software
made by others, and reserves the right to include them in future
releases. The Institute also recommends that this software not
be redistributed by third parties, but be obtained by all directly
from the ICSI server, so that all users of the software may
receive bug notifications and other relevant announcements.

The German Academic Exchange Service has supported the development
of this software. Their support is gratefully acknowledged.

* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static const char rcsid[]="$Id: MBoneSession.cc,v 1.4 1996/08/12 11:36:22 whd Exp $";

#include "MBoneSession.h"
#include "RTPMedia.h"
#include "VCRTcl.h"
#include "VCRDisplay.h"
#include "VCRFile.h"

MBoneSession::MBoneSession() :
session_timer(this, &ttimeout)
{
  VCRTcl& tcl=VCRTcl::instance();
  VCRDisplay& display=VCRDisplay::instance();

  #ifdef DEBUG1
  cerr << "MBoneSession::MBoneSession()\n";
  #endif

  display.set(0,1);
  session_time=0;
  session_speed=1;
  session_length=0;
  session_state=IDLE;
  blank_skip_time=0;
  blank_skip_media=0;
  old_state=IDLE;
  recording_started=0;

  recording_enabled=tcl.int_var((const char*) "session(rec_enabled)");
  if (tcl.int_var("session(loaded)")==1) {
    if (read_session_vars()!=SESSION_OK) session_state=INVALID;
  };
  if (session_state==IDLE) tcl.CreateCommand("vcr_ctrl", vcr_ctrl, (ClientData)this);
};

MBoneSession::~MBoneSession()
{
  Pix pix;

  #ifdef DEBUG1
  cerr << "MBoneSession::~MBoneSession()\n";
  #endif

  media_sched.clear();
  session_length=0;
  session_time=0;

  /* delete media objects */
  for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix)) delete p2mbmList(pix);
  p2mbmList.clear();

};

session_rcs MBoneSession::add_media(P2MBoneMedia p2mbm)
{
  VCRDisplay& display=VCRDisplay::instance();
  #ifdef DEBUG1
  cerr << "MBoneSession::add_media()\n";
  #endif
  p2mbmList.append(p2mbm);
  p2mbm->install_event_hdlr(this, &event_hdlr, p2mbmList.length());
  session_length=get_session_length();
  display.update_total_length(session_length);
  return SESSION_OK;
};

session_rcs MBoneSession::del_media(P2MBoneMedia p2mbm)
{
  #ifdef DEBUG1
  cerr << "MBoneSession::del_media()\n";
  #endif
  p2mbmList.del((void *&) p2mbm);
  session_length=get_session_length();
  return SESSION_OK;
};

session_rcs MBoneSession::set_mute_media(int index, int mute)
{
  Pix pix;
  int i;
  media_rcs rc;
  #ifdef DEBUG1
  cerr << "MBoneSession::set_mute_media(" << index << "," << mute << ")\n";
  #endif
  pix = p2mbmList.first();
  for (i=1; i<index; i++) p2mbmList.next(pix);
  rc=p2mbmList(pix)->set_mute(mute);
  switch (rc) {
  case MEDIA_OK:
    return SESSION_OK;
  default:
    return SESSION_ERROR;
  };
};



session_rcs MBoneSession::del_session()
{
  Pix pix;
  VCRDisplay& display=VCRDisplay::instance();
  VCRTcl& tcl=VCRTcl::instance();

  #ifdef DEBUG1
  cerr << "MBoneSession::del_session()\n";
  #endif
  if (session_state!=IDLE) this->stop();
  tcl.release_button();
  media_sched.clear();
  session_length=0;
  session_time=0;

  display.set(0,1);
  display.update_total_length(0);

  /* delete media objects */
  for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix)) delete p2mbmList(pix);
  p2mbmList.clear();

  return SESSION_OK;
};


session_rcs MBoneSession::play(float set_speed)
{
  media_rcs rc;
  int32_t signed_ts;
  u_int32_t ts=0;
  MediaSchedule ms;
  Pix pix;

  VCRTcl& tcl=VCRTcl::instance();
  VCRDisplay& display=VCRDisplay::instance();

  #ifdef DEBUG1
  cerr << "MBoneSession::play()\n";
  #endif

  if (nb_of_media()==0) return SESSION_EMPTY;

  if (set_speed!=0) session_speed=set_speed;

  switch (session_state) {
  case PLAYING:
    return SESSION_OK;
  case IDLE:

    if ((session_time==session_length) && (!tcl.int_var("session(loop_mode)"))) {
      #ifdef DEBUG1
      cerr << "MBoneSession::play() SESSION_EOF.\n";
      #endif
      return SESSION_EOF;
    };
    if (ttl_warning_sent==0) {
      if (nb_of_ttl_higher_16) {
	char x[1000];
	sprintf(x," Your session has %d media with a ttl higher than 16. \nAre you sure you want to play your session with this scope? \n\nIf not press \\\"Cancel\\\" and change the ttl in \\\"Edit Session\\\" in the \\\"Session\\\" menu. \n\nTo play your session with this ttl press \\\"Ok\\\".",nb_of_ttl_higher_16);
	int rc=1;
	/* for some reason the gcc compiler on an DEC Alpha with OSF
	   crashes with an internal error at this point !?
	   rc=tcl.WarningMsg(x);
	*/
	tcl.WarningMsg(x);
	if (rc) {
	  ttl_warning_sent=1;
	} else {
	  return SESSION_ERROR;
	}
      }
    }
    media_sched.clear();
    session_timer.cancel();
    old_state=session_state;
    session_state=PLAYING;
    nb_of_active_media=0;
    nb_of_inactive_media=0;
    last_ts=0;

    for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
    {
      rc=p2mbmList(pix)->media_change_state(PLAYING);

      #ifdef DEBUG2
      cerr << "MBoneSession::play("<<p2mbmList(pix)->get_media_name()<<
              ") media_change_state: " << int(rc) << "\n";
      #endif

      switch (rc) {
      case MEDIA_OK:
        rc=p2mbmList(pix)->read_next_sync_unit(&ts, session_speed, 1);

        #ifdef DEBUG2
	cerr << "MBoneSession::play("<<p2mbmList(pix)->get_media_name()<<
	        ")read_next_sync_unit rc: " << int(rc) << " ts:" << ts << "\n";
        #endif

	switch (rc) {
	case MEDIA_OK:
	  nb_of_active_media++;
	  last_ts=ts;
	  ms.ts=ts;
	  ms.pix=pix;
	  media_sched.enq(ms);

	  if (blank_skip_time) {
	    if ((blank_skip_media==pix) && (ms.ts>session_time) &&
		((ms.ts-session_time) >= blank_skip_time)) {
            #ifdef DEBUG1
	      cerr << "MBoneSession::play("<< p2mbmList(pix)->get_media_name() <<
	            ") BLANK_SKIP " << ms.ts-session_time << " ms.\n";
            #endif
	      char msg[100];
	      sprintf(msg,"Blank skiped %d secs.",int(int(ms.ts-session_time)/1000));
	      tcl.set_msg(msg, 3000);
	      goto_ts(p2mbmList(pix)->get_ts());
	      signed_ts=0;
	      break;
	    }
	  };

	  signed_ts=int32_t(media_sched.front().ts)-int32_t(session_time);

          #ifdef DEBUG2
	  cerr << "MBoneSession::play("<<p2mbmList(ms.pix)->get_media_name()<<
	  ") next packet scheduled in "<<int32_t(signed_ts/session_speed)
	  <<" ms st: "<<session_time<<".\n";
          #endif
	  break;
	case MEDIA_EMPTY:
	case MEDIA_EOF:
	  nb_of_inactive_media++;
	  break;
	default:
	  cancel_operation(pix);
	  media_msg(rc,"MBoneSession::play()");
	  VCRTcl& tcl=VCRTcl::instance();
	  tcl.ErrorMsg(p2mbmList(pix)->get_error_msg());
	  return SESSION_ERROR;
	};
      default:
	break;
      };
    };

    #ifdef DEBUG2
    cerr << "MBoneSession::play(): nb_of_active_media: " << nb_of_active_media << "\n";
    cerr << "MBoneSession::play(): nb_of_inactive_media: " << nb_of_inactive_media << "\n";
    #endif

    if (nb_of_active_media>0) {
      display.start(session_time, session_speed);
    } else {
      if (nb_of_inactive_media>0) {
	stop();
	goto_end();
	if (session_length && tcl.int_var("session(loop_mode)")) {
	  goto_start();
	  play();
	  return SESSION_OK;
	} else {
	  tcl.release_button();
	  tcl.set_msg("Nothing to play.", 3000);
	  return SESSION_EOF;
	};
      };
      return SESSION_EMPTY;
    };

    if (signed_ts<=0) ttimeout(); else session_timer.msched(int(signed_ts/session_speed));

    break;
  case FF:
  case REW:
    stop();
    return play(session_speed);
    break;
  default:
    return SESSION_INVALID_STATE;
  };
  return SESSION_OK;
}

session_rcs MBoneSession::rew(int factor)
{
  media_rcs rc;
  Pix pix;
  VCRDisplay& display=VCRDisplay::instance();

  if (nb_of_media()==0) return SESSION_EMPTY;

  if (session_state==IDLE)
  {
    #ifdef DEBUG1
    cerr << "MBoneSession::rew("<<factor<<")\n";
    #endif
    ff_rew_factor=factor;
    old_state=session_state;
    session_state=REW;
    for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
    {
      rc=p2mbmList(pix)->media_change_state(REW);
      if (rc>MEDIA_OK) media_msg(rc,"MBoneSession::rew()");
    };
    session_timer.msched(25);
    return SESSION_OK;
  } else if (session_state==REW) {
    if ((25*ff_rew_factor)<=session_time) {
      session_time-=25*ff_rew_factor;
      display.set(session_time,session_speed);
      session_timer.msched(25);
      return SESSION_OK;
    } else {
      session_time=0;
      VCRTcl& tcl=VCRTcl::instance();
      tcl.release_button();
      tcl.set_msg("Begin of session reached.",3000);
      return stop();
    };
  } else if (session_state == RECORDING) {
    return SESSION_ERROR;
  } else {
    if (stop()==SESSION_OK) return rew(ff_rew_factor);
  }
  return SESSION_OK;
};

session_rcs MBoneSession::ff(int factor)
{
  media_rcs rc;
  Pix pix;
  VCRDisplay& display=VCRDisplay::instance();

  if (nb_of_media()==0) return SESSION_EMPTY;

  if (session_state==IDLE)
  {
    #ifdef DEBUG1
    cerr << "MBoneSession::ff("<<factor<<")\n";
    #endif
    ff_rew_factor=factor;
    old_state=session_state;
    session_state=FF;

    for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
    {
      rc=p2mbmList(pix)->media_change_state(FF);
      if (rc>MEDIA_OK) media_msg(rc,"MBoneSession::ff()");
    }
    session_timer.msched(25);
    return SESSION_OK;
  } else if (session_state==FF) {
    if ((session_time+25*ff_rew_factor)<=session_length) {
      session_time+=25*ff_rew_factor;
      display.set(session_time,session_speed);
      session_timer.msched(25);
      return SESSION_OK;
    } else {
      session_time=session_length;
      VCRTcl& tcl=VCRTcl::instance();
      tcl.set_msg("End of session reached.",3000);
      tcl.release_button();
      return stop();
    };
  } else if (session_state == RECORDING) {
    return SESSION_ERROR;
  } else {
    if (stop()==SESSION_OK) return ff(ff_rew_factor);
  }
  return SESSION_OK;
};


session_rcs MBoneSession::rec(int start_rec)
{
  media_rcs rc;
  Pix pix;
  VCRDisplay& display=VCRDisplay::instance();
  VCRTcl& tcl=VCRTcl::instance();

  #ifdef DEBUG1
  cerr << "MBoneSession::rec()\n";
  #endif

  nb_of_active_media=0;

  if (nb_of_media()==0) return SESSION_EMPTY;
  if (!is_writable()) return SESSION_IS_READ_ONLY;

  if (session_state==IDLE)
  {
    if (session_time<session_length) {
      tcl.ErrorMsg("Recording can only be started\nat the end of the session.");
      return SESSION_ERROR;
    };
    goto_end();
    old_state=session_state;
    for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
    {
      rc=p2mbmList(pix)->media_change_state(RECORDING, session_time, start_rec);
      switch (rc) {
      case MEDIA_OK:
	nb_of_active_media++;
	break;
      default:
	tcl.ErrorMsg(p2mbmList(pix)->get_error_msg());
	media_msg(rc,"MBoneSession::rec()");
	cancel_operation(pix);
	return SESSION_ERROR;
      };
    };

    #ifdef DEBUG2
    cerr << "MBoneSession::rec(): nb_of_active_media: " << nb_of_active_media << "\n";
    #endif
    /* yes, it is only one "=" */
    if ((recording_started=start_rec)) {
      display.start(session_time, 1, 1);
    } else {
      tcl.set_msg("Waiting for the first data to arrive...",0);
    };
    session_state=RECORDING;
    for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix)) p2mbmList(pix)->record_now();
    return SESSION_OK;
  } else if (session_state==RECORDING) {
    return SESSION_OK;
  } else {
    return SESSION_INVALID_STATE;
  }
};

session_rcs MBoneSession::stop()
{
  media_rcs rc;
  Pix pix;
  MediaSchedule ms;
  VCRDisplay& display=VCRDisplay::instance();
  VCRTcl& tcl=VCRTcl::instance();

  if (nb_of_media()==0) return SESSION_EMPTY;

  #ifdef DEBUG1
  cerr << "MBoneSession::stop()\n";
  #endif

  tcl.set_msg("", 0);
  nb_of_active_media=0;
  switch (session_state) {
  case IDLE:
    old_state=IDLE;
    return SESSION_OK;
    break;
  case PLAYING:
    session_timer.cancel();
    old_state=session_state;
    session_state=IDLE;
    display.stop();

    /* go back to packets that were not played yet */
    while (media_sched.length()) {
      ms=media_sched.deq();
      #ifdef DEBUG1
      cout << "media         : " << p2mbmList(ms.pix)->get_media_name() << endl; 
      cout << "media ts      : " << p2mbmList(ms.pix)->get_ts() << endl;
      cout << "session ts    : " << get_session_ts() << endl;
      cout << "media length  : " << p2mbmList(ms.pix)->get_length() << endl;
      cout << "session length: " << get_session_length() << endl << endl;
      #endif
      if (p2mbmList(ms.pix)->get_ts()<get_session_ts()) {
	#ifdef DEBUG1
	cout << "MBoneSession::stop() media time < session time!\n";
	#endif
	if (nb_of_media()==1) p2mbmList(ms.pix)->goto_ts(ms.ts);
      } else if (p2mbmList(ms.pix)->get_length()<get_session_length()) {
	#ifdef DEBUG1
	cout << "MBoneSession::stop() media length < session length!\n";
	#endif
	if (nb_of_media()==1) p2mbmList(ms.pix)->goto_ts(ms.ts);
      } else {
	#ifdef DEBUG1
	cout << "MBoneSession::stop() goto_ts("<< ms.ts << ")\n";
	#endif
	p2mbmList(ms.pix)->goto_ts(ms.ts);
      }
    };
    media_sched.clear();

    for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
    {
      #ifdef DEBUG1
      cout << "media         : " << p2mbmList(pix)->get_media_name() << endl; 
      cout << "media ts      : " << p2mbmList(pix)->get_ts() << endl;
      cout << "session ts    : " << get_session_ts() << endl;
      cout << "media length  : " << p2mbmList(pix)->get_length() << endl;
      cout << "session length: " << get_session_length() << endl << endl;
      #endif
      rc=p2mbmList(pix)->media_change_state(IDLE);
      if (rc>MEDIA_OK) media_msg(rc,"MBoneSession::stop()");
    };
    session_time=get_session_ts();
    display.set(session_time, session_speed);
    break;
  case RECORDING:
    old_state=session_state;
    session_state=IDLE;
    if (recording_started) {
      for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
      {
	rc=p2mbmList(pix)->media_change_state(IDLE);
	if (rc>MEDIA_OK) media_msg(rc,"MBoneSession::stop()");
      };
      display.stop();
      session_time=session_length=get_session_length();
      display.update_total_length(session_time);
      recording_started=0;
    };
    break;
  case FF:
  case REW:
    session_timer.cancel();
    old_state=session_state;
    session_state=IDLE;
    for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
    {
      rc=p2mbmList(pix)->media_change_state(IDLE);
      if (rc>MEDIA_OK) media_msg(rc,"MBoneSession::stop()");
    };
    if (((old_state==FF) && (session_time<session_length)) ||
	((old_state==REW) && (session_time>0))) {
      goto_ts(session_time);
    } else if (old_state==FF) {
      goto_end();
    } else if (old_state==REW) {
      goto_start();
    }
    break;
  default:
    return SESSION_INVALID_STATE;
  };
  return SESSION_OK;
};

session_rcs MBoneSession::goto_end()
{
  media_rcs rc;
  Pix pix;
  VCRDisplay& display=VCRDisplay::instance();

  #ifdef DEBUG1
  cerr << "MBoneSession::goto_end()\n";
  #endif

  if (nb_of_media()==0) return SESSION_EMPTY;

  if (session_state == RECORDING) {
    return SESSION_ERROR;
  };

  for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
  {
    rc=p2mbmList(pix)->goto_end();
    if (rc>MEDIA_OK) media_msg(rc,"MBoneSession::goto_end()");
    session_time=session_length;
  }
  display.set(session_time,session_speed);
  return SESSION_OK;
};

session_rcs MBoneSession::goto_ts(u_int32_t ts)
{
  media_rcs rc;
  Pix pix;
  VCRDisplay& display=VCRDisplay::instance();

  #ifdef DEBUG1
  cerr << "MBoneSession::goto_ts(" << ts << ")\n";
  #endif

  if (nb_of_media()==0) return SESSION_EMPTY;
  if (session_state == RECORDING) {
    return SESSION_ERROR;
  };

  if (ts>session_length) {
    return goto_end();
  };

  session_time=ts;
  display.set(session_time, session_speed);

  for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
  {
    rc=p2mbmList(pix)->goto_ts(ts);
    if (rc>MEDIA_OK) media_msg(rc,"MBoneSession::goto_ts()");
  };
  return SESSION_OK;
};



String MBoneSession::set_index(int media, u_int8_t idx)
{
  Pix pix;
  media_rcs rc;
  VCR_Index vcr_idx;
  VCRDisplay& display=VCRDisplay::instance();

  #ifdef DEBUG1
  cerr << "MBoneSession::set_index(" << media << ", " << int(idx) << ")\n";
  #endif

  if (nb_of_media()==0) return "";
  pix = p2mbmList.first();
  int i;
  for (i = 1; i<media; i++) p2mbmList.next(pix);
  rc=p2mbmList(pix)->set_index(idx, &vcr_idx);
  switch (rc) {
  case MEDIA_OK:
    return idx2str(&vcr_idx, idx);
    break;
  case MEDIA_IDX_TOO_LATE:
    #ifdef DEBUG1
    cerr << "MBoneSession::set_index() Cannot set index: " << int(idx) << ".\ntimestamp " <<
             display.get_time() << " ms is too late for media " << media <<
	     ".\nmedia is only " << p2mbmList(pix)->get_length() << " ms long.\n\n";
    #endif
    return "";
  default:
    media_msg(rc,"MBoneSession::set_index()");
    return "";
    break;
  };
};

session_rcs MBoneSession::delete_index(int media, u_int32_t pos, u_int8_t idx)
{
  Pix pix;
  media_rcs rc;
  VCRTcl& tcl=VCRTcl::instance();

  #ifdef DEBUG1
  cerr << "MBoneSession::delete_index(" << media << ", " << pos << ", " << int(idx) << ")\n";
  #endif

  if ((nb_of_media()==0) || (media==0)) return SESSION_EMPTY;
  if (session_state!=IDLE) {
    tcl.ErrorMsg(" Sorry, you cannot delete indices \nwhile playing/recording.");
    return SESSION_ERROR;
  };
  pix = p2mbmList.first();
  int i;
  for (i = 1; i<media; i++) p2mbmList.next(pix);
  rc=p2mbmList(pix)->delete_index(pos, idx);
  switch (rc) {
  case MEDIA_OK:
    return SESSION_OK;
    break;
  default:
    media_msg(rc,"MBoneSession::delete_index()");
    return SESSION_ERROR;
    break;
  };
};



session_rcs MBoneSession::delete_all_indices(int media,  u_int8_t idx, int start)
{
  Pix pix;
  media_rcs rc;
  VCRTcl& tcl=VCRTcl::instance();

  #ifdef DEBUG1
  cerr << "MBoneSession::delete_all_indices(" << media << ", " << int(idx) << ")\n";
  #endif

  if ((nb_of_media()==0) || (media==0)) return SESSION_EMPTY;
  if (session_state!=IDLE) {
    tcl.ErrorMsg(" Sorry, you cannot delete indices \nwhile playing/recording.");
    return SESSION_ERROR;
  };
  pix = p2mbmList.first();
  int i;
  for (i = 1; i<media; i++) p2mbmList.next(pix);
  rc=p2mbmList(pix)->delete_all_indices(idx, start);
  switch (rc) {
  case MEDIA_OK:
    return SESSION_OK;
    break;
  default:
    media_msg(rc,"MBoneSession::delete_all_indices()");
    return SESSION_ERROR;
    break;
  };
};



String MBoneSession::get_first_index(int media, u_int8_t idx, int start)
{
  Pix pix;
  media_rcs rc;
  VCR_Index vcr_idx;

  #ifdef DEBUG2
  cerr << "MBoneSession::get_first_index(" << media << ", " << int(idx) << ")\n";
  #endif

  if ((nb_of_media()==0) || (media == 0)) return "";
  pix = p2mbmList.first();
  int i;
  for (i = 1; i<media; i++) p2mbmList.next(pix);

  rc=p2mbmList(pix)->get_first_index(&vcr_idx, idx, start);
  if (rc==MEDIA_OK) {
    return idx2str(&vcr_idx, idx);
  } else {
    return "";
  }
};

String MBoneSession::get_next_index(int media, u_int8_t idx)
{
  Pix pix;
  media_rcs rc;
  VCR_Index vcr_idx;

  #ifdef DEBUG2
  cerr << "MBoneSession::get_next_index(" << media << ")\n";
  #endif

  /* ... give X a chance ... */
  Tk_DoOneEvent(TK_X_EVENTS|TK_DONT_WAIT);

  if ((nb_of_media()==0) || (media == 0)) return "";
  pix = p2mbmList.first();
  int i;
  for (i = 1; i<media; i++) p2mbmList.next(pix);

  rc=p2mbmList(pix)->get_next_index(&vcr_idx);
  if (rc==MEDIA_OK) {
    return idx2str(&vcr_idx, idx);
  } else {
    return "";
  }
};

session_rcs MBoneSession::goto_index(int direction, int media, u_int8_t idx)
{
  media_rcs rc;
  Pix pix;
  VCRDisplay& display=VCRDisplay::instance();
  VCRTcl& tcl=VCRTcl::instance();

  #ifdef DEBUG1
  cerr << "MBoneSession::goto_index(" << direction << ", " << media << ", "
       << int(idx) << ")\n";
  #endif

  if (nb_of_media()==0) return SESSION_EMPTY;
  if (session_state == RECORDING) {
    return SESSION_ERROR;
  };

  pix = p2mbmList.first();
  int i;
  for (i = 1; i<media; i++) p2mbmList.next(pix);
  if (direction > 0) {
    rc=p2mbmList(pix)->goto_next_index(idx);
  } else {
    rc=p2mbmList(pix)->goto_prev_index(idx);
  }
  switch (rc) {
  case MEDIA_OK:
    session_time=p2mbmList(pix)->get_ts();
    goto_ts(session_time);
    display.set(session_time,session_speed);
    tcl.set_msg("Jumped to index.", 3000);
    return SESSION_OK;
    break;
  case MEDIA_IDX_NOT_FOUND:
    #ifdef DEBUG1
    cerr << "MBoneSession::goto_index(): media returns: IDX_NOT_FOUND.\n";
    #endif
    tcl.set_msg("Index not found.", 3000);
    if (direction > 0) {
      goto_end();
    } else {
      goto_start();
    }
    return SESSION_IDX_NOT_FOUND;
    break;
  default:
    media_msg(rc,"MBoneSession::goto_index()");
    return SESSION_ERROR;
    break;
  };

  return SESSION_OK;
};

session_rcs MBoneSession::goto_blank(int direction, int media, u_int32_t blank_time)
{
  media_rcs rc;
  Pix pix;
  VCRDisplay& display=VCRDisplay::instance();

  #ifdef DEBUG1
  cerr << "MBoneSession::goto_blank(" << direction << ", " << media << ", "
       << blank_time << ")\n";
  #endif

  if (nb_of_media()==0) return SESSION_EMPTY;
  if (session_state == RECORDING) {
    return SESSION_ERROR;
  };

  pix = p2mbmList.first();
  int i;
  for (i = 1; i<media; i++) p2mbmList.next(pix);

  if (direction > 0) {
    rc=p2mbmList(pix)->goto_next_blank(blank_time);
  } else {
    rc=p2mbmList(pix)->goto_prev_blank(blank_time);
  };

  u_int32_t ts;
  VCRTcl& tcl=VCRTcl::instance();
  char msg[100];
  
  switch (rc) {
  case MEDIA_OK:
    ts=p2mbmList(pix)->get_ts();
    sprintf(msg,"Blank skiped %d secs.", int(int(ts-session_time)/1000));
    tcl.set_msg(msg, 3000);
    session_time=ts;
    goto_ts(session_time);
    display.set(session_time,session_speed);
    return SESSION_OK;
    break;
  case MEDIA_IDX_NOT_FOUND:
    #ifdef DEBUG1
    cerr << "MBoneSession::goto_blank(): media returns: IDX_NOT_FOUND.\n";
    #endif
    tcl.set_msg("No more blank gaps found.", 3000);
    if (direction > 0) {
      goto_end();
    } else {
      goto_start();
    }
    return SESSION_IDX_NOT_FOUND;
    break;
  default:
    media_msg(rc,"MBoneSession::goto_blank()");
    return SESSION_ERROR;
    break;
  };

  return SESSION_OK;
};

session_rcs MBoneSession::set_auto_blank(int media_nb, u_int32_t time)
{
  #ifdef DEBUG1
  cerr << "MBoneSession::set_auto_blank(" <<  media_nb << ", " << time << ")\n";
  #endif

  if (media_nb==0) {
    blank_skip_media=0;
    blank_skip_time=time;
  } else {
    blank_skip_media = p2mbmList.first();
    int i;
    for (i=1; i<media_nb; i++) p2mbmList.next(blank_skip_media);
    blank_skip_time=time;
  }
  return SESSION_OK;
}

session_rcs MBoneSession::set_rec_timeout(int timeout)
{
  Pix pix;
  #ifdef DEBUG1
  cerr << "MBoneSession::set_rec_timeout(" << timeout << ")\n";
  #endif

  for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
  {
    p2mbmList(pix)->set_rec_timeout(timeout);
  };

  return SESSION_OK;
}


session_rcs MBoneSession::goto_start()
{
  media_rcs rc;
  Pix pix;
  VCRDisplay& display=VCRDisplay::instance();

  #ifdef DEBUG1
  cerr << "MBoneSession::goto_start()\n";
  #endif

  if (nb_of_media()==0) return SESSION_EMPTY;

  for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
  {
    rc=p2mbmList(pix)->goto_start();
    if (rc>MEDIA_OK) media_msg(rc,"MBoneSession::goto_start()");
  };
  session_time=0;
  display.set(session_time, session_speed);
  return SESSION_OK;
};

int MBoneSession::is_writable()
{
  Pix pix;

  if (nb_of_media()==0) return 0;

  for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
  {
    if (p2mbmList(pix)->is_writable()==0) return 0;
  };
  return 1;
}

u_int32_t MBoneSession::get_session_ts()
{
  u_int32_t ts=0;
  u_int32_t max=0;
  u_int32_t check=0;
  int end=0;
  Pix pix;

  #ifdef DEBUG1
  cerr << "MBoneSession:get_session_ts()" << endl;
  #endif

  if (nb_of_media()==0) return 0;

  for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
  {
    ts=p2mbmList(pix)->get_ts();
    check=p2mbmList(pix)->get_length();
    #ifdef DEBUG2
    cerr << "MBoneSession:get_session_ts() ts: " << ts << " max: " << max << endl;
    #endif
    if (ts!=check && ts>max) max=ts; else end++;
  };
  if (end==nb_of_media()) max=ts;
  return max;
}

u_int32_t MBoneSession::get_media_ts(int media)
{
  Pix pix;

  #ifdef DEBUG1
  cerr << "MBoneSession::get_media_ts(" << media << ")\n";
  #endif

  if (nb_of_media()==0) return 0;
  if (media==0) return get_session_ts();
  pix = p2mbmList.first();
  int i;
  for (i = 1; i<media; i++) p2mbmList.next(pix);
  return p2mbmList(pix)->get_ts();
};


u_int32_t MBoneSession::get_session_length()
{
  u_int32_t max=0,l=0;
  Pix pix;

  #ifdef DEBUG1
  cerr << "MBoneSession::get_session_length()\n";
  #endif

  if (nb_of_media()==0) return 0;

  for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
  {
    l=p2mbmList(pix)->get_length();
    if (max<l) max=l;
  };
  return max;
};

u_int32_t MBoneSession::get_media_length(int media)
{

  Pix pix;

  #ifdef DEBUG1
  cerr << "MBoneSession::get_media_length(" << media << ")\n";
  #endif

  if (nb_of_media()==0) return SESSION_EMPTY;
  if (media == 0) return get_session_length();
  pix = p2mbmList.first();
  int i;
  for (i = 1; i<media; i++) p2mbmList.next(pix);
  return p2mbmList(pix)->get_length();
};

void MBoneSession::event_hdlr(int media_idx, media_rcs event, u_int32_t cargo)
{
  Pix pix;
  VCRDisplay& display=VCRDisplay::instance();
  VCRTcl& tcl=VCRTcl::instance();

  switch (event)
  {
  case MEDIA_RECORDING_STARTED:
    #ifdef DEBUG1
    cerr << "MBoneSession::event_hdlr(): event (MEDIA_RECORDING_STARTED)\n";
    #endif
    recording_started=1;
    display.start(session_time, 1, 1);
    tcl.set_msg("Received first data, recording started...", 3000);
    for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
    {
      p2mbmList(pix)->set_recording_started(1);
    };
    break;
  case MEDIA_EOF:
    #ifdef DEBUG1
    cerr << "MBoneSession::event_hdlr(): event (MEDIA_EOF).\n";
    #endif
    nb_of_active_media--;
    if (nb_of_active_media==0) {
      stop();
      goto_end();
      if (session_length && tcl.int_var("session(loop_mode)")) {
	goto_start();
	play();
      } else {
	tcl.release_button();
	tcl.set_msg("End of session reached.",3000);
      };
    };
    break;
  case MEDIA_BOF:
    #ifdef DEBUG1
    cerr << "MBoneSession::event_hdlr(): event (MEDIA_BOF).\n";
    #endif
    if (nb_of_active_media>0) nb_of_active_media--;
    break;
  case MEDIA_ERROR:
    #ifdef DEBUG1
    cerr << "MBoneSession::event_hdlr(): event (MEDIA_ERROR).\n";
    #endif
    tcl.release_button();
    stop();
    break;
  case MEDIA_GONE:
    #ifdef DEBUG1
    cerr << "MBoneSession::event_hdlr(): event (MEDIA_GONE).\n";
    #endif
    if (nb_of_active_media>0) nb_of_active_media--;
    break;
  case MEDIA_UPDATE_TIME:
    #ifdef DEBUG1
    cerr << "MBoneSession::event_hdlr(): event (MEDIA_UPDATE_TIME).\n";
    #endif
    session_time=cargo;
    if (session_time>session_length) session_length=session_time;
    display.set_time(session_time, session_speed);
    break;
  case MEDIA_NO_DATA_TIMEOUT:
    int force_timeout,ft;
    Pix pix;
    #ifdef DEBUG1
    cerr << "MBoneSession::event_hdlr(): event (MEDIA_NO_DATA_TIMEOUT).\n";
    #endif

    force_timeout=1;
    for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
	{
	    ft=p2mbmList(pix)->is_timed_out();
	    if (ft==0) force_timeout=0;
	};
    if (force_timeout) {
      stop();
      rec(0);
      VCRTcl& tcl=VCRTcl::instance();
      tcl.set_msg("Recording timed out, waiting for next data...", 0);
    };
    break;
  default:
    cerr << "MBoneSession::event_hdlr(): unknown event (" << event << ").\n";
    break;
  };
  return;
};



void MBoneSession::ttimeout()
{
  MediaSchedule ms;
  int32_t signed_ts;
  u_int32_t ts;
  media_rcs rc;
  VCRDisplay& display=VCRDisplay::instance();
  VCRTcl& tcl=VCRTcl::instance();

  #ifdef DEBUG2
  cerr << "MBoneSession::ttimeout().\n";
  #endif

  switch (session_state)
  {
  case PLAYING:
    while (1) {

      /* ... give X a chance ... */
      Tk_DoOneEvent(TK_X_EVENTS|TK_DONT_WAIT);

      if (media_sched.length()==0) return;

      ms=media_sched.deq();
      p2mbmList(ms.pix)->send_pending_sync_unit();
      last_ts=ms.ts;
      rc=p2mbmList(ms.pix)->read_next_sync_unit(&ts, session_speed);

      #ifdef DEBUG2
      cerr << "MBoneSession::ttimeout("<<p2mbmList(ms.pix)->get_media_name()<<
	      ") read_next_sync_unit: " << int(rc) << " ts: " << ts << "\n";
      #endif

      switch (rc) {

      case MEDIA_OK:
	session_time=display.get_time();

	ms.ts=ts;
	media_sched.enq(ms);

	/* send packets with the same ts as the last one immediately */
	if (ms.ts == last_ts) {
          #ifdef DEBUG2
	  cerr << "MBoneSession::ttimeout("<<p2mbmList(ms.pix)->get_media_name()<<
	          ") ts == last_media_ts == " << last_ts << "\n";
          #endif
	  continue;
	};

	/* check if we have to blank skip */
	if (blank_skip_time) {
	  if ((blank_skip_media==ms.pix) && (ms.ts>session_time) &&
	      ((ms.ts-session_time)>=blank_skip_time)) {
            #ifdef DEBUG1
	    cerr << "MBoneSession::ttimeout("<< p2mbmList(ms.pix)->get_media_name() <<
	    ") BLANK_SKIP " << ms.ts-session_time << " ms.\n";
            #endif
	    char msg[100];
	    sprintf(msg,"Blank skiped %d secs.", int(int(ms.ts-session_time)/1000));
	    tcl.set_msg(msg, 3000);
	    goto_ts(p2mbmList(ms.pix)->get_ts());
	    continue;
	  };
	};

	break;

      case MEDIA_EOF:
        #ifdef DEBUG1
	cerr << "MBoneSession::ttimeout("<<p2mbmList(ms.pix)->get_media_name()<<
	        ") read_next_sync_unit() returns MEDIA_EOF.\n";
        #endif
	nb_of_active_media--;
	if ((nb_of_active_media==0) || (media_sched.length()==0)) {
	  display.stop();
	  stop();
	  goto_end();
	  if (session_length && tcl.int_var("session(loop_mode)")) {
	    goto_start();
	    play();
	  } else {
	    tcl.set_msg("End of session reached.",3000);
	    tcl.release_button();
	  };
	  return;
	} else {
	  session_time=display.get_time();
	};
	break;

      case MEDIA_ERROR:
        #ifdef DEBUG1
	cerr << "MBoneSession::ttimeout("<<p2mbmList(ms.pix)->get_media_name()<<
	        ") read_next_sync_unit() returns MEDIA_ERROR.\n";
        #endif
	media_sched.clear();
	stop();
	tcl.release_button();
	tcl.ErrorMsg(p2mbmList(ms.pix)->get_error_msg());
	return;

      default:
        #ifdef DEBUG1
	cerr << "MBoneSession::ttimeout("<<p2mbmList(ms.pix)->get_media_name()<<
	        ") unknow return code from read_next_sync_unit(): " << int(rc) << ".\n";
        #endif
	media_sched.clear();
	stop();
	tcl.release_button();
	tcl.ErrorMsg(p2mbmList(ms.pix)->get_error_msg());
	return;
      };

      /* here we calculate when to send the next packet */
      signed_ts=int32_t(media_sched.front().ts)-int32_t(session_time);

      #ifdef DEBUG2
      cerr << "MBoneSession::ttimeout("<<p2mbmList(ms.pix)->get_media_name()<<
	      ") next packet scheduled in "<<int(signed_ts/session_speed)<<" ms\n" <<
	      "                    session time: " << session_time <<"\n"<<
	      "                 msched.length(): " << media_sched.length() << "\n"<<
	      "          media_sched.front().ts: " << media_sched.front().ts << "\n"<<
	      "       media_sched.front().media: " <<
	      p2mbmList(media_sched.front().pix)->get_media_name() << "\n";
      #endif

      if (signed_ts>0) {
	/* and we actually have to take the session_speed into account */
	session_timer.msched(int(signed_ts/session_speed));
	return;
      };

    }; /* while */
    break;
  case RECORDING:
    break;
  case FF:
    ff(ff_rew_factor);
    break;
  case REW:
    rew(ff_rew_factor);
    break;
  default:
    display.set(get_session_ts(),session_speed);
  };
};


void MBoneSession::media_msg(media_rcs mrc, String msg, u_int32_t x)
{
  String my_msg;
  if (msg.length()==0) my_msg="MBoneSession::media_msg "; else my_msg=msg;
  switch (mrc) {
  case MEDIA_LAST_PACKET:
    #ifdef DEBUG1
    cerr << my_msg << " returns MEDIA_LAST_PACKET.\n";
    #endif
    break;
  case MEDIA_OK:
    #ifdef DEBUG1
    cerr << my_msg << " returns MEDIA_OK.\n";
    #endif
    break;
  case MEDIA_BOF:
    #ifdef DEBUG1
    cerr << my_msg << " returns MEDIA_BOF.\n";
    #endif
    break;
  case MEDIA_EOF:
    #ifdef DEBUG1
    cerr << my_msg << " returns MEDIA_EOF.\n";
    #endif
    break;
  case MEDIA_PACKET_PENDING:
    #ifdef DEBUG1
    cerr << my_msg << " returns MEDIA_PACKET_PENDING.\n";
    #endif
    break;
  case MEDIA_NO_PACKET_PENDING:
    #ifdef DEBUG1
    cerr << my_msg << " returns MEDIA_NO_PACKET_PENDING.\n";
    #endif
    break;
  case MEDIA_NOT_READY:
    #ifdef DEBUG1
    cerr << my_msg << " returns MEDIA_NOT_READY.\n";
    #endif
    break;
  case MEDIA_INVALID_STATE:
    #ifdef DEBUG1
    cerr << my_msg << " returns MEDIA_INVALID_STATE.\n";
    #endif
    break;
  case MEDIA_ERROR:
    #ifdef DEBUG1
    cerr << my_msg << " returns MEDIA_ERROR.\n";
    #endif
    break;
  case MEDIA_GONE:
    #ifdef DEBUG1
    cerr << my_msg << " returns MEDIA_GONE.\n";
    #endif
    break;
  default:
    #ifdef DEBUG1
    cerr << my_msg << " Unknow return code ("<<mrc<<").\n";
    #endif
    break;
  };
};

void MBoneSession::cancel_operation(Pix j, media_states state)
{
  media_rcs rc;
  Pix pix;
  VCRTcl& tcl=VCRTcl::instance();

  for (pix = p2mbmList.first(); (pix<j && pix!=0); p2mbmList.next(pix))
  {
    rc=p2mbmList(pix)->media_change_state(state);
    if (rc>MEDIA_OK) media_msg(rc,"MBoneSession::cancel_operation()");
  };
  old_state=session_state;
  session_state=state;
  tcl.release_button();
};

session_rcs MBoneSession::set_rec_enabled(int rec_enabled)
{
  media_rcs rc;
  int error=0;
  Pix pix;
  VCRTcl& tcl=VCRTcl::instance();

  if (!is_writable() && rec_enabled==1) return SESSION_IS_READ_ONLY;

  if (rec_enabled==recording_enabled) return SESSION_OK;

  recording_enabled=rec_enabled;
  for (pix = p2mbmList.first(); pix!=0; p2mbmList.next(pix))
  {
    rc=p2mbmList(pix)->set_rec_enabled(rec_enabled);
    if (rc>MEDIA_OK) {
      media_msg(rc,"MBoneSession::rec_enabled()");
      error++;
    };
  };
  if (error) {
    recording_enabled=0;
    return SESSION_ERROR;
  };
  if (recording_enabled) {
    tcl.set_msg("Recording enabled.",3000);
  } else {
    tcl.set_msg("Recording disabled.",3000);
  };
  return SESSION_OK;
};


session_rcs MBoneSession::read_session_vars()
{
  int		i,nom;
  char		media_var[500];

  media_fmts	media_type=RTP;
  String	media_protocol;
  String	media_ip;
  int		media_port=0;
  u_int8_t	media_ttl=0;
  u_int16_t	media_cid=0;
  String	media_nb;
  String	media_file;
  String	session_file;
  int32_t	blank_time;
  int		idx_media;
  int		session_read_only;

  VCRTcl& tcl=VCRTcl::instance();
  VCRDisplay& display = VCRDisplay::instance();

  if (nb_of_media()>0) {
    error_msg="MBoneSession::read_session_vars(): session already open.";
    return SESSION_IS_OPEN;
  };

  #ifdef DEBUG1
  cerr << "MBoneSession::read_session_vars():\n";
  #endif

  ttl_warning_sent=0;
  nb_of_ttl_higher_16=0;
  session_speed=1;
  blank_skip_media=0;
  blank_skip_time=0;
  session_read_only=0;

  session_data_dir=tcl.var("session(data_dir)");
  #ifdef DEBUG2
  cerr << "session(data_dir): " << session_data_dir << "\n";
  #endif

  session_file=tcl.var("session(file_prefix)");
  #ifdef DEBUG2
  cerr << "session(file_prefix): " << session_file << "\n";
  #endif

  nom=tcl.int_var("session(nb_of_media)");
  #ifdef DEBUG2
  cerr << "session(nb_of_media): " << nom << "\n";
  #endif

  session_max_ttl=tcl.int_var("session(max_ttl)");
  #ifdef DEBUG2
  cerr << "session(max_ttl): " << int(session_max_ttl) << "\n";
  #endif

  session_read_only=(strlen(tcl.var("session(read_only)"))!=0);
  #ifdef DEBUG2
  cerr << "session(read_only): " << session_read_only << "\n";
  #endif

  blank_time=tcl.int_var("auto_blank_skip_time");
  #ifdef DEBUG2
  cerr << "blank_time: " << blank_time << "\n";
  #endif

  idx_media=tcl.int_var("session(index_media)");
  #ifdef DEBUG2
  cerr << "idx_media: " << idx_media << "\n";
  #endif

  for (i=1;i<=nom;i++) {
    #ifdef DEBUG2
    cerr << "\nmedia #" << i << " :\n";
    #endif

    sprintf(media_var,"session(media_list,%d,media_type)",i);
    media_type=media_fmts(tcl.int_var(media_var));
    #ifdef DEBUG2
    cerr << "\tmedia_type: " << media_type << "\n";
    #endif

    sprintf(media_var,"protocols(%d)",media_type);
    media_protocol=tcl.var(media_var);
    #ifdef DEBUG2
    cerr << "\tmedia_protocol: " << media_protocol << "\n";
    #endif

    sprintf(media_var,"session(media_list,%d,media_ip)",i);
    media_ip=tcl.var(media_var);
    #ifdef DEBUG2
    cerr << "\tmedia_ip: " << media_ip << "\n";
    #endif

    sprintf(media_var,"session(media_list,%d,media_port)",i);
    media_port=tcl.int_var(media_var);
    #ifdef DEBUG2
    cerr << "\tmedia_port: " << media_port << "\n";
    #endif

    sprintf(media_var,"session(media_list,%d,media_ttl)",i);
    media_ttl=tcl.int_var(media_var);
    if (media_ttl>16) nb_of_ttl_higher_16++;
    #ifdef DEBUG2
    cerr << "\tmedia_ttl: " << int(media_ttl) << "\n";
    #endif

    sprintf(media_var,"session(media_list,%d,media_cid)",i);
    media_cid=tcl.int_var(media_var);
    #ifdef DEBUG2
    cerr << "\tmedia_cid: " << media_cid << "\n";
    #endif

    sprintf(media_var,"%d",i);
    media_nb=media_var;
    #ifdef DEBUG2
    cerr << "\tmedia_nb: " << media_nb << "\n";
    #endif

    media_file=session_data_dir+"/"+session_file+"_"+media_protocol+"_"+media_nb;
    #ifdef DEBUG2
    cerr << "\tmedia_file: " << media_file << "\n";
    #endif

    switch (media_type) {
    case VAT:  
    case RTP:
      P2RTPMedia prtp;
      prtp = new RTPMedia(media_file, session_read_only,
			  media_ip, media_port, media_cid, media_ttl, i);
      if (prtp->get_media_state()!=IDLE) {
	prtp->install_event_hdlr(this, &event_hdlr, p2mbmList.length()+1);
	error_msg="MBoneSession::read_session_vars():\ncreate media RTP: ";
	error_msg+=prtp->get_error_msg();
	delete prtp;
	tcl.ErrorMsg(error_msg);
	return SESSION_ERROR;
      };
      add_media(prtp);
      break;
    default:
      cerr << "MBoneSession::read_session_vars(): protocol not supported!\n";
      break;
    };
  };
  session_length=get_session_length();
  display.update_total_length(session_length);
  set_auto_blank(idx_media, blank_time);
  return SESSION_OK;
};

int MBoneSession::vcr_ctrl(ClientData cd, Tcl_Interp* t, int argc, char **argv)
{

  MBoneSession* mbp = (MBoneSession*)cd;
  VCRTcl& tcl=VCRTcl::instance();

  #ifdef DEBUG2
  int i;
  cerr << "MBoneSession::vcr_ctrl() cmd: ";
  for (i=0; i<argc; i++) {
    cerr << argv[i] << " ";
  };
  cerr << "\n";
  #endif

  if (strcmp(argv[1], "play") == 0) {
    tcl.release_button();
    mbp->stop();
    if (mbp->play()==SESSION_OK) {
      tcl.press_button("play");
      Tcl_AppendResult(t, "playing", 0);
    } else {
      Tcl_AppendResult(t, "error", 0);
    }
    return TCL_OK;


  } else if (strcmp(argv[1], "stop") == 0) {
    if (argc < 3) tcl.release_button();
    if (mbp->stop()==SESSION_OK) {
      Tcl_AppendResult(t, "idle", 0);
    } else {
      Tcl_AppendResult(t, "error", 0);
    }
    return TCL_OK;


  } else if (strcmp(argv[1], "rec") == 0) {
    int start_rec=0;
    if (argc == 3) sscanf(argv[2],"%d",&start_rec);
    tcl.release_button();
    if (mbp->rec(start_rec)==SESSION_OK) {
      tcl.press_button("rec");
      Tcl_AppendResult(t, "recording", 0);
    } else {
      Tcl_AppendResult(t, "error", 0);
    }
    return TCL_OK;


  } else if (strcmp(argv[1], "set_ff_rew_factor") == 0) {
    int factor=10;
    if (argc == 3) sscanf(argv[2],"%d",&factor);
    if (mbp->set_ff_rew_factor(factor)==SESSION_OK) {
      Tcl_AppendResult(t, "1", 0);
    } else {
      Tcl_AppendResult(t, "0", 0);
    }
    return TCL_OK;


  } else if (strcmp(argv[1], "ff") == 0) {
    int factor=10;
    if (argc == 3) sscanf(argv[2],"%d",&factor);
    tcl.release_button();
    mbp->stop();
    if (mbp->ff(factor)==SESSION_OK) {
      tcl.press_button("ff");
      Tcl_AppendResult(t, "forwarding", 0);
    } else {
      Tcl_AppendResult(t, "error", 0);
    }
    return TCL_OK;


  } else if (strcmp(argv[1], "rew") == 0) {
    int factor=10;
    if (argc == 3) sscanf(argv[2],"%d",&factor);
    tcl.release_button();
    mbp->stop();
    if (mbp->rew(factor)==SESSION_OK) {
      tcl.press_button("rew");
      Tcl_AppendResult(t, "rewinding", 0);
    } else {
      Tcl_AppendResult(t, "error", 0);
    }
    return TCL_OK;


  } else if (strcmp(argv[1], "get_ts") == 0) {
    char x[20];
    int media_nb = 0;
    if (argc < 2) return TCL_OK;
    if (argc == 3) sscanf(argv[2],"%d",&media_nb);
    sprintf(x,"%d",mbp->get_media_ts(media_nb));
    Tcl_AppendResult(t, x, 0);
    return TCL_OK;


  } else if (strcmp(argv[1], "get_length") == 0) {
    char x[20];
    int media_nb=0;
    if (argc==3) sscanf(argv[2],"%d",&media_nb);
    sprintf(x,"%d",mbp->get_media_length(media_nb));
    Tcl_AppendResult(t, x, 0);
    return TCL_OK;


  } else if (strcmp(argv[1], "goto") == 0) {
    u_int8_t idx=0;
    int index_level;
    int media_nb=0;
    int direction=0;
    u_int32_t time;
    session_rcs rc;


    if (strcmp(argv[2], "start") == 0) {
      if (mbp->get_state() == RECORDING) return TCL_OK;
      mbp->stop();
      mbp->goto_start();

    } else if (strcmp(argv[2], "end") == 0) {
      if (mbp->get_state() == RECORDING) return TCL_OK;
      mbp->stop();
      mbp->goto_end();

    } else if (strcmp(argv[2], "ts") == 0) {
      long pos=0;

      if (argc >= 4) sscanf(argv[3],"%ld",&pos); else return TCL_OK;
      if (mbp->get_state() == RECORDING) return TCL_OK;
      if (argc<5) mbp->stop();
      /* pos*1000 since pos is in secs and ts in ms */
      rc=mbp->goto_ts(pos*1000);
      if (rc != SESSION_OK) {
	tcl.release_button();
	Tcl_AppendResult(t, "error", 0);
	return TCL_OK;
      };

    } else if (strcmp(argv[2], "ms") == 0) {
      long pos=0;

      if (argc >= 4) sscanf(argv[3],"%ld",&pos); else return TCL_OK;
      if (mbp->get_state() == RECORDING) return TCL_OK;
      if (argc<5) mbp->stop();
      rc=mbp->goto_ts(pos);
      if (rc != SESSION_OK) {
	tcl.release_button();
	Tcl_AppendResult(t, "error", 0);
	return TCL_OK;
      };

    } else if ((strcmp(argv[2], "prev_index") == 0) ||
	       (strcmp(argv[2], "next_index") == 0)) {

      if (argc >= 5) {
	sscanf(argv[3],"%d",&media_nb);
	sscanf(argv[4],"%d",&index_level);
	idx = u_int8_t(index_level);
	if (media_nb == 0) {
	  Tcl_AppendResult(t, "1", 0);
	  return TCL_OK;
	}
      } else {
	Tcl_AppendResult(t, "0", 0);
	return TCL_OK;
      };
      if ( strcmp(argv[2], "prev_index") == 0) direction=0; else direction=1;

      mbp->stop();
      rc=mbp->goto_index(direction,media_nb,idx);
      if ((rc == SESSION_OK) && (argc == 6) && (strcmp(argv[5], "dbl") == 0))
        rc=mbp->goto_index(direction,media_nb,idx);

      if (direction) {
	tcl.evalc(".vcr.middle.right.bottom.skip_forw configure -relief raised");
      } else {
	tcl.evalc(".vcr.middle.right.bottom.skip_back configure -relief raised");
      };

      if (rc == SESSION_ERROR) {
	tcl.release_button();
	Tcl_AppendResult(t, "error", 0);
	return TCL_OK;
      } else if (rc == SESSION_IDX_NOT_FOUND) {
	Tcl_AppendResult(t, "not_found", 0);
      } else {
	Tcl_AppendResult(t, "found", 0);
      };

    } else if ((strcmp(argv[2], "prev_blank") == 0) ||
	       (strcmp(argv[2], "next_blank") == 0)) {

      if (argc<4) return TCL_OK;
      sscanf(argv[3],"%d",&media_nb);
      sscanf(argv[4],"%d",&time);
      if (media_nb == 0) {
	Tcl_AppendResult(t, "1", 0);
	return TCL_OK;
      }

      if ( strcmp(argv[2], "prev_blank") == 0) direction=0; else direction=1;

      mbp->stop();
      rc=mbp->goto_blank(direction, media_nb, time);

      if (rc == SESSION_ERROR) {
	tcl.release_button();
	Tcl_AppendResult(t, "error", 0);
	return TCL_OK;
      } else if (rc == SESSION_IDX_NOT_FOUND) {
	Tcl_AppendResult(t, "not_found", 0);
      } else {
	Tcl_AppendResult(t, "found", 0);
      };

    } else {

      Tcl_AppendResult(t, "\"",argv[0], " ", argv[1], " ", argv[2], "\": unknown command.",0);
      return TCL_ERROR;
    };

    return mbp->set_old_state(mbp, t);

  } else if (strcmp(argv[1], "index") == 0) {
    int media_nb;
    long start,pos;
    int index_level;
    static u_int8_t idx_mask;

    if (strcmp(argv[2], "get_first") == 0) {
      if (argc<6) return TCL_OK;
      sscanf(argv[3],"%d",&media_nb);
      sscanf(argv[4],"%d",&index_level);
      sscanf(argv[5],"%ld",&start);
      if (media_nb == 0) { return TCL_OK; }
      idx_mask = u_int8_t(index_level);
      Tcl_AppendResult(t, mbp->get_first_index(media_nb, idx_mask, start).chars(), 0);
      return TCL_OK;

    } else if (strcmp(argv[2], "get_next") == 0) {
      if (argc<4) return TCL_OK;
      sscanf(argv[3],"%d",&media_nb);
      if (media_nb == 0) { return TCL_OK; }
      Tcl_AppendResult(t, mbp->get_next_index(media_nb, idx_mask).chars(), 0);
      return TCL_OK;

    } else if (strcmp(argv[2], "set") == 0) {
      if (argc<5) return TCL_OK;

      sscanf(argv[3],"%d",&media_nb);
      sscanf(argv[4],"%d",&index_level);
      if (media_nb == 0) { return TCL_OK; }
      Tcl_AppendResult(t, mbp->set_index(media_nb, u_int8_t(index_level)).chars(), 0);
      return TCL_OK;

    } else if (strcmp(argv[2], "delete") == 0) {
      if (argc<6) return TCL_OK;
      sscanf(argv[3],"%d",&media_nb);
      sscanf(argv[4],"%ld",&pos);
      sscanf(argv[5],"%d",&index_level);
      if (media_nb == 0) {
	Tcl_AppendResult(t, "1", 0);
	return TCL_OK;
      }
      mbp->stop();
      if (mbp->delete_index(media_nb, pos, u_int8_t(index_level))==SESSION_OK) {
	Tcl_AppendResult(t, "1", 0);
      } else {
	Tcl_AppendResult(t, "0", 0);
      }
      mbp->set_old_state(mbp,t,0);
      return TCL_OK;

    } else if (strcmp(argv[2], "delete_all") == 0) {
      int start=0;
      if (argc<5) return TCL_OK;
      mbp->stop();
      sscanf(argv[3],"%d",&media_nb);
      sscanf(argv[4],"%d",&index_level);
      if (argc==6) sscanf(argv[4],"%d",&start);
      if (media_nb == 0) {
	Tcl_AppendResult(t, "1", 0);
	return TCL_OK;
      }
      mbp->delete_all_indices(media_nb, u_int8_t(index_level), start);
      Tcl_AppendResult(t, "1", 0);
      mbp->set_old_state(mbp,t,0);
      return TCL_OK;

    } else {
      Tcl_AppendResult(t, "\"",argv[0], " ", argv[1], " ", argv[2], "\": unknown command.",0);
      return TCL_ERROR;
    }
  } else if (strcmp(argv[1], "is_writable") == 0) {
    if (mbp->is_writable()) {
      Tcl_AppendResult(t, "1", 0);
    } else {
      Tcl_AppendResult(t, "0", 0);
    };
    return TCL_OK;


  } else if (strcmp(argv[1], "get_rec_enabled") == 0) {
    if (mbp->get_rec_enabled()) {
      Tcl_AppendResult(t, "1", 0);
    } else {
      Tcl_AppendResult(t, "0", 0);
    };
    return TCL_OK;


  } else if (strcmp(argv[1], "set_rec_enabled") == 0) {
    mbp->set_rec_enabled(1);
    if (mbp->get_rec_enabled()) {
      Tcl_AppendResult(t, "1", 0);
    } else {
      Tcl_AppendResult(t, "0", 0);
    };
    return TCL_OK;


  } else if (strcmp(argv[1], "set_rec_disabled") == 0) {
    mbp->set_rec_enabled(0);
    if (mbp->get_rec_enabled()) {
      Tcl_AppendResult(t, "1", 0);
    } else {
      Tcl_AppendResult(t, "0", 0);
    };
    return TCL_OK;


  } else if (strcmp(argv[1], "set_auto_blank_skip") == 0) {
    int media_nb=0;
    long time;

    if (argc<3) return TCL_OK;
    sscanf(argv[2],"%d",&media_nb);
    if (argc==4) sscanf(argv[3],"%ld",&time);
    mbp->set_auto_blank(media_nb, time);
    Tcl_AppendResult(t, "1", 0);
    return TCL_OK;

  } else if (strcmp(argv[1], "set_rec_timeout") == 0) {
    int rec_timeout=0;

    if (argc<3) return TCL_OK;
    sscanf(argv[2],"%d",&rec_timeout);
    mbp->set_rec_timeout(rec_timeout);
    Tcl_AppendResult(t, "1", 0);
    return TCL_OK;

  } else if (strcmp(argv[1], "no_auto_blank_skip") == 0) {
    mbp->set_auto_blank(0,0);
    Tcl_AppendResult(t, "1", 0);
    return TCL_OK;

  } else if (strcmp(argv[1], "set_ttl_warning") == 0) {
    int send_warning;
    if (argc<3) return TCL_OK;
    sscanf(argv[2],"%d",&send_warning);
    mbp->set_warning_sent(send_warning);
    Tcl_AppendResult(t, "1", 0);
    return TCL_OK;

  } else if (strcmp(argv[1], "speed") == 0) {
    float speed;
    if (argc == 3) sscanf(argv[2],"%f",&speed); else return TCL_OK;
    if (mbp->set_speed(speed)<0) {
      tcl.release_button();
      return TCL_OK;
    };
    char msg[100];
    sprintf(msg,"Set session speed to %1.1f.", speed);
    tcl.set_msg(msg, 3000);
    return TCL_OK;

  } else if (strcmp(argv[1], "get_error") == 0) {
    Tcl_AppendResult(t, mbp->get_error_msg(), 0);
    return TCL_OK;


  } else if (strcmp(argv[1], "read") == 0) {
    if (mbp->read_session_vars()==SESSION_OK) {
      Tcl_AppendResult(t, "1", 0);
    } else {
      Tcl_AppendResult(t, "0", 0);
    };
    return TCL_OK;


  } else if (strcmp(argv[1], "delete") == 0) {
    if (mbp->del_session() == SESSION_OK) {
      Tcl_AppendResult(t, "0", 0);
    } else {
      Tcl_AppendResult(t, "1", 0);
    };
    return TCL_OK;

  } else if (strcmp(argv[1], "mute_media") == 0) {
    int index;
    if (argc >= 3) sscanf(argv[2],"%d",&index); else return TCL_OK;

    if (mbp->set_mute_media(index,1) == SESSION_OK) {
      Tcl_AppendResult(t, "0", 0);
    } else {
      Tcl_AppendResult(t, "1", 0);
    };
    return TCL_OK;

  } else if (strcmp(argv[1], "unmute_media") == 0) {
    int index;
    if (argc >= 3) sscanf(argv[2],"%d",&index); else return TCL_OK;

    if (mbp->set_mute_media(index,0) == SESSION_OK) {
      Tcl_AppendResult(t, "0", 0);
    } else {
      Tcl_AppendResult(t, "1", 0);
    };
    return TCL_OK;

  } else {
    Tcl_AppendResult(t, "\"", argv[0], " ",argv[1], "\": unknown command.", 0);
    return TCL_ERROR;
  };
}

int MBoneSession::set_old_state(MBoneSession* mbp, Tcl_Interp* t, int set_result)
{

  VCRTcl& tcl=VCRTcl::instance();

  switch (mbp->get_old_state()) {
  case PLAYING:
    if (mbp->play()==SESSION_OK) {
      if (set_result) Tcl_AppendResult(t, "playing", 0);
    } else {
      mbp->stop();
      tcl.release_button();
      if (set_result) Tcl_AppendResult(t, "error", 0);
    };
    break;
  case FF:
    if (mbp->ff()==SESSION_OK) {
      if (set_result) Tcl_AppendResult(t, "forwarding", 0);
    } else {
      mbp->stop();
      tcl.release_button();
      if (set_result) Tcl_AppendResult(t, "error", 0);
    };
    break;
  case REW:
    if (mbp->rew()==SESSION_OK) {
      if (set_result) Tcl_AppendResult(t, "rewinding", 0);
    } else {
      mbp->stop();
      tcl.release_button();
      if (set_result) Tcl_AppendResult(t, "error", 0);
    };
    break;
  case IDLE:
    break;
  default:
    tcl.release_button();
    if (set_result) Tcl_AppendResult(t, "idle", 0);
    /* do nothing */
    break;
  };
  return TCL_OK;
};

String MBoneSession::idx2str(VCR_Index *vcr_idx, u_int8_t mask) {
  char entry[200];
  char desc[100];
  String str_entry;
  int index;
  u_int8_t test_idx;

  #ifdef DEBUG2
  cerr << "MBoneSession::idx2str(): ";
  #endif

  if (vcr_idx->idx==0) {
    #ifdef DEBUG2
    cerr << "NULL\n";
    #endif
    return "";
  }

  str_entry = "";
  test_idx=(vcr_idx->idx & mask);

  int hours=ms2hour(vcr_idx->ts);
  int mins=ms2mod_min(vcr_idx->ts);
  int secs=ms2mod_sec(vcr_idx->ts);

  while (test_idx) {
    if (test_idx & VCR_FILE_USER1) {
      sprintf(desc, "USER-INDEX (1)");
      test_idx ^= VCR_FILE_USER1;
      index=8;
    } else if (test_idx & VCR_FILE_USER2) {
      sprintf(desc, "USER-INDEX (2)");
      test_idx ^= VCR_FILE_USER2;
      index=16;
    } else if (test_idx & VCR_FILE_USER3) {
      sprintf(desc, "USER-INDEX (3)");
      test_idx ^= VCR_FILE_USER3;
      index=32;
    } else if (test_idx & VCR_FILE_USER4) {
      sprintf(desc, "USER-INDEX (4)");
      test_idx ^= VCR_FILE_USER4;
      index=64;
    } else if (test_idx & VCR_FILE_USER5) {
      sprintf(desc, "USER-INDEX (5)");
      test_idx ^= VCR_FILE_USER5;
      index=128;
    } else if (test_idx & VCR_FILE_NEWFMT) {
      sprintf(desc, "NEW-FORMAT");
      test_idx ^= VCR_FILE_NEWFMT;
      index=4;
    } else if (test_idx & VCR_FILE_NEWSP) {
      sprintf(desc, "NEW-SPEAKER");
      test_idx ^= VCR_FILE_NEWSP;
      index=2;
    } else if (test_idx & VCR_FILE_NEWTS) {
      sprintf(desc, "NEW-SYNC-UNIT");
      test_idx ^= VCR_FILE_NEWTS;
      index=1;
    } else {
      test_idx=0;
      index=0;
      return "";
    };
    sprintf(entry,"\"%ld %d %d%d:%d%d:%d%d %s\" ", long(vcr_idx->ts), index,
	    hours/10, hours%10, mins/10, mins%10, secs/10, secs%10, desc);
    str_entry += entry;
  };
  #ifdef DEBUG2
  cerr << str_entry << "\n";
  #endif
  return str_entry;
};



