/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

Copyright (c) 1995 International Computer Science Institute
All Rights Reserved.

Created by Wieland Holfelder (whd@icsi.berkeley.edu)
Tenet Group
International Computer Science Institute

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research and non-profit purposes,
without fee, and without a written agreement is hereby granted,
provided that the above copyright notice, this paragraph and the
following five paragraphs appear in all copies.

Permission to incorporate this software into commercial products may
be obtained from the International Computer Science Institute,
1947 Center Street, Suite 600, Berkeley, CA 94704.

IN NO EVENT SHALL THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE BE
LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE INTERNATIONAL
COMPUTER SCIENCE INSTITUTE  HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE SPECIFICALLY DISCLAIMS
ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE INTERNATIONAL
COMPUTER SCIENCE INSTITUTE  HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE,
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

The International Computer Science Institute will appreciate
receiving any enhancements and modifications of this software
made by others, and reserves the right to include them in future
releases. The Institute also recommends that this software not
be redistributed by third parties, but be obtained by all directly
from the ICSI server, so that all users of the software may
receive bug notifications and other relevant announcements.

The German Academic Exchange Service has supported the development
of this software. Their support is gratefully acknowledged.

* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* $Id: MBoneSession.h,v 1.1 1996/07/26 15:02:20 whd Exp $ */

#ifndef _MBoneSession_h
#define _MBoneSession_h

#include <Pix.h>
#include <DLList.h>
#include <String.h>

#include "MediaSchedule.PHPQ.h"
#include "SessionTimer.h"
#include "MBoneMedia.h"
#include "it_depends.h"

enum session_rcs {
  SESSION_EMPTY		= 0,
  SESSION_OK		= 1,
  SESSION_ERROR		= 2,
  SESSION_BOF		= 3,
  SESSION_EOF		= 4,
  SESSION_INVALID_STATE	= 5,
  SESSION_IS_OPEN	= 6,
  SESSION_NOT_OPEN	= 7,
  SESSION_IS_READ_ONLY	= 8,
  SESSION_IDX_NOT_FOUND	= 9
};

class MBoneSession {
public:
  MBoneSession();
  ~MBoneSession();

  static int vcr_ctrl(ClientData, Tcl_Interp*, int argc, char **argv);

  session_rcs 	add_media(P2MBoneMedia p2mbm);
  session_rcs	del_media(P2MBoneMedia p2mbm);
  session_rcs	set_mute_media(int, int);

  inline int 	nb_of_media(){ return p2mbmList.length(); };
  session_rcs	read_session_vars();
  session_rcs	del_session();

  session_rcs	play(float =0);
  session_rcs	rew(int =10);
  session_rcs	ff(int =10);
  session_rcs	rec(int =0);
  session_rcs	stop();
  session_rcs	goto_start();
  session_rcs	goto_end();
  session_rcs	goto_ts(u_int32_t =0);


  String	get_first_index(int =0, u_int8_t =0, int =0);
  String	get_next_index(int =0, u_int8_t =0);
  session_rcs	goto_index(int =1, int =0, u_int8_t =0);
  session_rcs	goto_blank(int =1, int =0, u_int32_t =1000);
  session_rcs	set_auto_blank(int =0, u_int32_t =0);
  session_rcs	set_rec_timeout(int =0);
  String	set_index(int =0, u_int8_t =0);
  session_rcs	delete_index(int =0, u_int32_t =0, u_int8_t =0);
  session_rcs	delete_all_indices(int =0, u_int8_t =0, int =0);

  session_rcs	set_rec_enabled(int);
  int		is_writable();

  u_int32_t	get_session_ts();
  u_int32_t	get_media_ts(int =0);

  u_int32_t	get_session_length();
  u_int32_t	get_media_length(int =0);

  inline session_rcs set_ff_rew_factor(int factor)
  {
    ff_rew_factor=factor;
  };

  inline int get_rec_enabled() {
    return recording_enabled;
  };

  inline int set_warning_sent(int ws) {
    return ttl_warning_sent=ws;
  };

  inline media_states get_state() {
    return session_state;
  };
  inline media_states get_old_state() {
    return old_state;
  };
  inline float get_speed() {
    return session_speed;
  };

  inline float set_speed(float speed =1) {
    if(speed==0) return session_speed;
    if ((session_state==PLAYING) && (session_speed!=speed)) {
      stop();
      if (play(speed)!=SESSION_OK) {
	return -1;
      };
    } else {
      session_speed=speed;
    };
    return session_speed;
  };

  inline const char *get_error_msg() {
    return (const char*) error_msg;
  };

protected:
  String	error_msg;

private:
  DLList<P2MBoneMedia>  	p2mbmList;
  MediaSchedulePHPQ		media_sched;
  SessionTimer			session_timer;
  String			session_data_dir;
  media_states 			session_state;
  media_states 			old_state;
  float 			session_speed;
  u_int32_t			session_time;
  u_int32_t			session_length;
  u_int32_t			last_ts;
  int				session_max_ttl;
  int				recording_enabled;
  int				nb_of_active_media;
  int				nb_of_inactive_media;
  int				recording_started;
  Pix				blank_skip_media;
  u_int32_t			blank_skip_time;
  int				ff_rew_factor;
  int				ttl_warning_sent;
  int				nb_of_ttl_higher_16;

  void	event_hdlr(int, media_rcs, u_int32_t =0);
  void	ttimeout();
  void 	media_msg(media_rcs, String ="", u_int32_t =0);
  void 	cancel_operation(Pix, media_states =IDLE);
  String idx2str(VCR_Index *, u_int8_t =0);
  int set_old_state(MBoneSession*, Tcl_Interp*, int =1);
};

#endif


