/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 
Copyright (c) 1995 International Computer Science Institute
All Rights Reserved.

Created by Wieland Holfelder (whd@icsi.berkeley.edu)
Tenet Group
International Computer Science Institute

Permission to use, copy, modify, and distribute this software and its 
documentation for educational, research and non-profit purposes, 
without fee, and without a written agreement is hereby granted, 
provided that the above copyright notice, this paragraph and the 
following five paragraphs appear in all copies.  

Permission to incorporate this software into commercial products may 
be obtained from the International Computer Science Institute,
1947 Center Street, Suite 600, Berkeley, CA 94704.  

IN NO EVENT SHALL THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE BE 
LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR 
CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE 
USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE INTERNATIONAL
COMPUTER SCIENCE INSTITUTE  HAS BEEN ADVISED OF THE POSSIBILITY OF 
SUCH DAMAGE.

THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE SPECIFICALLY DISCLAIMS 
ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE 
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE INTERNATIONAL
COMPUTER SCIENCE INSTITUTE  HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, 
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

The International Computer Science Institute will appreciate 
receiving any enhancements and modifications of this software 
made by others, and reserves the right to include them in future 
releases. The Institute also recommends that this software not 
be redistributed by third parties, but be obtained by all directly 
from the ICSI server, so that all users of the software may 
receive bug notifications and other relevant announcements.

The German Academic Exchange Service has supported the development 
of this software. Their support is gratefully acknowledged.

* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static const char rcsid[]="$Id: MCSocket.cc,v 1.3 1996/08/09 13:42:45 whd Exp $";

extern "C" {
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/time.h>
}
#include "MCSocket.h"

MCSocket::MCSocket() {

  #ifdef DEBUG1
  cerr << "MCSocket::MCSocket()\n";
  #endif
  my_state	=MCS_CLOSED;
  ipaddr	=DEFAULT_ADDR; 
  port		=DEFAULT_PORT; 
  ttl		=DEFAULT_TTL;
  loop		=DEFAULT_LOOP;
  reuseaddr	=DEFAULT_REUSE;

  read_hdlr_mask=0;
  write_hdlr_mask=0;

  tv= &tvs; 
  tv->tv_sec=0;
  tv->tv_usec=0;
  FD_ZERO(&read_fds);
  use_mcast=1;

};

MCSocket::~MCSocket() {
  #ifdef DEBUG1
  cerr << "MCSocket::~MCSocket()\n";
  #endif
  close();
};
  
int MCSocket::set_ttl(u_int8_t t) {
  ttl=t;
  return (is_open()?reopen():1);
};

int MCSocket::set_loop(u_int8_t l) {
  if ((l==0)||(l==1)) {
    loop=l;
    return (is_open()?reopen():1);
  } else {
    cerr << "invalid parameter \"" << l << "\" for set_loop()\n";
    return -1;
  };
};

int MCSocket::set_reuse(int r) {
  if ((r==0)||(r==1)) {
    reuseaddr=r;
    return (is_open()?reopen():1);
  } else {
    cerr << "invalid parameter \"" << r << "\" for set_reuse()\n";
    return -1;
  };
};

int MCSocket::set_ipaddr(String i) { 

  unsigned int i1, i2, i3, i4;
  sscanf((const char *)i,"%u.%u.%u.%u",&i1,&i2,&i3,&i4);
  use_mcast = 0;
  if (i1 < 224) {
    ipaddr=i;
  } else {
    if (!((i1&0xE0)==0xE0)) {
      cerr << "MCSocket.cc:set_ipaddr(): " <<i<<" is not a valid multicast address (because "<<
	i1<<"&0xE0!=0xE0)\n";
      return -1;
    } else if (!((i1&0x10)==0)) {
      cerr << "MCSocket.cc:set_ipaddr(): " <<i<<" is not a valid multicast address (because "<<
	i1<<"&0x10!=0)\n";
      return -1;
    } else if (!((i2>=0) && (i2<=255))) {
      cerr << "MCSocket.cc:set_ipaddr(): " <<i<<" is not a valid multicast address (because "<<
	i2<<"not between 0 and 255)\n";
      return -1;
    } else if (!((i3>=0) && (i3<=255))) {
      cerr << "MCSocket.cc:set_ipaddr(): " <<i<<" is not a valid multicast address (because "<<
	i3<<"not between 0 and 255)\n";
      return -1;
    } else if (!((i4>=0) && (i4<=255))) {
      cerr << "MCSocket.cc:set_ipaddr(): " <<i<<" is not a valid multicast address (because "<<
	i4<<"not between 0 and 255)\n";
      return -1;
    } else if (!(i2+i3!=0)) {
      cerr << "MCSocket.cc:set_ipaddr(): " <<i<<" is not a valid multicast address (because "<<
	i2<<"+"<<i3<<"==0)\n";
      return -1;
    } else {
      use_mcast = 1;
      ipaddr=i;
    }
  }

  return (is_open()?reopen():1);

};

int MCSocket::set_port(int p) {
  port=p;
  return (is_open()?reopen():1);
};

int MCSocket::open(String i, int p, u_int8_t t, u_int8_t l, int r) {

  #ifdef DEBUG4
  cerr << "MCSocket::open() ...\n"; 
  #endif

  if (is_open()) {
    #ifdef DEBUG4
    cerr << "MCSocket::open() socket already open.\n"; 
    #endif
    return -1;
  }
  if (set_ipaddr(i)<0) {
    #ifdef DEBUG4
    cerr << "MCSocket::open() set_ipaddr failed.\n"; 
    #endif
    return -1;
  }
  if (set_port(p)<0) {
    #ifdef DEBUG4
    cerr << "MCSocket::open() set_port() failed.\n"; 
    #endif
    return -1;
  }
  if (set_ttl(t)<0) {
    #ifdef DEBUG4
    cerr << "MCSocket::open() set_ttl() failed.\n"; 
    #endif
    return -1;
  }
  if (set_loop(l)<0) {
    #ifdef DEBUG4
    cerr << "MCSocket::open() set_loop() failed.\n"; 
    #endif
    return -1;
  }
  if (set_reuse(r)<0) {
    #ifdef DEBUG4
    cerr << "MCSocket::open() set_reuse() failed.\n"; 
    #endif
    return -1;
  }

  if (open_read() && open_write()) {
    /* this was just a test if we actually can open this socket... */
    my_state=MCS_OPEN;
    /* but we do not want to deal with data we don't need yet */
    close_read();
    close_write();
    return 1;
  } else {
    close();
    return -1;
  };
};

int MCSocket::open_read() {
#ifndef NO_MULTICAST
  struct ip_mreq mreq;
#endif
  struct hostent *h;
  struct sockaddr_in addr, bind_addr;

  #ifdef DEBUG4
  cerr << "MCSocket::open_read() ...\n"; 
  #endif

  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);

  if ((addr.sin_addr.s_addr = inet_addr((const char*) ipaddr)) == -1) {
    if ((h = gethostbyname((const char *) ipaddr)) == NULL) {
      perror("invalid hostname:");
      return -1;
    }
    addr.sin_addr.s_addr = *((int *) h->h_addr_list[0]);
  }
  
  /* Open multicast socket for receiving... */
  if((mcs_read = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    perror("can't open socket mcs_read");
    return -1;
  }

#ifndef NO_MULTICAST
  if (use_mcast) {
    if(IN_MULTICAST(ntohl(addr.sin_addr.s_addr))) {
      if (setsockopt(mcs_read, IPPROTO_IP, IP_MULTICAST_TTL, (char *)&ttl, sizeof(ttl)) < 0) {
	perror("mcs_read: setsockopt ttl");
	return -1;
      }
      mreq.imr_multiaddr = addr.sin_addr;
      mreq.imr_interface.s_addr = htonl(INADDR_ANY);
      if (setsockopt(mcs_read, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&mreq, sizeof(mreq)) < 0) {
	perror("mcs_read: setsockopt ip_add_membership");
	return -1;
      }
    } 
    
    if (setsockopt(mcs_read, IPPROTO_IP, IP_MULTICAST_LOOP, (char *)&loop, sizeof(loop)) < 0) {
      perror("mcs_read: setsockopt multicastloop");
      return -1;
    }
  } 
  if (setsockopt(mcs_read, SOL_SOCKET, SO_REUSEADDR, (char *)&reuseaddr, sizeof(reuseaddr)) < 0) {
    perror("mcs_read: setsockopt reuseaddr");
    return -1;
  }
#endif
  
  bind_addr.sin_family = AF_INET;
  bind_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  bind_addr.sin_port = htons(port);
  
  if(bind(mcs_read, (struct sockaddr *) &bind_addr, sizeof(bind_addr)) < 0) {
    perror("bind mcs_read");
    return -1;
  }
  
  FD_ZERO(&read_fds);
  FD_SET(mcs_read, &read_fds);
  mcs_read_tcl=Tcl_GetFile((ClientData) mcs_read, TCL_UNIX_FD);
  read_state=MCS_OPEN;
  return 1;
};

int MCSocket::open_write() {
#ifndef NO_MULTICAST
  struct ip_mreq mreq;
#endif
  struct hostent *h;
  struct sockaddr_in addr;
  
  #ifdef DEBUG4
  cerr << "MCSocket::open_write() ...\n"; 
  #endif

  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);

  if ((addr.sin_addr.s_addr = inet_addr((const char*) ipaddr)) == -1) {
    if ((h = gethostbyname((const char *) ipaddr)) == NULL) {
      perror("invalid hostname:");
      return -1;
    }
    addr.sin_addr.s_addr = *((int *) h->h_addr_list[0]);
  };

  /* Open multicast socket for sending... */
  if((mcs_write = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
    perror("can't open socket mcs_write");
    return -1;
  }

#ifndef NO_MULTICAST
  if (use_mcast) {
    if(IN_MULTICAST(ntohl(addr.sin_addr.s_addr))) {
      if (setsockopt(mcs_write, IPPROTO_IP, IP_MULTICAST_TTL, (const char *) &ttl, sizeof(ttl)) < 0) {
	perror("mcs_write: setsockopt ttl");
	return -1;
      }
      mreq.imr_multiaddr = addr.sin_addr;
      mreq.imr_interface.s_addr = htonl(INADDR_ANY);
      if (setsockopt(mcs_write, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char*)&mreq, sizeof(mreq)) < 0) {
	perror("mcs_write: setsockopt ip_add_membership");
	return -1;
      }
    } 
  } 
#endif
  
  if (connect(mcs_write, (struct sockaddr *) &addr, sizeof(addr))==-1) {
    perror("connect mcs_write");
    return -1;
  };
  
  mcs_write_tcl=Tcl_GetFile((ClientData) mcs_write, TCL_UNIX_FD);
  write_state=MCS_OPEN;
  return 1;
};

int MCSocket::close() {

  #ifdef DEBUG4
  cerr << "MCSocket::close() ...\n"; 
  #endif

  deactivate_read_hdlr();
  deactivate_write_hdlr();
  deactivate_exception_hdlr();

  FD_ZERO(&read_fds);

  close_write();
  close_read();

  my_state=MCS_CLOSED;

  return 1;
};

int MCSocket::close_read() {
  if (read_state==MCS_OPEN) {
    #ifdef DEBUG4
    cerr << "MCSocket::close_read() ...\n"; 
    #endif
    Tcl_FreeFile(mcs_read_tcl);
    ::close(mcs_read);
    read_state=MCS_CLOSED;
    return 1;
  }
  #ifdef DEBUG4
  cerr << "MCSocket::close_read() socket not open.\n"; 
  #endif
  return 1;
};

int MCSocket::close_write() {
  if (read_state==MCS_OPEN) {
    #ifdef DEBUG4
    cerr << "MCSocket::close_write() ...\n"; 
    #endif
    Tcl_FreeFile(mcs_write_tcl);
    ::close(mcs_write);
    read_state=MCS_CLOSED;
    return 1;
  }
  #ifdef DEBUG4
  cerr << "MCSocket::close_write() socket not open.\n"; 
  #endif
  return 1;
};

int MCSocket::reopen() {
  if (my_state==MCS_OPEN) {
    #ifdef DEBUG4
    cerr << "MCSocket::reopen() ...\n"; 
    #endif
    close();
    return open();
  };
  #ifdef DEBUG4
  cerr << "MCSocket::reopen() socket not open.\n"; 
  #endif
  return -1;
};

int MCSocket::write_str(String x) {
  static int count;
  static u_int16_t length;
  if (is_open()) {
    if (!write_is_open() && !open_write()) return -1;
    length=x.length()+1;
    length=htons(length);
    count = ::send(mcs_write, (char *) &length, sizeof(length), 0);
    const char *buffer = (const char *) x;
    count += ::send(mcs_write, (const char *) buffer, length, 0);
    #ifdef DEBUG4
    cerr << "MCSocket::write_str() wrote " << count << " bytes.\n"; 
    #endif
    return count;
  } else {
    #ifdef DEBUG4
    cerr << "MCSocket::write_str() socket not open.\n"; 
    #endif
    return -1;
  }
};

int MCSocket::read_str(String *x) {
  static struct sockaddr_in sin;
  static int slen,len,length;
  static char *buffer;

  if (is_open()) {
    if (!read_is_open() && !open_read()) return -1;

    select(FD_SETSIZE, &read_fds, NULL, NULL, tv);
    
    if (FD_ISSET(mcs_read, &read_fds)) {
      slen = sizeof(sin);
#if defined(AIX) && OS_VER == 4 && OS__VER == 2
      len=recvfrom(mcs_read, (char*)&length, sizeof(length), 0, (struct sockaddr *) &sin, 
		   (long unsigned int *)&slen);
#else
      len=recvfrom(mcs_read, (char*)&length, sizeof(length), 0, (struct sockaddr *) &sin, 
		   &slen);
#endif
      buffer = (char*) malloc(length+1);
#if defined(AIX) && OS_VER == 4 && OS__VER == 2
      len += recvfrom(mcs_read, buffer, length, 0, (struct sockaddr *) &sin, 
		      (long unsigned int *)&slen);
#else
      len += recvfrom(mcs_read, buffer, length, 0, (struct sockaddr *) &sin, 
		      &slen);
#endif
      *x = buffer;
      free(buffer);
      #ifdef DEBUG4
      cerr << "MCSocket::read_str() read " << len << " bytes.\n"; 
      #endif
      return len;
    } else {
      #ifdef DEBUG4
      cerr << "MCSocket::read_str() nothing to read.\n"; 
      #endif
      return 0;
    };
  } else {
    #ifdef DEBUG4
    cerr << "MCSocket::read_str() socket not open.\n"; 
    #endif
    return -1;
  }
};

int MCSocket::mcs_send(char *buffer, int length) {
  static int len;
  if (is_open()) {
    if (!write_is_open() && !open_write()) return -1;
    len = ::send(mcs_write, (const char *) buffer, length, 0);
    #ifdef DEBUG4
    cerr << "MCSocket::mcs_send() sent " << length << " bytes.\n"; 
    #endif
    return  len;
  } else {
    #ifdef DEBUG4
    cerr << "MCSocket::mcs_send() socket not open.\n"; 
    #endif
    return -1;
  }
};

int MCSocket::mcs_recv(char **buffer, int length, sockaddr_in *psin) {
  static int slen,len;
  if (is_open()) {
    if (!read_is_open() && !open_read()) return -1;
    slen = sizeof(*psin);
#if defined(AIX) && OS_VER == 4 && OS__VER == 2
    len = recvfrom(mcs_read, *buffer, length, 0, (struct sockaddr *) psin, 
		   (long unsigned int *)&slen);
#else
    len = recvfrom(mcs_read, *buffer, length, 0, (struct sockaddr *) psin, 
		   &slen);
#endif
    #ifdef DEBUG4
    cerr << "MCSocket::mcs_recv() received " << len << " bytes.\n"; 
    #endif
    return len;
  } else {
    #ifdef DEBUG4
    cerr << "MCSocket::mcs_recv() socket not open.\n"; 
    #endif
    return -1;
  }
};

int MCSocket::activate_read_hdlr() {
  if (is_open() && !read_hdlr_active()) {
    if (!read_is_open() && !open_read()) return -1;
    Tcl_DeleteFileHandler(mcs_read_tcl);
    read_hdlr_mask|=TK_READABLE;
    Tcl_CreateFileHandler(mcs_read_tcl, read_hdlr_mask, event_handler, (ClientData) this);
    #ifdef DEBUG4
    cerr << "MCSocket::activate_read_hdlr() read handler activated.\n"; 
    #endif
    return 1;
  };
  #ifdef DEBUG4
  cerr << "MCSocket::activate_read_hdlr() socket not open or hdlr already active.\n"; 
  #endif
  return -1;
};

int MCSocket::activate_write_hdlr() {
  if (is_open && !write_hdlr_active()) {
    if (!write_is_open() && !open_write()) return -1;
    Tcl_DeleteFileHandler(mcs_write_tcl);
    write_hdlr_mask|=TK_WRITABLE;
    Tcl_CreateFileHandler(mcs_write_tcl, write_hdlr_mask, event_handler, (ClientData) this);
    #ifdef DEBUG4
    cerr << "MCSocket::activate_write_hdlr() write handler activated.\n"; 
    #endif
    return 1;
  };
  #ifdef DEBUG4
  cerr << "MCSocket::activate_write_hdlr() socket not open or hdlr already active.\n"; 
  #endif
  return -1;
};

int MCSocket::activate_exception_hdlr() {
  if (is_open() && !exception_hdlr_active()) {
    if (!read_is_open() && !open_read()) return -1;
    if (!write_is_open() && !open_write()) return -1;

    Tcl_DeleteFileHandler(mcs_read_tcl);
    Tcl_DeleteFileHandler(mcs_write_tcl);
    read_hdlr_mask|=TK_EXCEPTION;
    write_hdlr_mask|=TK_EXCEPTION;
    Tcl_CreateFileHandler(mcs_read_tcl, read_hdlr_mask, event_handler, (ClientData) this);
    Tcl_CreateFileHandler(mcs_write_tcl, write_hdlr_mask, event_handler, (ClientData) this);
    #ifdef DEBUG4
    cerr << "MCSocket::activate_exeption_hdlr() exeption handler activated.\n"; 
    #endif
    return 1;
  };
  #ifdef DEBUG4
  cerr << "MCSocket::activate_exeption_hdlr() socket not open or hdlr already active.\n"; 
  #endif
  return -1;
};

int MCSocket::deactivate_read_hdlr() {
  if (is_open() && read_hdlr_active()) {
    if (!read_is_open()) return 1;
    Tcl_DeleteFileHandler(mcs_read_tcl);
    read_hdlr_mask&=(~TK_READABLE);
    if (read_hdlr_mask!=0) {
      Tcl_CreateFileHandler(mcs_read_tcl, read_hdlr_mask, event_handler, (ClientData) this);
    } else {
      close_read();
    };
    #ifdef DEBUG4
    cerr << "MCSocket::deactivate_read() read handler deactivated.\n"; 
    #endif
    return 1;
  }
  #ifdef DEBUG4
  cerr << "MCSocket::deactivate_read_hdlr() socket not open or hdlr not active.\n"; 
  #endif
  return -1;
};

int MCSocket::deactivate_write_hdlr() {
  if (is_open() && write_hdlr_active()) {
    if (!write_is_open()) return 1;
    Tcl_DeleteFileHandler(mcs_write_tcl);
    write_hdlr_mask&=(~TK_WRITABLE);
    if (write_hdlr_mask!=0) {
      Tcl_CreateFileHandler(mcs_write_tcl, write_hdlr_mask, event_handler, (ClientData) this);
    }
    #ifdef DEBUG4
    cerr << "MCSocket::deactivate_read() read handler deactivated.\n"; 
    #endif
    return 1;
  };
  #ifdef DEBUG4
  cerr << "MCSocket::deactivate_write_hdlr() socket not open or hdlr not active.\n"; 
  #endif
  return -1;
};

int MCSocket::deactivate_exception_hdlr() {
  if (is_open() && exception_hdlr_active()) {
    if (read_is_open()) {
      Tcl_DeleteFileHandler(mcs_read_tcl);
      read_hdlr_mask&=(~TK_EXCEPTION);
      if (read_hdlr_mask!=0) {
	Tcl_CreateFileHandler(mcs_read_tcl, read_hdlr_mask, event_handler, (ClientData) this);
      }
    };
    if (write_is_open()) {
      Tcl_DeleteFileHandler(mcs_write_tcl);
      write_hdlr_mask&=(~TK_EXCEPTION);
      if (write_hdlr_mask!=0) {
	Tcl_CreateFileHandler(mcs_write_tcl, write_hdlr_mask, event_handler, (ClientData) this);
      }
    };
    #ifdef DEBUG4
    cerr << "MCSocket::deactivate_exeption_hdlr() exeption handler deactivated.\n"; 
    #endif
    return 1;
  }
  #ifdef DEBUG4
  cerr << "MCSocket::deactivate_exeption_hdlr() socket not open or hdlr not active.\n"; 
  #endif
  return -1;
};

void MCSocket::event_handler(ClientData cd, int mask) {

  MCSocket *mcsp = (MCSocket *) cd;

  #ifdef DEBUG4
  cerr << "MCSocket::event_handler().\n"; 
  #endif

  if (mask&TK_READABLE)  mcsp->handle(MCS_READ);
  if (mask&TK_WRITABLE)  mcsp->handle(MCS_WRITE);
  if (mask&TK_EXCEPTION) mcsp->handle(MCS_EXCEPTION);

}

void MCSocket::handle(eventtype event) {
  /* this function is supposed to be overloaded */
  switch (event) {
    case MCS_READ: 
      cerr << "MCSocket::handle(READ)\n";
      break;
    case MCS_WRITE: 
      cerr << "MCSocket::handle(WRITE)\n";
      break;
    case MCS_EXCEPTION: 
      cerr << "MCSocket::handle(EXCEPTION)\n";
      break;
    default:
      break;
  };
};

