/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

Copyright (c) 1995 International Computer Science Institute
All Rights Reserved.

Created by Wieland Holfelder (whd@icsi.berkeley.edu)
Tenet Group
International Computer Science Institute

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research and non-profit purposes,
without fee, and without a written agreement is hereby granted,
provided that the above copyright notice, this paragraph and the
following five paragraphs appear in all copies.

Permission to incorporate this software into commercial products may
be obtained from the International Computer Science Institute,
1947 Center Street, Suite 600, Berkeley, CA 94704.

IN NO EVENT SHALL THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE BE
LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE INTERNATIONAL
COMPUTER SCIENCE INSTITUTE  HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.

THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE SPECIFICALLY DISCLAIMS
ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE INTERNATIONAL
COMPUTER SCIENCE INSTITUTE  HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE,
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

The International Computer Science Institute will appreciate
receiving any enhancements and modifications of this software
made by others, and reserves the right to include them in future
releases. The Institute also recommends that this software not
be redistributed by third parties, but be obtained by all directly
from the ICSI server, so that all users of the software may
receive bug notifications and other relevant announcements.

The German Academic Exchange Service has supported the development
of this software. Their support is gratefully acknowledged.

* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
static const char rcsid[]="$Id: RTPMedia.cc,v 1.5 1996/08/12 11:36:07 whd Exp $";

extern "C" {
#include <sys/time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <netinet/in.h>
}


#include "RTPMedia.h"
#include "VCRDisplay.h"
#include "VCRTcl.h"
#include "random.h"
#include "it_depends.h"

RTPMedia::RTPMedia(String	_filename,
		   int		_read_only,
		   String	_ipaddr,
		   int		_port,
		   u_int16_t	_confid,
		   u_int8_t	_ttl,
		   int		_number):

  MBoneMedia(_filename, _read_only, _ipaddr, _port, _confid, _ttl, 1, 4096, 0, 0, _number)

{
  #ifdef DEBUG1
  cout << "RTPMedia::RTPMedia()\n";
  #endif
  media_fmt=RTP;
  media_name="RTPMedia";
};

RTPMedia::~RTPMedia()
{
  #ifdef DEBUG1
  cout << "RTPMedia::~RTPMedia()\n";
  #endif
  send_msg(SEND_DONE);
};

media_rcs RTPMedia::read_next_sync_unit(u_int32_t *ts, float new_speed, int restart)
{

  static int pin=0;

  VCRPacket	p,p2;
  int		first, new_data=0, new_session=0;
  vcr_file_rc	frc;
  struct rtphdr	*rtp;
  struct rtcphdr * rtcp;
  vat_hdr_t	*vat;
  CtrlMsgHdr	*ctrl;
  media_rcs	rc=MEDIA_OK;

  #ifdef DEBUG1
  cout << "RTPMedia::read_next_sync_unit()\n";
  #endif

  /* init p.data */
  p.data=NULL;

  /* init p2.data */
  p2.data=NULL;

  /* delete old/unsent packets */
  while (data_queue.length()>0) {
    if (data_queue.front().data) free(data_queue.front().data);
    data_queue.del_front();
  };
  while (session_queue.length()>0) {
    if (session_queue.front().data) free(session_queue.front().data);
    session_queue.del_front();
  };

  if (new_speed!=0) media_speed=new_speed;

  if (restart) {
    new_session=new_data=1;
  } else new_session=new_data=0;
  #ifdef DEBUG2
    cout << "RTPMedia::read_next_sync_unit() restart="<<restart<<" new_data="<<new_data<<endl;
  #endif
  
  if (ts) *ts=0;
  first=1;

  do {
    frc=media_file.read(&(p.vcr_hdr), &(p.data));
    p2.data=NULL;
    
    if ( frc > VCR_FILE_OK) {
      if (frc == VCR_FILE_EOF) {
#ifdef DEBUG2
	cerr << "RTPMedia::read_next_sync_unit(): MEDIA_EOF.\n";
#endif
	if (!pin) {
	  rc=MEDIA_EOF;
	  pin=1; 
	} else {
	  pin=0;
	  return MEDIA_EOF;
	};
      } else if (frc == VCR_FILE_EMPTY) {
#ifdef DEBUG2
	cerr << "RTPMedia::read_next_sync_unit(): MEDIA_EMPTY.\n";
#endif
	return MEDIA_EMPTY;
      } else {
#ifdef DEBUG1
	cerr << "RTPMedia::read_next_sync_unit() media_file.read() failed ("<<frc<<").\n";
#endif
	return media_return(frc);
      };
    };
#ifdef DEBUG2
    if (frc==VCR_FILE_LAST_PACKET) {
      cerr << "RTPMedia::read_next_sync_unit(): LAST_PACKET.\n";
    };
#endif
    
    /* remember the timestamp of this packet just in */
    /* case we shouldn't get a real data packet ... */
    if (first && ts) *ts=p.vcr_hdr.ts;

    if (p.data) {
      rtp = (struct rtphdr *) p.data;
      rtp_version=(ntohs(rtp->rh_flags)&RTP_T);
    }
    
    switch (p.vcr_hdr.ch) {
      
    case SESSION_CHANNEL:
      if (p.data==NULL) break;
    
      if (media_speed==2) {
	/* speed up rtcp data */
	p.vcr_hdr.ts/=2;
      } else if (media_speed == 0.5) {
	/* slow down rtcp data */
	p.vcr_hdr.ts*=2;
      } else {
	/* play as is (speed == 1) */
      };

      if (rtp_version==RTPv0) {
	ctrl = (CtrlMsgHdr *) p.data;
	ctrl->confid=htons(media_confid);
	session_queue.enq(p);
      } else {
	/*
	   session_queue.enq(p);
	*/
      }
      
#ifdef DEBUG2
      rtcp = (struct rtcphdr *) p.data;
      cout << "RTPMedia::read_next_sync_unit(): SESSION_CHANNEL:\n" <<
              " " << p.vcr_hdr.ts << " " << ntohl(rtcp->rh_ssrc) << " " <<
	      (ntohs(rtcp->rh_flags)&RTP_PT) <<"\n";
#endif
      break;

    case DATA_CHANNEL:

      if (p.data==NULL) break;

      /* we better take this timestamp for scheduling since this 
	 is the critical data */
      if (first && ts) *ts=p.vcr_hdr.ts;
      
      switch (rtp_version) {


      case RTPv0:
#ifdef DEBUG2
        cout << "RTPMedia: sending RTPv0" << endl;
#endif
	/* get pointer to vat header */
	vat = (vat_hdr_t *) p.data;
	
	/* set correct confid for this packet */
	vat->confid=htons(media_confid);
	
	switch (ntohs(vat->flags)&VATHF_FMTMASK) {
	  
	  case AUDF_MULAW8:
	    if (media_speed==2) {
	      /* speed up audio data */
	      double_pcm(&p, new_data);
	    } else if (media_speed == 0.5) {
	      /* slow down */
	      half_pcm(&p, &p2, new_data);
	    } else {
	      /* play as is (speed == 1) */
	      
	      /* set new talkspurt bit of the first packet at restarts */
	      if (new_data) vat->flags=htons(ntohs(vat->flags)|VATHF_NEWTS);
	      
	    };
	    /* enqueue packet */
	    data_queue.enq(p);
	    if (p2.data) data_queue.enq(p2);
	    
	    break;
    
	  case AUDF_CELP:
  	  case AUDF_G721:
          case AUDF_GSM:
          case AUDF_G723:
          case AUDF_LPC4:
          case AUDF_LPC1:
          case AUDF_IDVI:
	    break;
          case AUDF_UNDEF:
          default:
	    char x[10];
	    sprintf(x,"%d",int(ntohs(vat->flags)&VATHF_FMTMASK));
	    error_msg="RTPMedia::read_next_sync_unit() unknown audio format (";
	    error_msg+=x;
	    error_msg+=").\n";
	    cerr << "RTPMedia::read_next_sync_unit() unknown audio format (" <<
	    int(ntohs(vat->flags)&VATHF_FMTMASK) << ").\n";
	    return MEDIA_ERROR;
	    break;
	};
	new_data=0;
      break;

      case RTPv1:
#ifdef DEBUG1
	cout << "RTPMedia::read_next_sync_unit(): sending RTPv1\n";
#endif
	{
	struct rtpv1hdr *rtpv1 = (struct rtpv1hdr *) p.data;
	p.vcr_hdr.seq=ntohs(rtpv1->rh_seq);
	
	if (frc==VCR_FILE_LAST_PACKET) {
#ifdef DEBUG2
	  cerr << "RTPMedia::rtpv1_data_enq(): LAST_PACKET.\n";
#endif
	  rtpv1->rh_sync=htons(1);
	};
	
	if (media_speed==2) {
	  /* speed up rtp data */
	  p.vcr_hdr.ts/=2;
	  rtpv1->rh_ts=htonl(ntohl(rtpv1->rh_ts)/2);
	} else if (media_speed == 0.5) {
	  /* slow down */
	  p.vcr_hdr.ts*=2;
	  rtpv1->rh_ts=htonl(ntohl(rtpv1->rh_ts)*2);
	} else {
	  /* play as is (speed == 1) */
	};
	
	data_queue.enq(p);
        };
        break;

      case RTPv2:
#ifdef DEBUG1
	cout << "RTPMedia::read_next_sync_unit(): sending RTPv2\n";
#endif
	{
        struct rtphdr *rtp = (struct rtphdr *) p.data;


	p.vcr_hdr.seq=ntohs(rtp->rh_seqno);
	
	if (new_data)                  set_rtp_first_marker(rtp);
	if (frc==VCR_FILE_LAST_PACKET) set_rtp_last_marker(rtp);
	
	if (media_speed==2) {
	  /* speed up rtp data */
	  p.vcr_hdr.ts/=2;
	  rtp->rh_ts=htonl(ntohl(rtp->rh_ts)/2);
	} else if (media_speed == 0.5) {
	  /* slow down */
	  p.vcr_hdr.ts*=2;
	  rtp->rh_ts = htonl(ntohl(rtp->rh_ts)*2);
	} else {
	  /* play as is (speed == 1) */
	};
	
	data_queue.enq(p);
        };
	break;

      default:
	cerr << "RTPMedia::read_next_sync_unit(): Unknown RTP version "
	     << int(ntohs(rtp->rh_flags)&RTP_T) << ".\n";
	break;
      };
      break;

    case TIME_CHANNEL:

#ifdef DEBUG1
    cerr << "RTP_MEDIA::read_next_sync_unit(): TIME_CHANNEL" << endl;
#endif

      /* time data, just wait and see ... */
      if (media_speed==2) {
	/* speed up */
	p.vcr_hdr.ts/=2;
      } else if (media_speed == 0.5) {
	/* slow down */
	p.vcr_hdr.ts*=2;
      } else {
	/* wait as is (speed == 1) */
      };
      /* enqueue packet */
      data_queue.enq(p);
      /* return and do not enque anymore... */
      return MEDIA_OK;
      break;
      
    default:
      cout << "RTPMedia::read_next_sync_unit(): invalid channel (" << int(p.vcr_hdr.ch) << ")\n";
      break;
    };
    
    new_data=0;
    first=1;
    
  } while ((!media_file.eof()) && (p.vcr_hdr.ch != DATA_CHANNEL));

  return rc;
};
 

media_rcs RTPMedia::send_pending_sync_unit() {
  struct rtphdr *rtp;
  struct rtcphdr *rtcp;
  VCRDisplay& display=VCRDisplay::instance();
  
#ifdef DEBUG2
  cout << "RTPMedia::send_pending_sync_unit()\n";
#endif
  
  if ((data_queue.length()==0) && (session_queue.length()==0)) 
    return MEDIA_NO_PACKET_PENDING;
  
  while (data_queue.length()>0) {
    if (get_mute()!=1) {
      if (data_queue.front().data) {

	rtp = (struct rtphdr *) data_queue.front().data;
	rtp_version = (ntohs(rtp->rh_flags)&RTP_T);

	switch (rtp_version) {
	case RTPv0:
#ifdef DEBUG2
	  cout << "RTPMedia: pre_sending RTPv0" << endl;
#endif
	  /* actually there is nothing to do for rtpv0 */
	break;

	case RTPv1:
#ifdef DEBUG2
	  cout << "RTPMedia: pre_sending RTPv1" << endl;
#endif
	  rtpv1_data_pre_send((struct rtpv1hdr*) rtp);
 	break;

	case RTPv2:
	  rtp_data_pre_send(rtp);
#ifdef DEBUG2
   	  cout << "RTPMedia: pre_sending RTPv2" << endl;
          cout << "RTPMedia::send_pending_sync_unit("<<ntohs(rtp->rh_seqno)<<")" <<
	  "\n                             display_time: " << display.get_time() <<
          "\n                               rtp->rh_ts: " << ntohl(rtp->rh_ts) <<
          "\n                            rtp->rh_seqno: " << ntohs(rtp->rh_seqno) <<
	  "\n                            rtp->rh_flags: " << ntohs(rtp->rh_flags) <<
	  "\n                              rtp->marker: " << ((ntohs(rtp->rh_flags)&RTP_M)>>7)<<endl<<endl; 
#endif
	break;

	default:
	  cerr << "RTPMedia::send_pending_sync_unit(): Unknown RTP version "
	       << int(ntohs(rtp->rh_flags)&RTP_T) << ".\n";
	  break;
	};

#ifdef DEBUG2
	cout << "send_pending_sync_unit: " << data_queue.front().vcr_hdr.ts << "\n";
#endif

	if (data_queue.front().data) {
	  VCRTcl& tcl=VCRTcl::instance();
	  tcl.set_led(get_media_nb(), 1);
	  
	  media_data_socket.mcs_send(data_queue.front().data, data_queue.front().vcr_hdr.l);
	  free(data_queue.front().data);
	};
      };
    };
    data_queue.del_front();
  };
  
  while (session_queue.length()>0) {

    if (session_queue.front().data) {

      rtcp = (rtcphdr *) session_queue.front().data;
      rtp_version=(ntohs(rtcp->rh_flags)&RTP_T);

      switch (rtp_version) {
        case RTPv0:
	  if (session_buffer_size()!=session_queue.front().vcr_hdr.l) {
	    init_session_buffer(&session_buffer,session_queue.front().vcr_hdr.l);
	  };
	  memcpy(session_buffer, session_queue.front().data, session_buffer_size());
	  
	  if (session_buffer_size()>0) {
	    VCRTcl& tcl=VCRTcl::instance();
	    tcl.set_led(get_media_nb(), 1);
	    media_session_socket.mcs_send(session_buffer, session_buffer_size());
	  }
	  
	  if (session_queue.length()==0) {
	    media_timer.cancel();
	    media_timer.msched(refresh_media_msg);
	  }
	break;
        case RTPv1:
        case RTPv2:
	/*
	   rtp_session_pre_send(rtcp);
	   VCRTcl& tcl=VCRTcl::instance();
	   tcl.set_led(get_media_nb(), 1);
	   media_session_socket.mcs_send(session_queue.front().data,
	                                 session_queue.front().vcr_hdr.l);
	*/				    
	break;
        default:;
      };
      free(session_queue.front().data);
    };
    session_queue.del_front();
  };
  return MEDIA_OK;
};


void RTPMedia::rtpv1_data_pre_send(struct rtpv1hdr *rtp) {
  static u_int16_t last_seq;
  static Pix last_pix=0;
  SrcInfo src;
  u_int16_t ssrc;
  Pix pix;
  
  #ifdef DEBUG1
  cout << "RTPMedia::rtpv1_data_pre_send()\n";
  #endif

  if (ssrc=src.ssrc16=rtpv1_get_ssrc(data_queue.front().data)) {
    src.ssrc32=0;
    src.new_ssrc32=0;
    src.addr.s_addr=0;

    /* Did we send this source already ? */
    if ((pix=src_set.seek(src))==0) {
      src.seq=last_seq;
      src.last_rtp_ts=0;
      src.new_rtp_ts=0;
      src.end_of_sync_sent=0;
      src.send_end_of_sync=0;
      pix=src_set.add(src);
    };
    last_pix=pix;
  } else if (last_pix!=0) {
    /* it's hard to find the right source for an RTPv1 packet without options present... */
    /* so we just assume that it's the same as the last one... (that's at least true if  */
    /* we only have one sender :)                                                        */
    pix=last_pix;
  } else {
    /* just wait for the first option packet */
    free(data_queue.front().data);
    data_queue.front().data=NULL;
    return;
  };

  rtp->rh_seq=htons(src_set(pix).seq);
  last_seq=src_set(pix).seq++;
  
  rtp->rh_chanid=htons(media_confid);
  
  if (src_set(pix).end_of_sync_sent) {
    free(data_queue.front().data);
    data_queue.front().data=NULL;
  } else if (src_set(pix).send_end_of_sync && (data_queue.length()>0)) {
    rtp->rh_sync=htons(1);
    src_set(pix).end_of_sync_sent=1;
#ifdef DEBUG2
    cout << "RTPMedia::rtpv1_data_pre_send(): send END_OF_SYNC\n";
#endif
  };

};


void RTPMedia::rtp_data_pre_send(struct rtphdr *rtp) {
  static u_int16_t last_seq=0;
  SrcInfo src;
  Pix pix;
  VCRDisplay& display=VCRDisplay::instance();
  float clock_rate=1;

  #ifdef DEBUG1
  cout << "RTPMedia::rtp_data_pre_send()\n";
  #endif

  src.addr.s_addr=0;
  src.new_ssrc32=0;
  src.ssrc32=ntohl(rtp->rh_ssrc);

  /* Did we send this source already ? */
  if ((pix=src_set.seek(src))==0) {
  #ifdef DEBUG1
    cout << "RTPMedia::rtp_data_pre_send() source NOT seen yet.\n";
    cout << "RTPMedia::rtp_data_pre_send() last_seq=" << last_seq << "\n";
  #endif
    src.seq=last_seq;
    src.last_rtp_ts=0;
    src.new_rtp_ts=0;
    src.end_of_sync_sent=0;
    src.send_end_of_sync=0;
    pix=src_set.add(src);
  } 
  #ifdef DEBUG2
  else {
    cout << "RTPMedia::rtp_data_pre_send() source already seen (seq:" << src_set(pix).seq <<")\n";
  }
  #endif
  ;


  rtp->rh_seqno=htons(src_set(pix).seq);
  last_seq=src_set(pix).seq++;

  if (src_set(pix).end_of_sync_sent) {
    free(data_queue.front().data);
    data_queue.front().data=NULL;
  } else if (src_set(pix).send_end_of_sync && (data_queue.length()>0)) {
    set_rtp_last_marker(rtp);
    src_set(pix).end_of_sync_sent=1;
    #ifdef DEBUG2
    cout << "RTPMedia::rtp_data_pre_send(): sending END_OF_SYNC\n";
    #endif
  };

};

void RTPMedia::rtp_session_pre_send(struct rtcphdr *rtcp) {
  SrcInfo src;
  Pix pix;

  #ifdef DEBUG1
  cout << "RTPMedia::rtp_session_pre_send()\n";
  #endif

  src.addr.s_addr=0;
  src.new_ssrc32=0;
  src.ssrc32=ntohl(rtcp->rh_ssrc);

  /* Did we send this source already ? */
  if ((pix=src_set.seek(src))==0) {
    src.seq=0;
    src.last_rtp_ts=0;
    src.new_rtp_ts=0;
    src.end_of_sync_sent=0;
    src.send_end_of_sync=0;
    pix=src_set.add(src);
  };
};

void RTPMedia::media_data_handle(struct sockaddr_in *sinp, int length)
{
  vcr_file_rc frc;
  VCR_Hdr vcr_hdr;
  struct rtphdr *rtp;
  Pix pix;

  #ifdef DEBUG2
  cerr << "RTPMedia::media_data_handle().\n";
  #endif

  rtp = (struct rtphdr *) media_buffer;
  rtp_version = ntohs(rtp->rh_flags)&RTP_T;

  vcr_hdr.l=length;
  vcr_hdr.ch=DATA_CHANNEL;

  switch (rtp_version) {
  case RTPv0:
#ifdef DEBUG2
    cout << "RTPMedia: receiving RTPv0 data" << endl;
#endif
    pix=rtpv0_data_handle(&vcr_hdr, sinp);
    break;
  case RTPv1:
#ifdef DEBUG2
    cout << "RTPMedia: receiving RTPv1 data" << endl;
#endif
    pix=rtpv1_data_handle(&vcr_hdr, sinp);
    break;
  case RTPv2:
#ifdef DEBUG2
    cout << "RTPMedia: receiving RTPv2 data" << endl;
#endif
    pix=rtp_data_handle(&vcr_hdr, sinp);
    break;
  default:
    cerr << "RTPMedia::media_data_handle(): Unknown RTP version "
         << int(ntohs(rtp->rh_flags)&RTP_T) << " flags: " << ntohs(rtp->rh_flags)
	 << "/" << ntohs(rtp->rh_flags) << ".\n";
    pix=0;
    return;
  };

  if (pix) {

    if (vcr_hdr.ts>=(src_set(pix).last_update+RTP_UPDATE_TIME)) {
      src_set(pix).last_update=vcr_hdr.ts;
#ifdef DEBUG1
      cerr << "vcr_hdr.ts: " << vcr_hdr.ts << "\n";
#endif
      
      /* this seems to cause problems sometimes, so you may comment it out */
#ifdef DEBUG1
      cerr << "Calling event handler to update time.\n";
#endif 
      call_event_hdlr(MEDIA_UPDATE_TIME, vcr_hdr.ts);
      
      /* update inactive sources */
      Pix i=src_set.first();
      while (i) {
	if ((src_set(i).active==1) && (vcr_hdr.ts>=(src_set(i).last_media_ts+RTP_MAX_IDLE_TIME))) {
	  src_set(i).active=0;
#ifdef DEBUG1
	  cerr << "src_set(" << int(i) << ").active=0\n";
#endif
	};
	src_set.next(i);
      };
    };
    
    if (rtp_version == RTPv0) {
      /* remember last timestamp for rtpv0 session packets */
      last_ts=src_set(pix).last_media_ts=vcr_hdr.ts; 
    };
    
    if (src_set(pix).active==1) {
      if ((frc=media_file.write(&vcr_hdr, media_buffer))!=VCR_FILE_OK) {
	cerr << "RTPMedia::media_data_handle() media_file.write() failed ("<<frc<<").\n";
	call_event_hdlr(MEDIA_ERROR);
	return;
      }
    }
  }
  return;
};


Pix RTPMedia::rtpv0_data_handle(VCR_Hdr *vcr_hdrp, struct sockaddr_in *sinp)
{
  static int old_fmt=-1;
  static int old_length=-1;
  static Pix last_pix=0;
  SrcInfo src;
  vat_hdr_t *vat;
  vcr_file_rc frc;
  Pix pix;
  VCRDisplay& display=VCRDisplay::instance();

  #ifdef DEBUG2
  cerr << "RTPMedia::rtpv0_data_handle().\n";
  #endif

  vat = (vat_hdr_t *) media_buffer;

  /* record only packets with the correct confid */
  if (ntohs(vat->confid)!=media_confid) return 0;
  vcr_hdrp->idx=0;

  src.addr.s_addr=ntohl(sinp->sin_addr.s_addr);
  src.ssrc32=0;
  src.new_ssrc32=0;

  if ((pix=src_set.seek(src))==0) {
    src.new_src=1;
    pix=src_set.add(src);

#ifdef DEBUG1
    cerr << "RTPMedia::rtpv0_data_handle("<<ntohl(vat->ts)<<"): NEW SOURCE, SSRC==" <<
    src_set(pix).new_ssrc16 << "\n";
  } else {
    if (src_set(pix).active==0)
    cerr << "RTPMedia::rtpv0_data_handle(): OLD SOURCE, SSRC==" <<
    src_set(pix).new_ssrc16 << "\n";
#endif
  };

  if (src_set(pix).new_src) {
    src_set(pix).new_src=0;
    src_set(pix).active=0;
    src_set(pix).nb_csrc=0;
    src_set(pix).expected=0;
    src_set(pix).cycles=0;
    src_set(pix).name="";
    vcr_hdrp->idx|=VCR_FILE_NEWSP;
  };

  /* Set new talkspurt index at restarts or if NEWTS is set */
  if ((src_set(pix).active==0) || (ntohs(vat->flags)&VATHF_NEWTS)) {
#ifdef DEBUG1
    if (ntohs(vat->flags) & VATHF_NEWTS)
    cerr << "RTPMedia::rtpv0_data_handle("<<ntohl(vat->ts)<<"): (ntohs(vat->flags) & VATHF_NEWTS)==1\n";
#endif

    vat->flags=htons(ntohs(vat->flags)|VATHF_NEWTS);
    vcr_hdrp->idx|=VCR_FILE_NEWTS;

    /* Times 0.125 to transfer from 8khz sample into ms timestamp */
    src_set(pix).media_offset=u_int32_t(ntohl(vat->ts)*0.125);
    vcr_hdrp->ts=src_set(pix).vcr_offset=src_set(pix).last_update=src_set(pix).last_media_ts=display.get_time();

  } else {

    /* Times 0.125 to transfer from 8khz sample into ms timestamp         */
    /* To get the vcr timestamp, subtract media offset and add vcr_offset */
    vcr_hdrp->ts = u_int32_t(ntohl(vat->ts)*0.125) - src_set(pix).media_offset + src_set(pix).vcr_offset;

  };


#ifdef DEBUG2
  cout << "RTPMedia::rtpv0_read_handle("<<ntohs(vat->confid)<<")"<<
  "\n                             vcr_hdrp->ts: " << vcr_hdrp->ts<<
  "\n                           display_time: " << display.get_time()<<
  "\n                                vat->ts: " << ntohl(vat->ts)<<
  "\n                           media_offset: " << src_set(pix).media_offset<<
  "\n                             vcr_offset: " << src_set(pix).vcr_offset<<"\n\n";
#endif

  /* Set new speaker index if sender differs from previous packets and sender was */
  /* not active				                                          */
  if ((pix!=last_pix) && (src_set(pix).active==0)) {
#ifdef DEBUG1
    cerr << "RTPMedia::rtpv0_data_handle("<<ntohl(vat->ts)<<"): SPEAKER HAS CHANGED.\n";
#endif
    vcr_hdrp->idx|=VCR_FILE_NEWSP;
  };
  last_pix=pix;

  /* Set fmt index if format or length differs from previous packets */
  if (((ntohs(vat->flags) & VATHF_FMTMASK) != old_fmt) || (vcr_hdrp->l != old_length)) {
    vcr_hdrp->idx|=VCR_FILE_NEWFMT;
    old_fmt=(ntohs(vat->flags) & VATHF_FMTMASK);
    old_length=vcr_hdrp->l;
  };

  /* now make this source an active source */
  src_set(pix).active=1;

  VCRTcl& tcl=VCRTcl::instance();
  tcl.set_led(get_media_nb(), 2);

  return pix;
};


Pix RTPMedia::rtpv1_data_handle(VCR_Hdr *vcr_hdrp, struct sockaddr_in *sinp) {
  static Pix last_pix=0;
  struct rtpv1hdr *rtp;
  Pix pix;
  SrcInfo src;
  u_int16_t ssrc;
  VCRDisplay& display=VCRDisplay::instance();

  #ifdef DEBUG1
  cout << "RTPMedia::rtpv1_data_handle()\n";
  #endif

  rtp = (struct rtpv1hdr *) media_buffer;

  /* record only packets with the correct confid              */
  /* however, since this does not work with vic we better don't do it ... */
  // if (ntohs(rtp->rh_chanid)!=media_confid) break;
  /* ... and to be able to play it back with nv just set it to the correct confid ... */
  rtp->rh_chanid=htons(media_confid);

  vcr_hdrp->idx=0;

  src.addr.s_addr=ntohl(sinp->sin_addr.s_addr);
  src.ssrc32=0;
  src.ssrc16=ssrc=rtpv1_get_ssrc(media_buffer);
  src.new_ssrc32=0;

  /* Did we see this source already ? */
  if ((pix=src_set.seek(src))==0) {
    src.new_src=1;
    do {
      src.ssrc16=random16();
    } while ((src.ssrc16==0) || (src_set.contains(src)));
    src.new_ssrc16=src.ssrc16;
    src.ssrc16=ssrc;
    pix=src_set.add(src);
#ifdef DEBUG1
    cerr << "RTPMedia::rtpv1_data_handle(): NEW SOURCE, NEW SSRC==" <<
    src_set(pix).new_ssrc16 << " OLD SSRC==" << src.ssrc16 << " ADDR==" << src.addr.s_addr << "\n";
  } else {
    if (src_set(pix).active==0)
    cerr << "RTPMedia::rtpv1_data_handle(): OLD SOURCE, SSRC==" <<
    src_set(pix).new_ssrc16 << "\n";
#endif
  };

  if (src_set(pix).new_src) {
    src_set(pix).new_src=0;;
    src_set(pix).active=0;;
    src_set(pix).nb_csrc=0;
    src_set(pix).expected=0;
    src_set(pix).cycles=0;
    src_set(pix).name="";
    src_set(pix).seq=ntohs(rtp->rh_seq);
    src_set(pix).start_seq=ntohs(rtp->rh_seq);
    vcr_hdrp->idx|=VCR_FILE_NEWSP;
    vcr_hdrp->idx|=VCR_FILE_NEWFMT;
  }
  
  if (src_set(pix).active==0) {
    #ifdef DEBUG1
    cerr << "RTPMedia::rtpv1_data_handle("<<int(pix)<<"): FIRST DATA\n";
    #endif
    src_set(pix).media_offset=ntohl(rtp->rh_ts);
    vcr_hdrp->ts=src_set(pix).vcr_offset=src_set(pix).last_update=src_set(pix).last_media_ts=display.get_time();
    /* set index for new talkspurt */
    vcr_hdrp->idx|=VCR_FILE_NEWTS;
  } else {
    /* To get the vcr timestamp, subtract media offset and add vcr_offset */
    vcr_hdrp->ts = ntohl(rtp->rh_ts)-src_set(pix).media_offset+src_set(pix).vcr_offset;
  };

  if ((ntohl(rtp->rh_ts)<src_set(pix).last_media_ts) &&
      (ntohs(rtp->rh_seq)>src_set(pix).seq)) {
    #ifdef DEBUG1
    cerr << "RTPMedia::rtpv1_data_handle(): RTP Timestamp wraped arround.\n";
    #endif
    src_set(pix).media_offset=ntohl(rtp->rh_ts);
    vcr_hdrp->ts=
	src_set(pix).vcr_offset=
	src_set(pix).last_update=
	src_set(pix).last_media_ts=
	display.get_time();
  };

  if ((ntohs(rtp->rh_seq)!=src_set(pix).seq+1) && (src_set(pix).active==1)) {
    if (ntohs(rtp->rh_seq)>src_set(pix).seq+1) {
      #ifdef DEBUG1
      cerr << "RTPMedia::rtpv1_data_handle("<<ntohs(rtp->rh_seq)<<"): "<<
      ntohs(rtp->rh_seq)-(src_set(pix).seq+1)<<" packet(s) missing ("<<int(pix)<<").\n";
      #endif
    } else if (ntohs(rtp->rh_seq)==src_set(pix).seq) {
      #ifdef DEBUG1
      cerr << "RTPMedia::rtpv1_data_handle("<<ntohs(rtp->rh_seq)<<"): "<<
      "duplicate packet received ("<<int(pix)<<").\n";
      #endif
      return pix;
    } else if (ntohs(rtp->rh_seq)<src_set(pix).seq) {
      #ifdef DEBUG1
      cerr << "RTPMedia::rtpv1_data_handle(): RTP Sequence number wraped arround.\n";
      #endif
      src_set(pix).seq=src_set(pix).start_seq=ntohs(rtp->rh_seq);
    }
  };

  /* Set new talk spurt index if sender differs from previous packets and */
  /* sender was not active						  */
  if ((src_set(pix).active==0) && (pix!=last_pix)) {
    vcr_hdrp->idx|=VCR_FILE_NEWTS;
    #ifdef DEBUG1
    cerr << "RTPMedia::rtpv1_data_handle(): RTP NEW_SYNC_UNIT index set.\n";
    #endif
  };
  last_pix=pix;

  /* add resp. change ssrc */
  vcr_hdrp->l=rtpv1_add_ssrc(src_set(pix).new_ssrc16, vcr_hdrp->l);

  vcr_hdrp->seq=ntohs(rtp->rh_seq);
  src_set(pix).seq=ntohs(rtp->rh_seq);

  /* remember last timestamp for session packets */
  last_ts=src_set(pix).last_media_ts=vcr_hdrp->ts;

  /* now make this source an active source */
  src_set(pix).active=1;

  #ifdef DEBUG2
  cout << "RTPMedia::rtpv1_data_handle("<<ntohs(rtp->rh_seq)<<")"<<
          "\n                             vcr_hdrp->ts: " << vcr_hdrp->ts<<
	  "\n                             display_time: " << display.get_time()<<
          "\n                               rtp->rh_ts: " << ntohl(rtp->rh_ts)<<
          "\n                             media_offset: " << src_set(pix).media_offset<<
	  "\n                               vcr_offset: " << src_set(pix).vcr_offset<<"\n\n";
  #endif

  VCRTcl& tcl=VCRTcl::instance();
  tcl.set_led(get_media_nb(), 2);

  return pix;
};


Pix RTPMedia::rtp_data_handle(VCR_Hdr *vcr_hdrp, struct sockaddr_in *sinp) {
  static Pix last_pix=0;
  Pix pix;
  float clock_rate=1;
  struct rtphdr *rtp;
  SrcInfo src;
  VCRDisplay& display=VCRDisplay::instance();

  #ifdef DEBUG1
  cout << "RTPMedia::rtp_data_handle()\n";
  #endif

  rtp=(struct rtphdr *) media_buffer;

  clock_rate=get_rtp_clock(rtp);

  vcr_hdrp->idx=0;
  src.addr.s_addr=ntohl(sinp->sin_addr.s_addr);
  src.ssrc32=ntohl(rtp->rh_ssrc);
  src.new_ssrc32=0;

  /* Did we see this source already ? */
  if ((pix=src_set.seek(src))==0) {
    src.new_src=1;
    do {
      src.new_ssrc32=random32();
    } while ((src.new_ssrc32==0) || (src_set.contains(src)));
    pix=src_set.add(src);
#ifdef DEBUG1
    cerr << "RTPMedia::rtp_data_handle(): NEW SOURCE, SSRC==" <<
    src_set(pix).new_ssrc32 << "\n";
  } else {
    if (src_set(pix).active==0)
    cerr << "RTPMedia::rtp_data_handle(): OLD SOURCE, SSRC==" <<
    src_set(pix).new_ssrc32 << "\n";
#endif
  };

  if (src_set(pix).new_src==1) {
    src_set(pix).new_src=0;;
    src_set(pix).active=0;;
    src_set(pix).nb_csrc=0;
    src_set(pix).expected=0;
    src_set(pix).cycles=0;
    src_set(pix).name="";
    src_set(pix).last_pt=(ntohs(rtp->rh_flags)&RTP_PT);
    src_set(pix).seq=src_set(pix).start_seq=ntohs(rtp->rh_seqno);
    vcr_hdrp->idx|=VCR_FILE_NEWSP;
    vcr_hdrp->idx|=VCR_FILE_NEWFMT;
  };

  if (src_set(pix).active==0) {
    #ifdef DEBUG1
    cerr << "RTPMedia::rtp_data_handle("<<int(pix)<<"): FIRST DATA\n";
    #endif
    src_set(pix).media_offset=ntohl(rtp->rh_ts);
    vcr_hdrp->ts=src_set(pix).vcr_offset=src_set(pix).last_update=src_set(pix).last_media_ts=display.get_time();
    /* set index for new talkspurt */
    vcr_hdrp->idx|=VCR_FILE_NEWTS;
  } else {
    /* To get the vcr timestamp, subtract media offset from the rtp timestamp,  */
    /* correct the time by clock_rate and finally add vcr_offset.               */
    vcr_hdrp->ts = u_int32_t((ntohl(rtp->rh_ts)-src_set(pix).media_offset)*clock_rate
			     +src_set(pix).vcr_offset);
  };

  if ((ntohl(rtp->rh_ts)<src_set(pix).last_media_ts) &&
      (ntohs(rtp->rh_seqno)>src_set(pix).seq)) {
    #ifdef DEBUG1
    cerr << "RTPMedia::rtp_data_handle(): RTP Timestamp wraped arround.\n";
    #endif
    src_set(pix).media_offset=ntohl(rtp->rh_ts);
    vcr_hdrp->ts=
	src_set(pix).vcr_offset=
	src_set(pix).last_update=
	src_set(pix).last_media_ts=
	display.get_time();
  };

  if ((ntohs(rtp->rh_seqno)!=src_set(pix).seq+1) && (src_set(pix).active==1)) {
    if (ntohs(rtp->rh_seqno)>src_set(pix).seq+1) {
      #ifdef DEBUG1
      cerr << "RTPMedia::rtp_data_handle("<<ntohl(rtp->rh_ts)<<"): "<<
      ntohs(rtp->rh_seqno)-(src_set(pix).seq+1)<<" packet(s) missing ("<<int(pix)<<").\n";
      #endif
    } else if (ntohs(rtp->rh_seqno)==src_set(pix).seq) {
      #ifdef DEBUG1
      cerr << "RTPMedia::rtp_data_handle("<<ntohs(rtp->rh_seqno)<<"): "<<
      "duplicate packet received ("<<int(pix)<<").\n";
      #endif
      return pix;
    } else if (ntohs(rtp->rh_seqno)<src_set(pix).seq) {
      #ifdef DEBUG1
      cerr << "RTPMedia::rtp_data_handle(): RTP Sequence number wraped arround.\n";
      #endif
      src_set(pix).seq=src_set(pix).start_seq=ntohs(rtp->rh_seqno);
    }
  };

  /* Set new talk spurt index if sender differs from previous packets and */
  /* sender was not active.						  */
  if ((src_set(pix).active==0) && (pix!=last_pix)) {
    vcr_hdrp->idx|=VCR_FILE_NEWTS;
    #ifdef DEBUG1
    cerr << "RTPMedia::rtp_data_handle(): RTP NEW_TALK_SPURT index set.\n";
    #endif
  };
  last_pix=pix;

  /* Set new format index if payload type differs from last packet seen from this source */
  if ((ntohs(rtp->rh_flags)&RTP_PT)!=src_set(pix).last_pt) {
    vcr_hdrp->idx|=VCR_FILE_NEWFMT;
  };
  src_set(pix).last_pt=(ntohs(rtp->rh_flags)&RTP_PT);

  set_exp_nb_of_packets(&src_set(pix), ntohs(rtp->rh_seqno));

  /* add resp. change ssrc */
  rtp->rh_ssrc=htonl(src_set(pix).new_ssrc32);

  vcr_hdrp->seq=ntohs(rtp->rh_seqno);
  src_set(pix).seq=ntohs(rtp->rh_seqno);

  /* remember last timestamp for session packets */
  last_ts=src_set(pix).last_media_ts=vcr_hdrp->ts;

  /* now make this source an active source */
  src_set(pix).active=1;

  #ifdef DEBUG2
  cout << "RTPMedia::rtp_data_handle("<<ntohs(rtp->rh_seqno)<<")"<<
          "\n                             vcr_hdrp->ts: " << vcr_hdrp->ts<<
	  "\n                             display_time: " << display.get_time()<<
          "\n                               rtp->rh_ts: " << ntohl(rtp->rh_ts)<<
          "\n                            rtp->rh_flags: " << ntohl(rtp->rh_flags)<<
          "\n                             media_offset: " << src_set(pix).media_offset<<
	  "\n                               vcr_offset: " << src_set(pix).vcr_offset<<
	  "\n                              rtp->marker: " << ((ntohs(rtp->rh_flags)&RTP_M)>>7)<<
	  "\n                               clock_rate: " << clock_rate << endl << endl;
  #endif

  VCRTcl& tcl=VCRTcl::instance();
  tcl.set_led(get_media_nb(), 2);

  return pix;
};







void RTPMedia::media_session_handle(struct sockaddr_in *sinp, int length)
{
  struct rtphdr *rtp;

  #ifdef DEBUG2
  cerr << "RTPMedia::media_session_handle().\n";
  #endif

  rtp = (struct rtphdr *) media_buffer;

  switch (ntohs(rtp->rh_flags)&RTP_T) {
  case RTPv0:
#ifdef DEBUG2
    cout << "RTPMedia: receiving RTPv0 session data" << endl;
#endif
    rtpv0_session_handle(sinp, length);
    break;
  case RTPv1:
#ifdef DEBUG2
    cout << "RTPMedia: receiving RTPv1 session data" << endl;
#endif
    rtp_session_handle(sinp, length);
    break;
  case RTPv2:
#ifdef DEBUG2
    cout << "RTPMedia: receiving RTPv2 session data" << endl;
#endif
    rtp_session_handle(sinp, length);
    break;
  default:
    cerr << "RTPMedia::media_session_handle(): Unknown RTP version "
         << int(ntohs(rtp->rh_flags)&RTP_T) << " flags: " << ntohs(rtp->rh_flags)
	 << "/" << ntohs(rtp->rh_flags) << ".\n";
  };
  return;
};


void RTPMedia::rtpv0_session_handle(struct sockaddr_in *sinp, int length)
{
  int		i,j;
  VCR_Hdr	vcr_hdr;
  vcr_file_rc	frc;
  CtrlMsgHdr	*chdr;
  IDMsgHdr	*ihdr;
  SrcInfo	src;
  Pix		pix;

  char *ptr;
  VCRDisplay& display=VCRDisplay::instance();

  #ifdef DEBUG2
  cerr << "RTPMedia::rtpv0_session_handle().\n";
  #endif

  chdr = (CtrlMsgHdr *) media_buffer;

  /* record only packets with the correct confid */
  if (ntohs(chdr->confid)!=media_confid) return;

  /* Don't record our own control messages starting with MEDIA_MAGIC ... */

  if (((ntohs(chdr->flags)&CTRHF_MSGTYPE) == CTRL_TYPE_ID) &&
      (strncmp(media_buffer+sizeof(CtrlMsgHdr), MEDIA_MAGIC, strlen(MEDIA_MAGIC))==0))
  return;

  vcr_hdr.ch=SESSION_CHANNEL;
  vcr_hdr.idx=0;
  vcr_hdr.l=u_int16_t(length);
  vcr_hdr.ts=display.get_time();

  src.addr.s_addr=ntohl(sinp->sin_addr.s_addr);
  src.ssrc16=0;
  src.new_ssrc16=0;

  if ((pix=src_set.seek(src))==0) {
    src.new_src=1;
    src.active=0;
    pix=src_set.add(src);
    src_set(pix).last_update=vcr_hdr.ts;
  } else {
    if (vcr_hdr.ts>=(src_set(pix).last_update+RTP_UPDATE_TIME)) {
      src_set(pix).last_update=vcr_hdr.ts;
      /* update inactive sources, so we won't store their session packets */
      Pix i=src_set.first();
      while (i) {
	if (src_set(i).active &&
	    ((vcr_hdr.ts-src_set(i).last_media_ts)>=RTP_MAX_IDLE_TIME)) {
	  src_set(i).active=0;
	};
	src_set.next(i);
      };
    };
  };
  /* remember last timestamp for session packets */
  last_ts=vcr_hdr.ts;

  if (src_set(pix).active) {

    if ((src_set(pix).name.length()==0) && ((ntohs(chdr->flags)&CTRHF_MSGTYPE) == CTRL_TYPE_ID)) {
      src_set(pix).name=(char *) media_buffer+sizeof(CtrlMsgHdr);
    };

    if ((ntohs(chdr->flags)&CTRHF_MSGTYPE) == CTRL_TYPE_IDLIST) {
      ihdr = (IDMsgHdr *) media_buffer;
      ptr=media_buffer+sizeof(IDMsgHdr);
      src_set(pix).nb_csrc=(ntohs(ihdr->flags)&IDMHF_NBOS);
      /* delete old entries */
      for (i=0; i<16; i++) src_set(pix).csrc16[i]=0;
      /* get new entries */
      for (i=0; i<(ntohs(ihdr->flags)&IDMHF_NBOS); i++) {
	src_set(pix).csrc16[i]=(u_int16_t) ptr;
	if (i!=(ntohs(ihdr->flags)&IDMHF_NBOS)) {
	  /* skip this site id */
	  ptr+=4;
	  /* get strlen of site's name */
	  j=strlen(ptr);
	  /* padded to next for byte boundary */
	  while (j%4) j++;
	  /* skip to next site id */
	  ptr+=j;
	};
      };
    };

    if ((ntohs(chdr->flags)&CTRHF_MSGTYPE) == CTRL_TYPE_DONE) {
      src_set.del(src_set(pix));
    };

    if ((frc=media_file.write(&vcr_hdr, media_buffer))!=VCR_FILE_OK) {
      cout << "RTPMedia::rtpv0_session_handle() media_file.write() failed ("<<frc<<").\n";
      call_event_hdlr(MEDIA_ERROR);
      return;
    };
  };
  VCRTcl& tcl=VCRTcl::instance();
  tcl.set_led(get_media_nb(), 2);
  return;
};



void RTPMedia::rtp_session_handle(struct sockaddr_in *sinp, int length)
{
  Pix pix;
  SrcInfo src;
  struct rtcphdr *rtcp;
  VCR_Hdr vcr_hdr;
  vcr_file_rc frc;
  VCRDisplay& display=VCRDisplay::instance();

  #ifdef DEBUG2
  cerr << "RTPMedia::media_session_handle().\n";
  #endif

  rtcp = (struct rtcphdr *) media_buffer;

  vcr_hdr.ch=SESSION_CHANNEL;
  vcr_hdr.idx=0;
  vcr_hdr.l=u_int16_t(length);
  vcr_hdr.ts=display.get_time();

  src.addr.s_addr=ntohl(sinp->sin_addr.s_addr);
  src.ssrc32=ntohl(rtcp->rh_ssrc);
  src.new_ssrc32=0;

  /* Did we see this source already ? */
  if ((pix=src_set.seek(src))==0) {
    src.new_src=1;

    do {
      src.new_ssrc32=random32();
    } while ((src.new_ssrc32==0) || (src_set.contains(src)));

    pix=src_set.add(src);
    src_set(pix).last_update=vcr_hdr.ts;
#ifdef DEBUG1
    cerr << "RTPMedia::media_session_handle(): NEW SOURCE, SSRC==" <<
    src_set(pix).new_ssrc32 << "\n";
#endif

  };

  /* add resp. change ssrc */
  rtcp->rh_ssrc=htonl(src_set(pix).new_ssrc32);

  /* remember last timestamp for session packets */
  last_ts=vcr_hdr.ts;

  if (src_set(pix).active==1) {
    if ((frc=media_file.write(&vcr_hdr, media_buffer))!=VCR_FILE_OK) {
      cerr << "RTPMedia::data_read_handle() media_file.write() failed ("<<frc<<").\n";
      call_event_hdlr(MEDIA_ERROR);
      return;
    };
  };

  VCRTcl& tcl=VCRTcl::instance();
  tcl.set_led(get_media_nb(), 2);
  return;
};



void RTPMedia::media_timeout() {
  #ifdef DEBUG3
  cout << "RTPMedia::media_timeout()\n";
  #endif

  if (session_buffer_size()>0) {
    media_session_socket.mcs_send(session_buffer, session_buffer_size());
    VCRTcl& tcl=VCRTcl::instance();
    tcl.set_led(get_media_nb(), 1);
  };
  
  MBoneMedia::media_timeout();
};

void RTPMedia::ctrl_timeout() {
  #ifdef DEBUG2
  cout << "RTPMedia::ctrl_timeout()\n";
  #endif
};

media_rcs RTPMedia::send_msg(msg_codes code, String msg) {
  struct CtrlMsgHdr hdr, *hdrp;
  VCRTcl& tcl=VCRTcl::instance();
  int length;

  switch (code) {
  case SEND_ID:
    switch (rtp_version) {
      case RTPv0:
        length=(msg.length()+sizeof(*hdrp)+1);
	if (id_buffer_size()<length) {
	  init_id_buffer(&id_buffer,length);
	};
	if (msg.length()>0) {
	  hdrp=(struct CtrlMsgHdr *) id_buffer;
	  hdrp->flags=htons(CTRL_TYPE_ID);
	  hdrp->confid=htons(media_confid);
	  
	  sprintf((id_buffer+sizeof(*hdrp)),"%s",(const char *) msg);
	  VCRTcl& tcl=VCRTcl::instance();
	  tcl.set_led(get_media_nb(), 1);
	  media_session_socket.mcs_send(id_buffer,length);
	};
      break;

      case RTPv1:
      case RTPv2:
        break;
      default:
        /* we don't know yet which RTP version we are using so we just ignore it. */
        ;
	break;
      }
    break;

  case SEND_DONE:
    switch (rtp_version) {
      case RTPv0:
        hdr.flags=htons(CTRL_TYPE_DONE);
	hdr.confid=htons(media_confid);
	tcl.set_led(get_media_nb(), 1);
	media_session_socket.mcs_send((char *) &hdr, sizeof(hdr));
      break;

      case RTPv1:
      case RTPv2:
        break;
      default:
        /* we don't know yet which RTP version we are using so we just ignore it. */
        ;
	break;
      }
  default:
    break;
  }

  return MEDIA_OK;
};


media_rcs RTPMedia::send_ctrl() {
  #ifdef DEBUG1
  cout << "RTPMedia::send_ctrl()\n";
  #endif
  return MEDIA_OK;
};

media_rcs RTPMedia::start_recording() {
  #ifdef DEBUG1
  cout << "RTPMedia::start_recording()\n";
  #endif
  src_set.clear();
  return MEDIA_OK;
};

media_rcs RTPMedia::stop_recording() {
  #ifdef DEBUG1
  cout << "RTPMedia::stop_recording()\n";
  #endif
  src_set.clear();
  return MEDIA_OK;
};

media_rcs RTPMedia::start_playing() {
  #ifdef DEBUG2
  cout << "RTPMedia::start_playing()\n";
  #endif
  src_set.clear();
  return MEDIA_OK;
};

media_rcs RTPMedia::stop_playing() {
  Pix pix;
  int ok=1;

  #ifdef DEBUG2
  cout << "RTPMedia::stop_playing()\n";
  #endif

  for (pix=src_set.first(); pix!=0; src_set.next(pix)) {
    if (!src_set(pix).end_of_sync_sent) {
#ifdef DEBUG1
      cout << "RTPMedia::stop_playing() end_of_sync_sent=="<< src_set(pix).end_of_sync_sent <<"\n";
#endif
      src_set(pix).send_end_of_sync=1;
      src_set(pix).end_of_sync_sent=0;
      send_pending_sync_unit();
      ok=0;
    };
  };

  if (ok) src_set.clear();

  return MEDIA_OK;
};


/* from draft-ietf-avt-rtp-05.ps [Page 33f] */
void RTPMedia::set_exp_nb_of_packets(SrcInfo *src, u_int16_t curr_seq) {

  #ifdef DEBUG1
  cout << "RTPMedia::set_exp_nb_of_packets()\n";
  #endif

  /* Determining the Expected Number of RTP Packets */
  src->expected=src->cycles*65536+curr_seq-src->start_seq+1;
  if (curr_seq > src->seq) {
    if (curr_seq - src->seq > 32768) {
      /* out-of-order packet with wrap-around (e.g., 65530 preceded by 3) */
      src->cycles--;
    }
  } else if (curr_seq < src->seq) {
    if (curr_seq-src->seq > 32768) {
      /* out-of-order packet (e.g., 2 preceded by 3) */
    } else {
      /* wrap-around (e.g., 3 preceded by 65530) */
      src->cycles++;
    };
  };
};

/* from draft-ietf-avt-rtp-06.ps [Page 39ff] */
int32_t RTPMedia::rtcp_period(int members, int nsenders, double bw, int we_sent, int packet_size)
{
  double t; /* interval */
  double rtcp_min_time = RTCP_MIN_RPT_TIME;

  #ifdef DEBUG1
  cout << "RTPMedia::rtcp_period()\n";
  #endif

  /* The avg. rtcp size is initialized to 128 bytes which is
   * conservative (it assumes everyone else is generating SRs instead
   * of RRs).
   */
  static double avg_rtcp_size = 128;

  int n;    /* number of members for computation */

  /* initial */
  if (members == 1) {
    rtcp_min_time /= 2;
  }

  /* Compute estimated number of session members. */
  n = members;
  if (nsenders > 0) {
    if (we_sent) {
      bw *= RTCP_SENDER_BW_FRACTION;
      n = nsenders;
    } else {
      bw *= RTCP_RECEIVER_BW_FRACTION;
      n -= nsenders;
    }
  }

  /* Update avg. size of message [Is this really helpful?] */
  avg_rtcp_size += (packet_size - avg_rtcp_size) * RTCP_SIZE_GAIN;

  /* compute interval */
  t = avg_rtcp_size * n / bw;

  /* enforce minimum spacing */
  if (t < rtcp_min_time) t = rtcp_min_time;

  /*
   * To avoid traffic bursts from unintended synchronization with
   * other sites, we then pick our actual next report interval as a
   * random number uniformly distributed between 0.5*t and 1.5*t.
   */
  return int32_t(t * (drand48() + 1.0));
}


u_int16_t RTPMedia::rtpv1_get_ssrc(char *buffer)
{

  #ifdef DEBUG1
  cout << "RTPMedia::rtpv1_get_ssrc()\n";
  #endif

  if (!buffer) return 0;

  int done;
  struct rtpv1hdr *rtp=(struct rtpv1hdr *)buffer;
  struct rtpv1opthdr *opt=(struct rtpv1opthdr *)(rtp+1);
  struct rtpv1ssrchdr *ssrcopt;

  done = !ntohs(rtp->rh_opts);
  while (!done) {
    if (opt->roh_optlen == 0) break;
    switch (opt->roh_type) {
    case RTPv1OPT_SSRC:
      ssrcopt = (struct rtpv1ssrchdr *) opt;
      return ntohs(ssrcopt->rsh_id);
      break;
    }
    done = opt->roh_fin;
    opt = (struct rtpv1opthdr *) (((u_int32_t *)opt)+opt->roh_optlen);
  };
  return 0;
};

u_int16_t RTPMedia::rtpv1_add_ssrc(u_int16_t ssrc, u_int16_t length)
{

  #ifdef DEBUG1
  cout << "RTPMedia::rtpv1_add_ssrc()\n";
  #endif

  int done;
  u_int32_t optlen=0;
  struct rtpv1hdr *rtp=(struct rtpv1hdr *) media_buffer;
  struct rtpv1opthdr *opt=(struct rtpv1opthdr *)(rtp+1);
  struct rtpv1ssrchdr *ssrcopt;
  char *new_data;

  done = !ntohs(rtp->rh_opts);
  while (!done) {
    if (opt->roh_optlen == 0) break;
    if (opt->roh_type==RTPv1OPT_SSRC) {
      ssrcopt = (struct rtpv1ssrchdr *) opt;
      #ifdef DEBUG2
      cerr << "RTPMedia::rtpv1_add_ssrc() changed SSRC: "<< ntohl(ssrcopt->rsh_id) <<
      " to SSRC: " << ssrc << "\n";
      #endif
      ssrcopt->rsh_id=htons(ssrc);
      return length;
    };
    done = opt->roh_fin;
    /* since we add an option we have to reset the roh_fin bit for the last option */
    if (done) opt->roh_fin=0;
    opt = (struct rtpv1opthdr *) (((u_int32_t *)opt)+opt->roh_optlen);
  };

  optlen=((char *)opt)-media_buffer;

  /* set the option present bit for the ssrc option */
  rtp->rh_opts=htons(1);

#ifndef MAX
#define MAX(x,y) (((x)>(y))?(x):(y))
#endif

  /* allocate memory for new data */
  new_data = (char *) malloc(MAX(length+sizeof(*ssrcopt), media_buffer_size()));

  /* copy old hdr */
  memcpy(new_data, media_buffer, optlen);

  /* create new ssrc option */
  ssrcopt=(struct rtpv1ssrchdr *) (new_data+optlen);
  ssrcopt->rsh_fin=1;
  ssrcopt->rsh_type=RTPv1OPT_SSRC;
  ssrcopt->rsh_optlen=1;
  ssrcopt->rsh_id=htons(ssrc);

  /* append old data */
  memcpy(new_data+optlen+sizeof(*ssrcopt), media_buffer+optlen, length);

  change_media_buffer(&media_buffer, &new_data, MAX(length+sizeof(*ssrcopt),
						       media_buffer_size()));

  #ifdef DEBUG2
  cerr << "RTPMedia::rtpv1_add_ssrc() added SSRC: "<< ntohl(ssrcopt->rsh_id) << "\n";
  #endif

  /* return new length */
  return length+sizeof(*ssrcopt);
};


int RTPMedia::double_pcm(VCRPacket *p, int new_data) {
  vat_hdr_t	*vat;
  int		i,offset;
  int		packet_length;


  vat = (vat_hdr_t *) p->data;
  /* set new talkspurt bit at restarts */
  if (new_data) vat->flags=htons(ntohs(vat->flags)|VATHF_NEWTS);

  /* calculate offset = length of vat header + NSID * 4 byte speaker id */
  offset=sizeof(vat_hdr_t)+(ntohs(vat->flags)&VATHF_NSID)*4;

  packet_length=p->vcr_hdr.l;
  /* send out only half of the packet */
  p->vcr_hdr.l-=u_int16_t((packet_length-offset)/2);

  /* set new vat timestamp */
  vat->ts=htonl(ntohl(vat->ts)/2);

  /* "compress" audio packet */
  for (i=0; i<(packet_length-offset)/2; i++) {
    p->data[i+offset] = p->data[i*2+offset+1];
  };

  return 1;
};

int RTPMedia::half_pcm(VCRPacket *p, VCRPacket *p2, int new_data) {
  vat_hdr_t	*vat;
  int		i,offset;
  int		packet_length;


  vat = (vat_hdr_t *) p->data;

#define ONE_PACKET
#ifdef ONE_PACKET

  /* set new vat timestamp */
  vat->ts=htonl(ntohl(vat->ts)*2);

  /* set new talkspurt bit of the first packet at restarts */
  if (new_data) vat->flags=htons(ntohs(vat->flags)|VATHF_NEWTS);

  /* calculate offset = length of vat header + NSID * 4 byte speaker id */
  offset=sizeof(vat_hdr_t)+(ntohs(vat->flags)&VATHF_NSID)*4;

  packet_length=p->vcr_hdr.l;
  /* create audio packet twice as large */
  p->vcr_hdr.l=u_int16_t(packet_length*2-offset);

  char *new_packet;
  new_packet=(char *) malloc(p->vcr_hdr.l);

  /* copy vat header */
  for (i=0; i<offset; new_packet[i]=p->data[i++]);

  /* stretch audio packet */
  for (i=packet_length-offset-1; i>=0; i--)
  new_packet[i*2+offset] = new_packet[i*2+offset+1] = p->data[i+offset];

  free(p->data);
  p->data=new_packet;

#else
  /* this is an implementation to slow down the speed without generating */
  /* one large but two regular sized packets. I guess something is still */
  /* not 100% ok, though ? */

  /* set new vat timestamp */
  vat->ts=htonl(ntohl(vat->ts)*2);

  /* calculate offset = length of vat header + NSID * 4 byte speaker id */
  offset=sizeof(vat_hdr_t)+(ntohs(vat->flags)&VATHF_NSID)*4;

  /* create second audio packet */
  memcpy(&(p2->vcr_hdr), &(p->vcr_hdr), sizeof(p->vcr_hdr));
  p2->data=(char *) malloc(p2->vcr_hdr.l);

  /* copy vat header */
  for (i=0; i<offset; p2->data[i]=p->data[i++]);

  /* get pointer to vat header of second packet*/
  vat = (vat_hdr_t *) p2->data;

  /* set vat timestamp for second packet */
  vat->ts=htonl(ntohl(vat->ts)+(p2->vcr_hdr.l-offset));

  /* calculate playout for second packet */
  p2->vcr_hdr.ts+=((p2->vcr_hdr.l-offset)*0.125)/2;

  /* stretch and distribute audio data */
  int j=p->vcr_hdr.l-offset;
  for (i=p->vcr_hdr.l-offset; i>(p->vcr_hdr.l-offset)/2; i--) {
    p2->data[j+offset] = p2->data[j+offset-1] = p->data[i+offset];
    j-=2;
  };

  j=p->vcr_hdr.l-offset;
  for (i=(p->vcr_hdr.l-offset)/2; i>0; i--) {
    p->data[j+offset] = p->data[j+offset-1] = p->data[i+offset];
    j-=2;
  };

  vat = (vat_hdr_t *) p->data;
  /* set new talkspurt bit of the first packet at restarts */
  if (new_data) vat->flags=htons(ntohs(vat->flags)|VATHF_NEWTS);

#endif

  return 1;

}


float RTPMedia::get_rtp_clock(rtphdr* rtp) {
float clock_rate=1;
 
#ifdef DEBUG1
  cerr << "RTPMedia:get_rtp_clock()\n";
#endif

  switch (ntohs(rtp->rh_flags)&RTP_PT) {
  
    case RTP_PT_PCMU:		
    /* 8kHz PCM mu-law mono */
      clock_rate=0.125; /* 1000/8000 */
      #ifdef DEBUG2
      cout << "RTPMedia:get_rtp_clock() 8kHz PCM mu-law mono\n";
      #endif
      break;

    case RTP_PT_1016:		
    /* 8kHz CELP (Fed Std 1016) mono */
      clock_rate=0.125; /* 1000/8000 */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() 8kHz CELP (Fed Std 1016) mono\n"; 
      #endif
      break;

    case RTP_PT_G721:		
    /* 8kHz G.721 ADPCM mono */
      clock_rate=0.125; /* 1000/8000 */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() 8kHz G.721 ADPCM mono\n"; 
      #endif
      break;

    case RTP_PT_GSM:
    /* 8kHz GSM mono */
      clock_rate=0.125; /* 1000/8000 */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() 8kHz GSM mono\n"; 
      #endif
      break;

    case RTP_PT_G723:
    /* 8kHz G.723 ADPCM mono */
      clock_rate=0.125; /* 1000/8000 */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() 8kHz G.723 ADPCM mono\n"; 
      #endif
      break;

    case RTP_PT_DVI:
    /* 8kHz Intel DVI ADPCM mono */
      clock_rate=0.125; /* 1000/8000 */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() 8 kHz Intel DVI ADPCM mono\n"; 
      #endif
      break;

    case RTP_PT_L16_16:
    /* 16kHz 16-bit linear mono */
      clock_rate=0.0625; /* 1000/16000 */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() 6: 16kHz 16-bit linear mono\n"; 
      #endif
      break;

    case RTP_PT_L16_44_2:
    /* 44.1kHz 16-bit linear stereo */
      clock_rate=0.022675736; /* 1000/44100 */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() 4_2: 44.1kHz 16-bit linear stereo\n"; 
      #endif
      break;

    case RTP_PT_SCR:
    /* simple conditional replenishment */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() simple conditional replenishment\n"; 
      #endif
      break;

    case RTP_PT_MPEG:
    /* MPEG II? */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() MPEG II?\n"; 
      #endif
      break;

    case RTP_PT_CELLB:
    /* Sun CellB */
      clock_rate=0.011111; /* 1000/90000 */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() Sun CellB\n"; 
      #endif
      break;

    case RTP_PT_JPEG:
    /* JPEG */
      clock_rate=0.011111; /* 1000/90000 */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() JPEG\n"; 
      #endif
      break;

    case RTP_PT_CUSEEME:
    /* Cornell CU-SeeMe */
      clock_rate=0.011111; /* 1000/90000 */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() Cornell CU-SeeMe\n"; 
      #endif
      break;

    case RTP_PT_NV:
    /* Xerox PARC nv */
      clock_rate=0.011111; /* 1000/90000 */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() Xerox PARC nv\n"; 
      #endif
      break;

    case RTP_PT_DVC:
    /* BBN dvc */
      clock_rate=0.011111; /* 1000/90000 */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() BBN dvc\n"; 
      #endif
      break;

    case RTP_PT_BOLT:
    /* Bolter */
      clock_rate=0.011111; /* 1000/90000 */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() Bolter\n"; 
      #endif
      break;

    case RTP_PT_H261:
    /* CCITT H.261 */                               
      clock_rate=0.011111; /* 1000/90000 */
      #ifdef DEBUG2
      cout << "RTPMedia::get_rtp_clock() CCITT H.261\n";
      #endif
      break;

    default:  
      cerr << "RTPMedia::get_rtp_clock() unknown payload type " << u_int32_t(ntohs(rtp->rh_flags)&RTP_P)           << endl;
  }

  return clock_rate;

};


void RTPMedia::set_rtp_first_marker(rtphdr* rtp) {
 
#ifdef DEBUG1
  cerr << "RTPMedia:set_rtp_first_marker()\n";
#endif

  switch (ntohs(rtp->rh_flags)&RTP_PT) {
  
    case RTP_PT_PCMU:		
    /* 8kHz PCM mu-law mono */
      #ifdef DEBUG2
      cout << "RTPMedia:set_rtp_first_marker() 8kHz PCM mu-law mono: Marker-Bit set.\n";
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_1016:		
    /* 8kHz CELP (Fed Std 1016) mono */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() 8kHz CELP (Fed Std 1016) mono: Marker-Bit set.\n"; 
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_G721:		
    /* 8kHz G.721 ADPCM mono */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() 8kHz G.721 ADPCM mono: Marker-Bit set.\n"; 
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_GSM:
    /* 8kHz GSM mono */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() 8kHz GSM mono: Marker-Bit set.\n"; 
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_G723:
    /* 8kHz G.723 ADPCM mono */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() 8kHz G.723 ADPCM mono: Marker-Bit set.\n"; 
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_DVI:
    /* 8kHz Intel DVI ADPCM mono */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() 8 kHz Intel DVI ADPCM mono: Marker-Bit set.\n"; 
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_L16_16:
    /* 16kHz 16-bit linear mono */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() 16kHz 16-bit linear mono: Marker-Bit set.\n"; 
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_L16_44_2:
    /* 44.1kHz 16-bit linear stereo */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() 44.1kHz 16-bit linear stereo: Marker-Bit set.\n"; 
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_SCR:
    /* simple conditional replenishment */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() simple conditional replenishment: Marker-Bit set.\n"; 
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_MPEG:
    /* MPEG II? */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() MPEG II: Marker-Bit NOT set.\n"; 
      #endif
      break;

    case RTP_PT_CELLB:
    /* Sun CellB */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() Sun CellB: Marker-Bit NOT set.\n"; 
      #endif
      break;

    case RTP_PT_JPEG:
    /* JPEG */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() JPEG: Marker-Bit NOT set.\n"; 
      #endif
      break;

    case RTP_PT_CUSEEME:
    /* Cornell CU-SeeMe */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() Cornell CU-SeeMe: Marker-Bit NOT set.\n\n"; 
      #endif
      break;

    case RTP_PT_NV:
    /* Xerox PARC nv */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() Xerox PARC nv: Marker-Bit NOT set.\n\n"; 
      #endif
      break;

    case RTP_PT_DVC:
    /* BBN dvc */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() BBN dvc: Marker-Bit NOT set.\n\n"; 
      #endif
      break;

    case RTP_PT_BOLT:
    /* Bolter */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() Bolter: Marker-Bit NOT set.\n\n"; 
      #endif
      break;

    case RTP_PT_H261:
    /* CCITT H.261 */                               
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_first_marker() CCITT H.261: Marker-Bit NOT set.\n\n";
      #endif
      break;

    default:  
      cerr << "RTPMedia::set_rtp_first_marker() unknown payload type " << u_int32_t(ntohs(rtp->rh_flags)&RTP_P)           << endl;
  }

};


void RTPMedia::set_rtp_last_marker(rtphdr* rtp) {
 
#ifdef DEBUG1
  cerr << "RTPMedia:set_rtp_last_marker()\n";
#endif

  switch (ntohs(rtp->rh_flags)&RTP_PT) {
  
    case RTP_PT_PCMU:		
    /* 8kHz PCM mu-law mono */
      #ifdef DEBUG2
      cout << "RTPMedia:set_rtp_last_marker() 8kHz PCM mu-law mono: Marker-Bit NOT set.\n\n";
      #endif
      break;

    case RTP_PT_1016:		
    /* 8kHz CELP (Fed Std 1016) mono */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() 8kHz CELP (Fed Std 1016) mono: Marker-Bit NOT set.\n\n"; 
      #endif
      break;

    case RTP_PT_G721:		
    /* 8kHz G.721 ADPCM mono */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() 8kHz G.721 ADPCM mono: Marker-Bit NOT set.\n\n"; 
      #endif
      break;

    case RTP_PT_GSM:
    /* 8kHz GSM mono */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() 8kHz GSM mono: Marker-Bit NOT set.\n\n"; 
      #endif
      break;

    case RTP_PT_G723:
    /* 8kHz G.723 ADPCM mono */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() 8kHz G.723 ADPCM mono: Marker-Bit NOT set.\n\n"; 
      #endif
      break;

    case RTP_PT_DVI:
    /* 8kHz Intel DVI ADPCM mono */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() 8 kHz Intel DVI ADPCM mono: Marker-Bit NOT set.\n\n"; 
      #endif
      break;

    case RTP_PT_L16_16:
    /* 16kHz 16-bit linear mono */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() 16kHz 16-bit linear mono: Marker-Bit NOT set.\n\n"; 
      #endif
      break;

    case RTP_PT_L16_44_2:
    /* 44.1kHz 16-bit linear stereo */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() 44.1kHz 16-bit linear stereo: Marker-Bit NOT set.\n\n"; 
      #endif
      break;

    case RTP_PT_SCR:
    /* simple conditional replenishment */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() simple conditional replenishment: Marker-Bit NOT set.\n\n"; 
      #endif
      break;

    case RTP_PT_MPEG:
    /* MPEG II? */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() MPEG II: Marker-Bit set.\n"; 
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_CELLB:
    /* Sun CellB */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() Sun CellB: Marker-Bit set.\n"; 
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_JPEG:
    /* JPEG */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() JPEG: Marker-Bit set.\n"; 
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_CUSEEME:
    /* Cornell CU-SeeMe */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() Cornell CU-SeeMe: Marker-Bit set.\n"; 
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_NV:
    /* Xerox PARC nv */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() Xerox PARC nv: Marker-Bit set.\n"; 
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_DVC:
    /* BBN dvc */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() BBN dvc: Marker-Bit set.\n"; 
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_BOLT:
    /* Bolter */
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() Bolter: Marker-Bit set.\n"; 
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    case RTP_PT_H261:
    /* CCITT H.261 */                               
      #ifdef DEBUG2
      cout << "RTPMedia::set_rtp_last_marker() CCITT H.261: Marker-Bit set.\n";
      #endif
      rtp->rh_flags=htons(ntohs(rtp->rh_flags)|RTP_M);
      break;

    default:  
      cerr << "RTPMedia::set_rtp_last_marker() unknown payload type " << u_int32_t(ntohs(rtp->rh_flags)&RTP_P)           << endl;
  }

};

