/*
 * RTPv1.h
 *
 * Constants and structures based on the 12/15/92 draft of the RTP protocol
 * Internet Draft. This information is still subject to change.
 *
 */

/* $Id: RTPv1.h,v 1.1 1996/07/26 15:02:20 whd Exp $ */

#ifndef _RTPv1_h
#define _RTPv1_h

#include "RTP.h"

/* Basic RTPv1 header */
struct rtpv1hdr {
#ifndef LITTLE_BITFIELDS
	u_int8_t	rh_vers:2;	/* version */
	u_int8_t	rh_chanid:6;	/* channel id */
	u_int8_t	rh_opts:1;	/* options present */
	u_int8_t	rh_sync:1;	/* end of synchronization unit */
	u_int8_t	rh_content:6;	/* content id */
#else
	u_int8_t	rh_chanid:6;	/* channel id */
	u_int8_t	rh_vers:2;	/* version */
	u_int8_t	rh_content:6;	/* content id */
	u_int8_t	rh_sync:1;	/* end of synchronization unit */
	u_int8_t	rh_opts:1;	/* options present */
#endif
	u_int16_t	rh_seq;		/* sequence number */
	u_int32_t	rh_ts;		/* time stamp (middle of NTP timestamp) */
};

/* Basic RTPv1 option header */
struct rtpv1opthdr {
#ifndef LITTLE_BITFIELDS
	u_int8_t	roh_fin:1;	/* final option flag */
	u_int8_t	roh_type:7;	/* option type */
#else
	u_int8_t	roh_type:7;	/* option type */
	u_int8_t	roh_fin:1;	/* final option flag */
#endif
	u_int8_t	roh_optlen;	/* option len */
};

/* Normal RTPv1 options */
#define RTPv1OPT_CSRC	0	/* Content source */
#define RTPv1OPT_SSRC	1	/* Synchronization source */
#define RTPv1OPT_BOP	2	/* Beginning of playout unit */

/* RTPv1 synchronization source (SSRC) option header */
struct rtpv1ssrchdr {
#ifndef LITTLE_BITFIELDS
	u_int8_t	rsh_fin:1;	/* final option flag */
	u_int8_t	rsh_type:7;	/* option type */
#else
	u_int8_t	rsh_type:7;	/* option type */
	u_int8_t	rsh_fin:1;	/* final option flag */
#endif
	u_int8_t	rsh_optlen;	/* option len (== 1) */
	u_int16_t	rsh_id;		/* source identifier */
};

/* RTPv1 BOP option header */
struct rtpv1bophdr {
#ifndef LITTLE_BITFIELDS
	u_int8_t	rbh_fin:1;	/* final option flag */
	u_int8_t	rbh_type:7;	/* option type */
#else
	u_int8_t	rbh_type:7;	/* option type */
	u_int8_t	rbh_fin:1;	/* final option flag */
#endif
	u_int8_t	rbh_optlen;	/* option len (== 1) */
	u_int16_t	rbh_seq;	/* sequence number of BOP */
};

/* RTCPv1 forward direction options */
#define RTPv1OPT_FMT	32	/* Format description */
#define RTPv1OPT_SDESC	33	/* Source description */
#define RTPv1OPT_BYE	35	/* Conference exit notification */

/* RTCPv1 CDESC option header */
struct rtcpv1fmthdr {
#ifndef LITTLE_BITFIELDS
	u_int8_t	rtfh_fin:1;	/* final option flag */
	u_int8_t	rtfh_type:7;	/* option type */
#else
	u_int8_t	rtfh_type:7;	/* option type */
	u_int8_t	rtfh_fin:1;	/* final option flag */
#endif
	u_int8_t	rtfh_optlen;	/* option len */
#ifndef LITTLE_BITFIELDS
	u_int8_t	rtfh_x1:2;	/* reserved (must be 0) */
	u_int8_t	rtfh_fmt:6;	/* format id */
#else
	u_int8_t	rtfh_fmt:6;	/* format id */
	u_int8_t	rtfh_x1:2;	/* reserved (must be 0) */
#endif
	u_int8_t	rtfh_cqual;	/* clock quality */
};

/* RTCPv1 SDESC option header */
struct rtcpv1sdeschdr {
#ifndef LITTLE_BITFIELDS
	u_int8_t	rtsh_fin:1;	/* final option flag */
	u_int8_t	rtsh_type:7;	/* option type */
#else
	u_int8_t	rtsh_type:7;	/* option type */
	u_int8_t	rtsh_fin:1;	/* final option flag */
#endif
	u_int8_t	rtsh_optlen;	/* option len */
	u_int16_t	rtsh_id;	/* content source id */
	u_int32_t	rtsh_addr;	/* IP address of host */
};

/* RTCPv1 BYE option header */
struct rtcpv1byehdr {
#ifndef LITTLE_BITFIELDS
	u_int8_t	rtbh_fin:1;	/* final option flag */
	u_int8_t	rtbh_type:7;	/* option type */
#else
	u_int8_t	rtbh_type:7;	/* option type */
	u_int8_t	rtbh_fin:1;	/* final option flag */
#endif
	u_int8_t	rtbh_optlen;	/* option len */
	u_int16_t	rtbh_id;	/* content source id */
};

#endif /*_RTPv1_h*/


