// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Doug Lea (dl@rocky.oswego.edu)

This file is part of the GNU C++ Library.  This library is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#ifndef _SrcInfoAVL_h
#ifdef __GNUG__
#pragma interface
#endif
#define _SrcInfoAVL_h 1

#include "SrcInfo.Set.h"

struct SrcInfoAVLNode
{
  SrcInfoAVLNode*         lt;
  SrcInfoAVLNode*         rt;
  SrcInfo                 item;
  char                stat;
                      SrcInfoAVLNode(SrcInfo& h, SrcInfoAVLNode* l=0, SrcInfoAVLNode* r=0);
                      ~SrcInfoAVLNode();
};

inline SrcInfoAVLNode::SrcInfoAVLNode(SrcInfo& h, SrcInfoAVLNode* l, SrcInfoAVLNode* r)
:lt(l), rt(r), item(h), stat(0) {}

inline SrcInfoAVLNode::~SrcInfoAVLNode() {}

typedef SrcInfoAVLNode* SrcInfoAVLNodePtr;


class SrcInfoAVLSet : public SrcInfoSet
{
protected:
  SrcInfoAVLNode*   root;

                SrcInfoAVLSet(SrcInfoAVLNode* p, int l);

  SrcInfoAVLNode*   leftmost();
  SrcInfoAVLNode*   rightmost();
  SrcInfoAVLNode*   pred(SrcInfoAVLNode* t);
  SrcInfoAVLNode*   succ(SrcInfoAVLNode* t);
  void          _kill(SrcInfoAVLNode* t);
  void          _add(SrcInfoAVLNode*& t);
  void          _del(SrcInfoAVLNode* p, SrcInfoAVLNode*& t);

public:
                SrcInfoAVLSet();
                SrcInfoAVLSet(SrcInfoAVLSet& a);
  inline               ~SrcInfoAVLSet();

  Pix           add(SrcInfo& item);
  void          del(SrcInfo& item);
  inline int           contains(SrcInfo& item);

  inline void          clear();

  inline Pix           first();
  inline void          next(Pix& i);
  inline SrcInfo&          operator () (Pix i);
  int           owns(Pix i);
  Pix           seek(SrcInfo& item);

  Pix           last();
  void          prev(Pix& i);

  void          operator |= (SrcInfoAVLSet& b);
  void          operator -= (SrcInfoAVLSet& b);
  void          operator &= (SrcInfoAVLSet& b);

  int           operator == (SrcInfoAVLSet& b);
  int           operator != (SrcInfoAVLSet& b);
  int           operator <= (SrcInfoAVLSet& b); 

  int           OK();
};

inline SrcInfoAVLSet::~SrcInfoAVLSet()
{
  _kill(root);
}

inline SrcInfoAVLSet::SrcInfoAVLSet()
{
  root = 0;
  count = 0;
}

inline SrcInfoAVLSet::SrcInfoAVLSet(SrcInfoAVLNode* p, int l)
{
  root = p;
  count = l;
}

inline int SrcInfoAVLSet::operator != (SrcInfoAVLSet& b)
{
  return ! ((*this) == b);
}

inline Pix SrcInfoAVLSet::first()
{
  return Pix(leftmost());
}

inline Pix SrcInfoAVLSet::last()
{
  return Pix(rightmost());
}

inline void SrcInfoAVLSet::next(Pix& i)
{
  if (i != 0) i = Pix(succ((SrcInfoAVLNode*)i));
}

inline void SrcInfoAVLSet::prev(Pix& i)
{
  if (i != 0) i = Pix(pred((SrcInfoAVLNode*)i));
}

inline SrcInfo& SrcInfoAVLSet::operator () (Pix i)
{
  if (i == 0) error("null Pix");
  return ((SrcInfoAVLNode*)i)->item;
}

inline void SrcInfoAVLSet::clear()
{
  _kill(root);
  count = 0;
  root = 0;
}

inline int SrcInfoAVLSet::contains(SrcInfo& key)
{
  return seek(key) != 0;
}

#endif
