/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 
Copyright (c) 1995 International Computer Science Institute
All Rights Reserved.

Created by Wieland Holfelder (whd@icsi.berkeley.edu)
Tenet Group
International Computer Science Institute

Permission to use, copy, modify, and distribute this software and its 
documentation for educational, research and non-profit purposes, 
without fee, and without a written agreement is hereby granted, 
provided that the above copyright notice, this paragraph and the 
following five paragraphs appear in all copies.  

Permission to incorporate this software into commercial products may 
be obtained from the International Computer Science Institute,
1947 Center Street, Suite 600, Berkeley, CA 94704.  

IN NO EVENT SHALL THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE BE 
LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR 
CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE 
USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE INTERNATIONAL
COMPUTER SCIENCE INSTITUTE  HAS BEEN ADVISED OF THE POSSIBILITY OF 
SUCH DAMAGE.

THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE SPECIFICALLY DISCLAIMS 
ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE 
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE INTERNATIONAL
COMPUTER SCIENCE INSTITUTE  HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, 
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

The International Computer Science Institute will appreciate 
receiving any enhancements and modifications of this software 
made by others, and reserves the right to include them in future 
releases. The Institute also recommends that this software not 
be redistributed by third parties, but be obtained by all directly 
from the ICSI server, so that all users of the software may 
receive bug notifications and other relevant announcements.

The German Academic Exchange Service has supported the development 
of this software. Their support is gratefully acknowledged.

* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* $Id: SrcInfo.h,v 1.1 1996/07/26 15:06:42 whd Exp $ */

#ifndef _SRCINFO_H_
#define _SRCINFO_H_

#include <sys/types.h>
#include <netinet/in.h>
#include <String.h>
#include "it_depends.h"

class SrcInfo {
 public:

  /* for administration */
  int		new_src;
  int		active;
  u_int8_t	last_pt;
  u_int32_t	last_update;
  u_int32_t	vcr_offset;

  /* for identification */
  struct in_addr addr;

  /* RTPv2 has 32 bit ssrc's and csrc's */
  /* RTPv1 has 16 bit ssrc's and csrc's */
  union {
    /* they need to be overlayed in order to be able to compare them */
    u_int32_t	S_ssrc32;
    u_int16_t	S_ssrc16;
  } S_ssrc;
  #define ssrc16 S_ssrc.S_ssrc16
  #define ssrc32 S_ssrc.S_ssrc32
  
  union {
    /* they need to be overlayed in order to be able to compare them */
    u_int32_t	S_new_ssrc32;
    u_int16_t	S_new_ssrc16;
  } S_new_ssrc;
  #define new_ssrc16 S_new_ssrc.S_new_ssrc16
  #define new_ssrc32 S_new_ssrc.S_new_ssrc32

  u_int16_t	csrc16[16];
  u_int32_t	csrc32[16];

  int		nb_csrc;
  String	name;

  /* for timing */
  u_int32_t	last_media_ts;
  u_int32_t	media_offset;

  /* for statistics */

  u_int16_t	start_seq;
  u_int16_t	seq;

  u_int32_t	cycles;
  u_int32_t	expected;
  u_int32_t	packet_size;

  /* for playback */
	
  int		send_end_of_sync;
  int		end_of_sync_sent;
  u_int32_t	last_rtp_ts;
  u_int32_t	new_rtp_ts;

  inline int operator==(const SrcInfo& a) {
    return ((addr.s_addr==a.addr.s_addr)&&
	    ((ssrc32==a.ssrc32) ||
	     (ssrc32==a.new_ssrc32) ||
	     (new_ssrc32==a.ssrc32) ||
	     (new_ssrc32==a.new_ssrc32)));
  };
  
  inline int operator<(const SrcInfo& a) {
    return ((addr.s_addr)==(a.addr.s_addr)?
	    ((ssrc32==a.ssrc32)?(new_ssrc32<a.new_ssrc32):(ssrc32<a.ssrc32)):
	    (addr.s_addr<a.addr.s_addr));
  };
  
  inline int operator<=(const SrcInfo& a) {
    return ((addr.s_addr)==(a.addr.s_addr)?
	    ((ssrc32==a.ssrc32)?(new_ssrc32<=a.new_ssrc32):(ssrc32<a.ssrc32)):
	    (addr.s_addr<a.addr.s_addr));
  };
  
  inline int operator>(const SrcInfo& a) {
    return ((addr.s_addr)==(a.addr.s_addr)?
	    ((ssrc32==a.ssrc32)?(new_ssrc32>a.new_ssrc32):(ssrc32>a.ssrc32)):
	    (addr.s_addr>a.addr.s_addr));
  };
  
  inline int operator>=(const SrcInfo& a) {
    return ((addr.s_addr)==(a.addr.s_addr)?
	    ((ssrc32==a.ssrc32)?(new_ssrc32>=a.new_ssrc32):(ssrc32>a.ssrc32)):
	    (addr.s_addr>a.addr.s_addr));
  };

};

#endif
