// This may look like C code, but it is really -*- C++ -*-
/* 
Copyright (C) 1988 Free Software Foundation
    written by Dirk Grunwald (grunwald@cs.uiuc.edu)
    adapted for libg++ by Doug Lea (dl@rocky.oswego.edu)

This file is part of the GNU C++ Library.  This library is free
software; you can redistribute it and/or modify it under the terms of
the GNU Library General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.  This library is distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the GNU Library General Public License for more details.
You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free Software
Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifndef VCRPacketPHPQ_h
#ifdef __GNUG__
#pragma interface
#endif
#define VCRPacketPHPQ_h 1

#include "VCRPacket.PQ.h"

#ifndef VCRPacketPHPQIndex
#define VCRPacketPHPQIndex unsigned short
#endif

struct VCRPacketPHPQNode
{
  VCRPacketPHPQIndex   sibling;
  VCRPacketPHPQIndex   children;
  VCRPacket            item;
  char           valid;
};


class VCRPacketPHPQ : public VCRPacketPQ
{
  VCRPacketPHPQNode*  storage;        // table -- freelist in storage[0].sibling
  int           root;
  int           size;

  void          prealloc(int);
  int           check_sibling_list(int);

public:

                VCRPacketPHPQ(int sz = DEFAULT_INITIAL_CAPACITY);
                VCRPacketPHPQ(VCRPacketPHPQ&);
  inline               ~VCRPacketPHPQ();

  Pix           enq(VCRPacket& item);
  inline VCRPacket           deq(); 

  inline VCRPacket&          front();
  void          del_front();

  inline int           contains(VCRPacket& item);

  void          clear(); 

  Pix           first(); 
  void          next(Pix& i);
  inline VCRPacket&          operator () (Pix i);
  void          del(Pix i);
  Pix           seek(VCRPacket& item);

  int           OK();                    // rep invariant
};


inline VCRPacketPHPQ::~VCRPacketPHPQ()
{
  delete [] storage;
}


inline VCRPacket VCRPacketPHPQ::deq()
{
  if (count == 0) error("deq of empty PQ");
  VCRPacket x = storage[root].item;
  del_front();
  return x;
}


inline VCRPacket& VCRPacketPHPQ::front()
{
  if (count == 0) error("front of empty PQ");
  return storage[root].item;
}

inline int VCRPacketPHPQ::contains(VCRPacket& item)
{
  return seek(item) != 0;
}

inline VCRPacket& VCRPacketPHPQ::operator() (Pix p)
{
  if (p == 0) error("null Pix");
  return storage[int(p)].item;
}

#endif
