/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 
Copyright (c) 1995 International Computer Science Institute
All Rights Reserved.

Created by Wieland Holfelder (whd@icsi.berkeley.edu)
Tenet Group
International Computer Science Institute

Permission to use, copy, modify, and distribute this software and its 
documentation for educational, research and non-profit purposes, 
without fee, and without a written agreement is hereby granted, 
provided that the above copyright notice, this paragraph and the 
following five paragraphs appear in all copies.  

Permission to incorporate this software into commercial products may 
be obtained from the International Computer Science Institute,
1947 Center Street, Suite 600, Berkeley, CA 94704.  

IN NO EVENT SHALL THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE BE 
LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR 
CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE 
USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE INTERNATIONAL
COMPUTER SCIENCE INSTITUTE  HAS BEEN ADVISED OF THE POSSIBILITY OF 
SUCH DAMAGE.

THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE SPECIFICALLY DISCLAIMS 
ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE 
PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE INTERNATIONAL
COMPUTER SCIENCE INSTITUTE  HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, 
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

The International Computer Science Institute will appreciate 
receiving any enhancements and modifications of this software 
made by others, and reserves the right to include them in future 
releases. The Institute also recommends that this software not 
be redistributed by third parties, but be obtained by all directly 
from the ICSI server, so that all users of the software may 
receive bug notifications and other relevant announcements.

The German Academic Exchange Service has supported the development 
of this software. Their support is gratefully acknowledged.

* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static const char rcsid[]="$Id: vcr.cc,v 1.1 1996/07/26 15:02:20 whd Exp whd $";

extern "C" {
#include <std.h>
#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <tk.h>

#include <ff.xbm>
#include <goto_end.xbm>
#include <goto_start.xbm>
#include <led.xbm>
#include <next_blank.xbm>
#include <next_index_1.xbm>
#include <next_index_2.xbm>
#include <next_index_3.xbm>
#include <next_index_4.xbm>
#include <next_index_5.xbm>
#include <next_speaker.xbm>
#include <next_time.xbm>
#include <next_ts.xbm>
#include <play.xbm>
#include <prev_blank.xbm>
#include <prev_index_1.xbm>
#include <prev_index_2.xbm>
#include <prev_index_3.xbm>
#include <prev_index_4.xbm>
#include <prev_index_5.xbm>
#include <prev_speaker.xbm>
#include <prev_time.xbm>
#include <prev_ts.xbm>
#include <rec.xbm>
#include <rew.xbm>
#include <stop.xbm>

#include "it_depends.h"

  /* vcr's tcl2c strings */
  extern const char tcl_alert[];
  extern const char tcl_edit_media[];
  extern const char tcl_edit_session[];
  extern const char tcl_fs_box[];
  extern const char tcl_how2start[];
  extern const char tcl_index_ctrl[];
  extern const char tcl_num_display[];
  extern const char tcl_sd_import[];
  extern const char tcl_tkerror[];
  extern const char tcl_vcr_init[];
  extern const char tcl_vcr[];

  /* tcl2c strings from the tcl library */
  extern const char tcl_init[];

  /* tcl2c strings from the tk library */
  extern const char tcl_button[];
  extern const char tcl_dialog[];
  extern const char tcl_entry[];
  extern const char tcl_focus[];
  extern const char tcl_listbox[];
  extern const char tcl_menu[];
  extern const char tcl_optMenu[];
  extern const char tcl_palette[];
  extern const char tcl_scale[];
  extern const char tcl_scrlbar[];
  extern const char tcl_tearoff[];
  extern const char tcl_text[];
  extern const char tcl_tk[];

  /* strings generated by the Makefile */
  extern const char version[];
  extern const char host_vendor[];
  extern const char host_cpu[];
  extern const char host_os[];
  extern const char host_target[];
  extern const char ps_cmd[];

  char *optarg;
  int optind;
  int opterr;
}

#include <String.h>
#include "VCRTcl.h"
#include "VCRDisplay.h"
#include "MBoneSession.h"

static MBoneSession *session_ptr=NULL;

static void adios(int sig)
{
  VCRTcl& tcl = VCRTcl::instance();
  if (session_ptr) delete session_ptr;
  tcl.evalc("exit");
#ifdef DEBUG1
  if (sig) cout << "vcr: received signal: " << sig << ", adios() called.\n"; 
#endif
  exit(0);
}

static int usage(ClientData cd =NULL, Tcl_Interp* t =NULL, int argc =0, char **argv =NULL) {
    cerr << (char *) version << " usage:\n\n";
    cerr << "\tvcr\t[-d[isplay] display] [-f init_file] [-h[elp]]\n";
    cerr << "\t\t[-l[oop]] [[-s] session_file[.vcr]]\n\n";
    return TCL_OK;
};

static int time2str(ClientData cd =NULL, Tcl_Interp* t =NULL, int argc =0, char **argv =NULL) {
  time_t ct;
  char *st;
  if (argc == 2 && argv[1]) {
    sscanf(argv[1],"%d",&ct);
    if (ct!=0) {
      st = ctime(&ct);
      Tcl_AppendResult(t, st, 0);
    } else {
      Tcl_AppendResult(t, "unknown", 0);
    }
    return TCL_OK;
  };
  return TCL_ERROR;
};

void loadbitmaps(Tcl_Interp* tcl)
{
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_led"),
			led_bits, led_width, led_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_play"),
			play_bits, play_width, play_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_rec"),
			rec_bits, rec_width, rec_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_stop"),
			stop_bits, stop_width, stop_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_goto_end"),
			goto_end_bits, goto_end_width, goto_end_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_goto_start"),
			goto_start_bits, goto_start_width, goto_start_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_ff"),
			ff_bits, ff_width, ff_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_rew"),
			rew_bits, rew_width, rew_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_next_speaker"),
			next_speaker_bits, next_speaker_width, next_speaker_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_next_index_1"),
			next_index_1_bits, next_index_1_width, next_index_1_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_next_index_2"),
			next_index_2_bits, next_index_2_width, next_index_2_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_next_index_3"),
			next_index_3_bits, next_index_3_width, next_index_3_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_next_index_4"),
			next_index_4_bits, next_index_4_width, next_index_4_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_next_index_5"),
			next_index_5_bits, next_index_5_width, next_index_5_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_next_ts"),
			next_ts_bits, next_ts_width, next_ts_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_next_time"),
			next_time_bits, next_time_width, next_time_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_next_blank"),
			next_blank_bits, next_blank_width, next_blank_height);

	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_prev_speaker"),
			prev_speaker_bits, prev_speaker_width, prev_speaker_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_prev_index_1"),
			prev_index_1_bits, prev_index_1_width, prev_index_1_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_prev_index_2"),
			prev_index_2_bits, prev_index_2_width, prev_index_2_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_prev_index_3"),
			prev_index_3_bits, prev_index_3_width, prev_index_3_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_prev_index_4"),
			prev_index_4_bits, prev_index_4_width, prev_index_4_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_prev_index_5"),
			prev_index_5_bits, prev_index_5_width, prev_index_5_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_prev_ts"),
			prev_ts_bits, prev_ts_width, prev_ts_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_prev_time"),
			prev_time_bits, prev_time_width, prev_time_height);
	Tk_DefineBitmap(tcl, Tk_GetUid("bmp_prev_blank"),
			prev_blank_bits, prev_blank_width, prev_blank_height);
}


int main(int argc, char** argv)
{
  char* display=NULL;
  int interactive = 0;
  int mkindex = 0;
  Tk_Window tk;
  
  signal(SIGHUP, adios);
  signal(SIGINT, adios);
  signal(SIGQUIT, adios);
  signal(SIGPIPE, adios);
  signal(SIGTERM, adios);

  /* this is to avoid the "Suspended (tty input)  vcr" error that causes      */
  /* vcr to core dump when started in the background (at least under solaris).*/
  close(0);

  /*
   * Parse command-line arguments processed in the C++ code.
   */
  char **arg = argv;
  int n=argc;
  int tcl_argc=0;
  String tcl_argv="";
  while (*arg && n) {
    if (strncmp((const char *) *arg, "-help", 2) == 0) {
      usage();
      exit(1);
    } else if (strncmp((const char *) *arg, "-display", 2) == 0) {
      arg++;
      if (*arg) display = strdup(*arg);
#ifdef DEVELOPMENT_VERSION
    } else if (strncmp((const char *) *arg, "-tcl", 2) == 0) {
      interactive=1;
    } else if (strncmp((const char *) *arg, "-mki", 2) == 0) {
      mkindex=1;
#endif
   } else {
      tcl_argc++;
      tcl_argv += "\"";
      tcl_argv += *arg; 
      tcl_argv += "\" ";
    };
    arg++;
    n--;
  };

  /* pass remaining commandline arguments to the tcl script */
  char new_tcl_argc[50], *new_tcl_argv;
  sprintf(new_tcl_argc,"%d",tcl_argc);
  /* ... and remember, we have to cleanup the memory later! */
  new_tcl_argv = (char *) malloc(tcl_argv.length()+1);
  strcpy(new_tcl_argv, tcl_argv.chars());

  VCRTcl::init("vcr");
  VCRTcl& tcl = VCRTcl::instance();

#ifdef TCL_MEM_DEBUG
  Tcl_InitMemory(tcl.interp());
#endif

#ifdef DEVELOPMENT_VERSION
  if (interactive) {
    tcl.evalc("source [info library]/init.tcl");
  } else {
#endif
    tcl.evalc(tcl_init);
#ifdef DEVELOPMENT_VERSION
  }
#endif

  loadbitmaps(tcl.interp());

  Tcl_SetVar(tcl.interp(), "VERSION", (char *) version, TCL_GLOBAL_ONLY);
  Tcl_SetVar(tcl.interp(), "PS_CMD", (char *) ps_cmd, TCL_GLOBAL_ONLY);
  Tcl_SetVar(tcl.interp(), "MAGIC", MBONE_VCR_MAGIC, TCL_GLOBAL_ONLY);

  Tcl_SetVar(tcl.interp(), "HOST_VENDOR", (char *) host_vendor, TCL_GLOBAL_ONLY);
  Tcl_SetVar(tcl.interp(), "HOST_CPU", (char *) host_cpu, TCL_GLOBAL_ONLY);
  Tcl_SetVar(tcl.interp(), "HOST_OS", (char *) host_os, TCL_GLOBAL_ONLY);
  Tcl_SetVar(tcl.interp(), "HOST_TARGET", (char *) host_target, TCL_GLOBAL_ONLY);

#ifdef DEMO_VERSION
  Tcl_SetVar(tcl.interp(), "DEMO_VERSION", "1", TCL_GLOBAL_ONLY);
#else
  Tcl_SetVar(tcl.interp(), "DEMO_VERSION", "0", TCL_GLOBAL_ONLY);
#endif

#ifdef DEVELOPMENT_VERSION
  Tcl_SetVar(tcl.interp(), "DEVELOPMENT_VERSION", "1", TCL_GLOBAL_ONLY);
#else
  Tcl_SetVar(tcl.interp(), "DEVELOPMENT_VERSION", "0", TCL_GLOBAL_ONLY);
#endif

  /*
   * If a display was specified, put it into the DISPLAY
   * environment variable so that it will be available for
   * any sub-processes created by us.
   */
  if (display != NULL) {
    Tcl_SetVar2(tcl.interp(), "env", "DISPLAY", display, TCL_GLOBAL_ONLY);
    free(display);
  }

  if (Tk_Init(tcl.interp()) == TCL_OK)
  tk = Tk_MainWindow(tcl.interp());

  if (tk == 0) {
    fprintf(stderr, "vcr: %s\n", tcl.result());
    exit(1);
  }

  tcl.tkmain(tk);

  Tk_MakeWindowExist(tk);

  /* initialize the display */
  VCRDisplay::init();

  tcl.CreateCommand("usage", usage, (ClientData) NULL);
  tcl.CreateCommand("time2str", time2str, (ClientData) NULL);

#ifdef DEVELOPMENT_VERSION
  if (interactive) {
    /* This is VCR's INTERACTIVE MODE which is only valid in the DEVELOPMENT_VERSION */

    Tcl_SetVar(tcl.interp(), "interactive", "0", TCL_GLOBAL_ONLY);

    /* Find out where vcr's tcltk library is located and add it to tcl's auto_path.    */
    /* We check the environment variable VCR_TCLTK_LIB first, if it's not set or not   */
    /* vaild check "." afterwards check "./tcl_tk" and finally "../tcl_tk" if we still */
    /* couldn't find it, give up.						       */
	
    tcl.evalc("set vcr_tcltk_lib \"\"");
    tcl.evalc("if {[info exists env(VCR_TCLTK_LIB)] == \"1\"} {set vcr_tcltk_lib $env(VCR_TCLTK_LIB)}");
    tcl.evalc("
	if {$vcr_tcltk_lib == \"\" || [file exists ${vcr_tcltk_lib}/vcr.tcl] == 0} {
		set vcr_tcltk_lib .
		if {[file exists ${vcr_tcltk_lib}/vcr.tcl] == 0} {
			set vcr_tcltk_lib ./tcl_tk
			if {[file exists ${vcr_tcltk_lib}/vcr.tcl] == 0} {
				set vcr_tcltk_lib ../tcl_tk
				if {[file exists ${vcr_tcltk_lib}/vcr.tcl] == 0} {
					puts \"$VERSION can't start in interactive mode\"
					puts \"because VCR's TCL/TK library could not be found.\"
					puts \"(set VCR_TCLTK_LIB to the correct path)\" 
					exit 1
				}
			}
		}
	}");
    tcl.evalc("set auto_path [linsert $auto_path 0 $vcr_tcltk_lib]");

    /* if we have to rebuild the tclIndex, do it now */
    if (mkindex) {
      tcl.evalc("puts -nonewline \"$VERSION rebuilding tclIndex in $vcr_tcltk_lib ... \"");
      tcl.evalc("auto_mkindex $vcr_tcltk_lib *.tcl");
      tcl.evalc("puts \"done.\n\"");
    }

    tcl.evalc("source ${tk_library}/tk.tcl");

    /* pass the commandline arguments to the tcl script not earlier then here */
    Tcl_SetVar(tcl.interp(), "argc", new_tcl_argc, TCL_GLOBAL_ONLY);
    Tcl_SetVar(tcl.interp(), "argv", new_tcl_argv, TCL_GLOBAL_ONLY);

    /* run this script last */
    tcl.evalc("source ${vcr_tcltk_lib}/vcr_init.tcl");

  } else {
#endif

    Tcl_SetVar(tcl.interp(), "interactive", "0", TCL_GLOBAL_ONLY);

    /* load tk library code */
    tcl.evalc(tcl_tk);
    tcl.evalc(tcl_button);
    tcl.evalc(tcl_dialog);
    tcl.evalc(tcl_entry);
    tcl.evalc(tcl_focus);
    tcl.evalc(tcl_listbox);
    tcl.evalc(tcl_menu);
    tcl.evalc(tcl_optMenu);
    tcl.evalc(tcl_palette);
    tcl.evalc(tcl_scale);
    tcl.evalc(tcl_scrlbar);
    tcl.evalc(tcl_tearoff);
    tcl.evalc(tcl_text);

    /* load vcr's tcl code */
    tcl.evalc(tcl_alert);
    tcl.evalc(tcl_edit_media);
    tcl.evalc(tcl_edit_session);
    tcl.evalc(tcl_fs_box);
    tcl.evalc(tcl_how2start);
    tcl.evalc(tcl_index_ctrl);
    tcl.evalc(tcl_num_display);
    tcl.evalc(tcl_sd_import);
    tcl.evalc(tcl_tkerror);
    tcl.evalc(tcl_vcr);

    /* pass the commandline arguments to the tcl script not earlier then here */
    Tcl_SetVar(tcl.interp(), "argc", new_tcl_argc, TCL_GLOBAL_ONLY);
    Tcl_SetVar(tcl.interp(), "argv", new_tcl_argv, TCL_GLOBAL_ONLY);

    /* run this script last */
    tcl.evalc(tcl_vcr_init);

#ifdef DEVELOPMENT_VERSION
  };
#endif

  /* don't forget to cleanup! */
  free(new_tcl_argv); 

  /* create the session */
  session_ptr=new MBoneSession();

  if (session_ptr->get_state()==INVALID) {
    cerr << (char *) version << " error:\n";
    cerr << session_ptr->get_error_msg() << "\n";
  } else {
    /* let tcl deal with the remaining command-line parameters... */
    tcl.evalc("parse_cmd_line");

    /* and now display the GUI */
    Tk_MapWindow(tk);
    Tk_MainLoop();

  };

  adios(0);
};

/* this is to avoid problems during linking in some environments */
void Tcl_AppInit() {};


