#/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#
# Copyright (c) 1995 International Computer Science Institute
# All Rights Reserved.
#
# Created by Wieland Holfelder (whd@icsi.berkeley.edu)
# Tenet Group
# International Computer Science Institute
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for educational, research and non-profit purposes,
# without fee, and without a written agreement is hereby granted,
# provided that the above copyright notice, this paragraph and the
# following five paragraphs appear in all copies.
#
# Permission to incorporate this software into commercial products may
# be obtained from the International Computer Science Institute,
# 1947 Center Street, Suite 600, Berkeley, CA 94704.
#
# IN NO EVENT SHALL THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE BE
# LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
# CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
# USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE INTERNATIONAL
# COMPUTER SCIENCE INSTITUTE  HAS BEEN ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE SPECIFICALLY DISCLAIMS
# ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
# PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE INTERNATIONAL
# COMPUTER SCIENCE INSTITUTE  HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE,
# SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
#
# The International Computer Science Institute will appreciate
# receiving any enhancements and modifications of this software
# made by others, and reserves the right to include them in future
# releases. The Institute also recommends that this software not
# be redistributed by third parties, but be obtained by all directly
# from the ICSI server, so that all users of the software may
# receive bug notifications and other relevant announcements.
#
# The German Academic Exchange Service has supported the development
# of this software. Their support is gratefully acknowledged.
#
#* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#
# $Id: edit_session.tcl,v 1.3 1996/08/12 11:35:25 whd Exp $ (ICSI)
#
#* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
global new_session init

# procedure edit_session
proc edit_session {{new "1"}} {

    global VERSION DEMO_VERSION session tmp_session media new_session

    clear_tmp_session
    clear_media
    fill_tmp_session

    if {$new == 1} {
	set tmp_session(read_only) ""
	set tmp_session(name) ""
	set tmp_session(saved) "0"
	set new_session "1"
    } else {
	set new_session "0"
    }

    # build widget .edit_session
    toplevel .edit_session -relief {raised}

    # Window manager configurations
    wm positionfrom .edit_session "program"
    wm sizefrom .edit_session "program"

    if { $tmp_session(read_only) != "" } {
	wm title .edit_session "$VERSION (View Session)"
    } elseif {$new == "0"} {
	wm title .edit_session "$VERSION (Edit Session)"
    } else {
	wm title .edit_session "$VERSION (New Session)"
    }

    trace_session

    # build widget .edit_session.top
    frame .edit_session.top \
	    -borderwidth {2} \
	    -relief {raised}

    # build widget .edit_session.top.title
    label .edit_session.top.title \
	    -height {2} \
	    -font {-*-helvetica-bold-r-normal--*-140-*}

    if { $tmp_session(read_only) != "" } {
	.edit_session.top.title configure -text "View Session"
    } elseif { $new == "0" } {
	.edit_session.top.title configure -text "Edit Session"
    } else {
	.edit_session.top.title configure -text "New Session"
    }

    # pack widget .edit_session.top
    pack append .edit_session.top \
	    .edit_session.top.title {top frame center padx 100 pady 10}

    # build widget .edit_session.middle
    frame .edit_session.middle \
	    -borderwidth {2} \
	    -relief {raised}

    # build widget .edit_session.middle.session_name
    frame .edit_session.middle.session_name \
	    -borderwidth {4} \
	    -relief {raised}

    # build widget .edit_session.middle.session_name.label
    label .edit_session.middle.session_name.label \
	    -text {Session Name:}

    # build widget .edit_session.middle.session_name.entry
    entry .edit_session.middle.session_name.entry \
	    -relief {sunken} \
	    -textvariable {tmp_session(name)}

    bind .edit_session.middle.session_name.entry <Return> {
	focus .edit_session.middle.session_description.text.text
    }

    # pack widget .edit_session.middle.session_name
    pack append .edit_session.middle.session_name \
	    .edit_session.middle.session_name.label {left frame center fill} \
	    .edit_session.middle.session_name.entry {left frame center expand fill}

    # build widget .edit_session.middle.session_description
    frame .edit_session.middle.session_description \
	    -borderwidth {4} \
	    -relief {raised}

    # build widget .edit_session.middle.session_description.header
    frame .edit_session.middle.session_description.header \
	    -borderwidth {2}

    # build widget .edit_session.middle.session_description.header.label
    label .edit_session.middle.session_description.header.label \
	    -borderwidth {0} \
	    -text {Session Description:}

    # pack widget .edit_session.middle.session_description.header
    pack append .edit_session.middle.session_description.header \
	    .edit_session.middle.session_description.header.label {left frame w}

    # build widget .edit_session.middle.session_description.text
    frame .edit_session.middle.session_description.text

    # build widget .edit_session.middle.session_description.text.scrollbar
    scrollbar .edit_session.middle.session_description.text.scrollbar \
	    -command {.edit_session.middle.session_description.text.text yview} \
	    -relief {raised}

    # build widget .edit_session.middle.session_description.text.text
    text .edit_session.middle.session_description.text.text \
	    -borderwidth {2} \
	    -height {5} \
	    -relief {sunken} \
	    -width {60} \
	    -wrap {word} \
	    -yscrollcommand {.edit_session.middle.session_description.text.scrollbar set}


    # pack widget .edit_session.middle.session_description.text
    pack append .edit_session.middle.session_description.text \
	    .edit_session.middle.session_description.text.scrollbar {left frame center filly} \
	    .edit_session.middle.session_description.text.text {top frame center expand fill}

    # pack widget .edit_session.middle.session_description
    pack append .edit_session.middle.session_description \
	    .edit_session.middle.session_description.header {top frame center fill} \
	    .edit_session.middle.session_description.text {bottom frame center padx 5 pady 5}

    # build widget .edit_session.middle.scope
    frame .edit_session.middle.scope \
	    -borderwidth {4} \
	    -relief {raised}

    # build widget .edit_session.middle.scope.label
    label .edit_session.middle.scope.label \
	    -text {Max Scope:}

    # build widget .edit_session.middle.scope.host
    radiobutton .edit_session.middle.scope.host \
	    -command {
	.edit_session.middle.scope.entry configure -state disabled
	set tmp_session(max_ttl) [scope2ttl $tmp_session(scope)]
    }\
	    -relief {flat} \
	    -text {host} \
	    -value {host} \
	    -variable {tmp_session(scope)}

    # build widget .edit_session.middle.scope.site
    radiobutton .edit_session.middle.scope.site \
	    -command {
	.edit_session.middle.scope.entry configure -state disabled
	set tmp_session(max_ttl) [scope2ttl $tmp_session(scope)]
    }\
	    -relief {flat} \
	    -text {site} \
	    -value {site} \
	    -variable {tmp_session(scope)}

    # build widget .edit_session.middle.scope.region
    radiobutton .edit_session.middle.scope.region \
	    -command {
	.edit_session.middle.scope.entry configure -state disabled
	set tmp_session(max_ttl) [scope2ttl $tmp_session(scope)]
    }\
	    -relief {flat} \
	    -text {region} \
	    -value {region} \
	    -variable {tmp_session(scope)}

    # build widget .edit_session.middle.scope.continent
    radiobutton .edit_session.middle.scope.continent \
	    -command {
	.edit_session.middle.scope.entry configure -state disabled
	set tmp_session(max_ttl) [scope2ttl $tmp_session(scope)]
    }\
	    -relief {flat} \
	    -text {cont.} \
	    -value {continent} \
	    -variable {tmp_session(scope)}

    # build widget .edit_session.middle.scope.world
    radiobutton .edit_session.middle.scope.world \
	    -command {
	.edit_session.middle.scope.entry configure -state disabled
	set tmp_session(max_ttl) [scope2ttl $tmp_session(scope)]
    }\
	    -relief {flat} \
	    -text {world} \
	    -value {world} \
	    -variable {tmp_session(scope)}

    # build widget .edit_session.middle.scope.other
    radiobutton .edit_session.middle.scope.other \
	    -command {
	.edit_session.middle.scope.entry configure -state normal
    }\
	    -relief {flat} \
	    -text {} \
	    -value {other} \
	    -variable {tmp_session(scope)}

    # build widget .edit_session.middle.scope.entry
    entry .edit_session.middle.scope.entry \
	    -relief {sunken} \
	    -textvariable {tmp_session(max_ttl)} \
	    -width {5}


    # pack widget .edit_session.middle.scope
    pack append .edit_session.middle.scope \
	    .edit_session.middle.scope.label {left frame center} \
	    .edit_session.middle.scope.host {left frame center} \
	    .edit_session.middle.scope.site {left frame center} \
	    .edit_session.middle.scope.region {left frame center} \
	    .edit_session.middle.scope.continent {left frame center} \
	    .edit_session.middle.scope.world {left frame center} \
	    .edit_session.middle.scope.other {left frame center} \
	    .edit_session.middle.scope.entry {left frame center}

    # build widget .edit_session.middle.media_list
    frame .edit_session.middle.media_list \
	    -borderwidth {2} \
	    -relief {raised}

    # build widget .edit_session.middle.media_list.label
    frame .edit_session.middle.media_list.label \
	    -borderwidth {2} \
	    -relief {raised}

    # build widget .edit_session.middle.media_list.label.label
    label .edit_session.middle.media_list.label.label \
	    -text {Media List:}

    # pack widget .edit_session.middle.media_list.label
    pack append .edit_session.middle.media_list.label \
	    .edit_session.middle.media_list.label.label {left frame center}

    # build widget .edit_session.middle.media_list.list
    frame .edit_session.middle.media_list.list

    # build widget .edit_session.middle.media_list.list.scrollbar_x
    scrollbar .edit_session.middle.media_list.list.scrollbar_x \
	    -command {.edit_session.middle.media_list.list.listbox xview} \
	    -orient {horizontal} \
	    -relief {raised}

    # build widget .edit_session.middle.media_list.list.scrollbar_y
    scrollbar .edit_session.middle.media_list.list.scrollbar_y \
	    -command {.edit_session.middle.media_list.list.listbox yview} \
	    -relief {raised}

    # build widget .edit_session.middle.media_list.list.listbox
    listbox .edit_session.middle.media_list.list.listbox \
	    -relief {groove} \
	    -selectmode {extended} \
	    -xscrollcommand {.edit_session.middle.media_list.list.scrollbar_x set} \
	    -yscrollcommand {.edit_session.middle.media_list.list.scrollbar_y set}

    bind .edit_session.middle.media_list.list.listbox <Double-Button-1> {
	if { $DEVELOPMENT_VERSION == "0" } {
	    vcr_grab release .edit_session
	}
	.edit_session.middle.media_list.buttons.edit invoke
    }

    # pack widget .edit_session.middle.media_list.list
    pack append .edit_session.middle.media_list.list \
	    .edit_session.middle.media_list.list.scrollbar_y {left frame center filly} \
	    .edit_session.middle.media_list.list.listbox {top frame center expand fill} \
	    .edit_session.middle.media_list.list.scrollbar_x {bottom frame center fillx}

    # build widget .edit_session.middle.media_list.buttons
    frame .edit_session.middle.media_list.buttons \
	    -borderwidth {2} \
	    -relief {raised}

    if { $tmp_session(read_only) == "" } {
	# build widget .edit_session.middle.media_list.buttons.add
	button .edit_session.middle.media_list.buttons.add \
		-text {Add Media}\
		-command {
	    vcr_grab release .edit_session
	    edit_media [expr $tmp_session(nb_of_media) + 1] 1
	}
    }

    # build widget .edit_session.middle.media_list.buttons.edit
    button .edit_session.middle.media_list.buttons.edit \
	    -text {Edit Media}\
	    -command {
	global session tmp_session VVERSION

	set media_nb [.edit_session.middle.media_list.list.listbox curselection]

	if { "$media_nb" == "" } {
	    if { [.edit_session.middle.media_list.list.listbox size] == 1 } {
		vcr_grab release .edit_session
		edit_media 1
	    } else {
		AlertBox "\n $VERSION error: \n\n No media selected. \n" "" ""\
			"$VERSION (Edit Media)" Ok
	    }
	} elseif { [llength "$media_nb"] > 1 } {
	    AlertBox "\n $VERSION error: \n\n More than one media selected. \n" "" ""\
		    "$VERSION (Edit Media)" Ok
	} else {
	    vcr_grab release .edit_session
	    edit_media [expr $media_nb + 1]
	}
    }


    if { $tmp_session(read_only) == "" } {
	# build widget .edit_session.middle.media_list.buttons.delete
	button .edit_session.middle.media_list.buttons.delete \
		-text {Delete Media}\
		-command {
	    global session tmp_session VVERSION

	    set selection [.edit_session.middle.media_list.list.listbox curselection]
	    if {$selection == ""} {
		AlertBox "\n $VERSION error: \n\n No media selected. \n" "" ""\
			"$VERSION (Delete Media)" Ok
	    } else {
		if { [AlertBox "\n Delete media $selection ? \n" "" ""\
			"$VERSION (Delete Media)" Ok Cancel] == "0"} {
		    set offset 0
		    foreach index $selection {
			.edit_session.middle.media_list.list.listbox delete [expr $index - $offset]
			incr offset
			if { [expr $index + 1] == $session(index_media) } {
			    switch_media 0
			}
			for {set i [expr $index + 1]} \
			    {$i < $tmp_session(nb_of_media)} \
			    {incr i} {
				foreach x [array names media] {
				    set tmp_session(media_list,$i,$x) \
					    $tmp_session(media_list,[expr $i+1],$x)
				}
			}
		    }
		    if {[.edit_session.middle.media_list.list.listbox size] == 0} {
			.edit_session.middle.media_list.buttons.edit configure -state disabled
			.edit_session.middle.media_list.buttons.delete configure -state disabled
		    }

		    set tmp_session(nb_of_media) [.edit_session.middle.media_list.list.listbox size]
		    update_index_menu 1
		}
	    }

	}
    }

    if { $tmp_session(read_only) == "" } {
	# pack widget .edit_session.middle.media_list.buttons
	pack append .edit_session.middle.media_list.buttons \
		.edit_session.middle.media_list.buttons.add \
		{left frame center padx 5 expand fill} \
		.edit_session.middle.media_list.buttons.edit \
		{left frame center padx 5 expand fill} \
		.edit_session.middle.media_list.buttons.delete \
		{left frame center padx 5 expand fill}
    } else {
	# pack widget .edit_session.middle.media_list.buttons
	pack append .edit_session.middle.media_list.buttons \
		.edit_session.middle.media_list.buttons.edit \
		{left frame center padx 5 expand fill}
    }

    # pack widget .edit_session.middle.media_list
    pack append .edit_session.middle.media_list \
	    .edit_session.middle.media_list.label {top frame center fill} \
	    .edit_session.middle.media_list.list {top frame center fill} \
	    .edit_session.middle.media_list.buttons {top frame center fill}

    # build widget .edit_session.middle.filename
    frame .edit_session.middle.filename \
	    -borderwidth {2} \
	    -relief {raised}

    # build widget .edit_session.middle.filename.frame
    frame .edit_session.middle.filename.frame \
	    -relief {raised}

    # build widget .edit_session.middle.filename.frame.label_button
    button .edit_session.middle.filename.frame.label_button \
	    -text {Session Filename:} \
	    -command {
	if { $tmp_session(filename) == "" } { set tmp_session(filename) [name2filename $tmp_session(name)] }
	set file [FSBox "Select MBone VCR Session" "$tmp_session(filename)" "" "" "*.vcr" "MBone VCR v0.1 (Load Session)"]
	if { $file != "" } {
	    set tmp_session(filename) $file
	}
    }

    # build widget .edit_session.middle.filename.frame.entry
    entry .edit_session.middle.filename.frame.entry \
	    -textvariable {tmp_session(filename)} \
	    -relief {sunken}

    # pack widget .edit_session.middle.filename.frame
    pack append .edit_session.middle.filename.frame \
	    .edit_session.middle.filename.frame.label_button {left frame center fill} \
	    .edit_session.middle.filename.frame.entry {right frame center expand fill}

    # pack widget .edit_session.middle.filename
    pack append .edit_session.middle.filename \
	    .edit_session.middle.filename.frame {top frame center fill}

    # pack widget .edit_session.middle
    pack append .edit_session.middle \
	    .edit_session.middle.session_name {top frame center fillx} \
	    .edit_session.middle.session_description {top frame center fill} \
	    .edit_session.middle.scope {top frame center fill} \
	    .edit_session.middle.media_list {top frame center fill} \
	    .edit_session.middle.filename {top frame center fillx}

    if {$tmp_session(read_only) == ""} {
	# build widget .edit_session.bottom
	frame .edit_session.import \
		-borderwidth {2} \
		-relief {raised}
	button .edit_session.import.sd_import \
		-text {Import Session from SD} \
		-width {59} \
		-command {
	    if { [winfo exists .sd_import] == 0} {
		if { [file exists "~/.sd_cache"] == 1 } {
		    sd_import
		} else {
		    AlertBox "\n $VERSION: \n\n Cannot import sessions from SD. \n\
			    File ~/.sd_cache not found. \n\n"\
			    "" "" "$VERSION (Import Session)" Ok
		}
	    } else {
		wm deiconify .sd_import
	    }
	}
	pack append .edit_session.import \
		.edit_session.import.sd_import {left frame center expand padx 20 pady 20} \
    }


    # build widget .edit_session.bottom
    frame .edit_session.bottom \
	    -borderwidth {2} \
	    -relief {raised}

    if {$tmp_session(read_only) == ""} {
	# build widget .edit_session.bottom.ok
	button .edit_session.bottom.ok \
		-text {Ok} \
		-width {15} \
		-command {
	    if {$tmp_session(name) == ""} {
		AlertBox "\n $VERSION error: \n\n Please specify a session name. \n"\
			"" "" "$VERSION (Edit Session)" Ok
		return
	    }

	    set $tmp_session(filename) [check_file_ext $tmp_session(filename)]

	    if { [scope_warning $tmp_session(max_ttl)] == 1 } {
		return
	    }

	    set x [.edit_session.middle.session_description.text.text get 0.0 end]
	    if {[string compare $tmp_session(desc) $x] != 0} {
		set tmp_session(desc) $x
	    }
	    untrace_session
	    fill_session
	    if { $new_session == 1 } {
		set ret [save_session]
		enable_recording
	    } else {
		set ret 1
	    }
	    if { $ret == 1 } {
		update_session
	    }
	    update_main_win
	    update_index_menu
	    if { [winfo exists .sd_import] == 1} {destroy .sd_import}
	    destroy .edit_session
	}
	# build widget .edit_session.bottom.save
	button .edit_session.bottom.save \
		-text {Save} \
		-width {15} \
		-command {
	    if {$tmp_session(name) == ""} {
		AlertBox "\n $VERSION error: \n\n Please specify a session name. \n"\
			"" "" "$VERSION (Edit Session)" Ok
		return
	    }
	    if { $tmp_session(filename) != [check_file_ext $tmp_session(filename)] } {
		AlertBox "\n\n Session filename: \n \
			\"$tmp_session(filename)\" \n \
			did not end on \".vcr\" (corrected). \n\n"\
			"" "" "$VERSION" Ok
		set tmp_session(filename) [check_file_ext $tmp_session(filename)]
		return
	    }

	    if { [scope_warning $tmp_session(max_ttl)] == 1 } {
		return
	    }

	    set x [.edit_session.middle.session_description.text.text get 0.0 end]
	    if {[string compare $tmp_session(desc) $x] != 0} {
		set tmp_session(desc) $x
	    }
	    untrace_session
	    set old_session_filename $session(filename)
	    fill_session
	    if {$session(filename) != $old_session_filename} {
		set ret [save_session $old_session_filename]
	    } else {
		set ret [save_session]
	    }
	    if { $ret == "1" } {
		update_session
	    }
	    update_main_win
	    trace_session
	}
    }

    # build widget .edit_session.bottom.cancel
    button .edit_session.bottom.cancel\
	    -text {Cancel}\
	    -width {15}\
	    -command {

	untrace_session
	update_main_win
	update_index_menu

	if { [winfo exists .sd_import] == 1} {destroy .sd_import}
	destroy .edit_session
    }

    if {$tmp_session(read_only) == ""} {
	# pack widget .edit_session.bottom
	pack append .edit_session.bottom \
		.edit_session.bottom.ok {left frame center expand padx 20 pady 20} \
		.edit_session.bottom.save {left frame center expand padx 20 pady 20} \
		.edit_session.bottom.cancel {left frame center expand padx 20 pady 20}
    } else {
	# pack widget .edit_session.bottom
	pack append .edit_session.bottom \
		.edit_session.bottom.cancel {expand frame center padx 60 pady 20}
    }

    if {$tmp_session(read_only) == ""} {
	# pack widget .edit_session
	pack append .edit_session \
		.edit_session.top {top frame center expand fill} \
		.edit_session.middle {top frame center expand fill} \
		.edit_session.import {top frame center expand fill} \
		.edit_session.bottom {top frame center expand fill}
    } else {
	# pack widget .edit_session
	pack append .edit_session \
		.edit_session.top {top frame center expand fill} \
		.edit_session.middle {top frame center expand fill} \
		.edit_session.bottom {top frame center expand fill}
    }

    if { $tmp_session(scope) != "other" } {
	.edit_session.middle.scope.entry configure -state disabled
    }

    update_session_win
    vcr_grab .edit_session
}

proc update_session_win {} {
    global tmp_session new_session

    if { [winfo exists .edit_session] == "0" } { return }

    if {$tmp_session(desc) != ""} {
	.edit_session.middle.session_description.text.text configure -state normal
	.edit_session.middle.session_description.text.text delete 0.0 end
	.edit_session.middle.session_description.text.text insert 0.0 "$tmp_session(desc)"
    }

    .edit_session.middle.media_list.list.listbox delete 0 end

    if {$tmp_session(nb_of_media) > 0} {
	for {set i 1} { $i <= $tmp_session(nb_of_media)} {incr i} {
	    if {$tmp_session(media_list,$i,media_desc) != ""} {
		.edit_session.middle.media_list.list.listbox insert end\
			[format "%s (%s)" $tmp_session(media_list,$i,media_name) \
			                  $tmp_session(media_list,$i,media_desc)]
	    } else {
		.edit_session.middle.media_list.list.listbox insert end\
			$tmp_session(media_list,$i,media_name)
	    }
	}
    }

    set tmp_session(scope) [ttl2scope $tmp_session(max_ttl)]

    if {$tmp_session(read_only)!=""} {
	.edit_session.top.title configure -text "View Session"
	.edit_session.middle.session_description.text.text configure -state disabled
	.edit_session.middle.session_name.entry configure -state disabled
 	.edit_session.middle.scope.host configure -state disabled
	.edit_session.middle.scope.site configure -state disabled
	.edit_session.middle.scope.region configure -state disabled
	.edit_session.middle.scope.world configure -state disabled
	.edit_session.middle.scope.other configure -state disabled
	.edit_session.middle.scope.entry configure -state disabled
	catch {.edit_session.middle.media_list.buttons.add configure -state disabled}
	.edit_session.middle.media_list.buttons.edit configure -text "View Media"
	catch {.edit_session.middle.media_list.buttons.delete configure -state disabled}
	catch {.edit_session.bottom.save configure -state disabled}
    } else {
	if { $new_session == "0" } {
	    .edit_session.top.title configure -text "Edit Session"
	} else {
	    .edit_session.top.title configure -text "New Session"
	}
	.edit_session.middle.session_description.text.text configure -state normal
	.edit_session.middle.session_name.entry configure -state normal
	.edit_session.middle.scope.host configure -state normal
	.edit_session.middle.scope.site configure -state normal
	.edit_session.middle.scope.region configure -state normal
	.edit_session.middle.scope.world configure -state normal
	.edit_session.middle.scope.other configure -state normal
	.edit_session.middle.scope.entry configure -state normal

	catch {.edit_session.middle.media_list.buttons.add configure -state normal}
	.edit_session.middle.media_list.buttons.edit configure -text "Edit Media"
	catch {.edit_session.middle.media_list.buttons.delete configure -state normal}
	catch {.edit_session.bottom.save configure -state normal}
    }

    if {[.edit_session.middle.media_list.list.listbox size] == 0} {
	catch {.edit_session.middle.media_list.buttons.edit configure -state disabled}
	catch {.edit_session.middle.media_list.buttons.delete configure -state disabled}
    } else {
	catch {.edit_session.middle.media_list.buttons.edit configure -state normal}
	catch {.edit_session.middle.media_list.buttons.delete configure -state normal}
    }
}


proc read_session_file { {filename ""} {force 0} } {
    global MAGIC READ_ONLY VERSION DEMO_VERSION
    global tmp_session session nb_of_indices index_list

    if {$filename == ""} {
	if { $force == 0} {
	    AlertBox "\n $VERSION error: \n\n No filename specified. \n"\
		    "" "" "$VERSION" Ok
	}
	return "0"
    }

    if {[file exist $filename] == 0} {
	if { $force == 0} {
	    AlertBox "\n $VERSION error: \n\n File $filename not found. \n"\
		    "" "" "$VERSION" Ok
	}
	return "0"
    }

    if {[file readable $filename] == 0} {
	if { $force == 0} {
	    AlertBox "\n $VERSION error: \n\n File $filename not readable. \n"\
		    "" "" "$VERSION" Ok
	}
	return "0"
    }

    if {[file writable $filename] == 1} {
	set tmp_session(read_only) ""
    } else {
	set tmp_session(read_only) $READ_ONLY
    }

    if { $DEMO_VERSION == "1" } {
	set tmp_session(read_only) $READ_ONLY
    }

    set tmp_session(nb_of_media) "0"
    set tmp_session(filename) $filename
    set tmp_session(data_dir) [file dirname $tmp_session(filename)]
    set tmp_session(file_prefix) [file rootname [file tail $tmp_session(filename)]]

    set sf [open $filename r]
    set line [get_next_line $sf]
    if {$line != ""} {
	if {$MAGIC != $line} {
	    if { $force == 0} {
		AlertBox "\n $VERSION error: \n\n File \"$filename\" has \n wrong type $line \n\n\
			(should be $MAGIC) \n\n" "" "" "$VERSION" Ok
	    }
	    return "0"

	}
    } else {
	if { $force == 0} {
	    AlertBox "\n $VERSION error: \n\n Read error on file \"$filename\". \n\n\
		    EOF reached. \n\n" "" "" "$VERSION" Ok
	}
	return "0"
    }

    set nb_of_media "0"
    set line [get_next_line $sf]
    while {$line !=""} {
	set pattern [string range $line 0 [expr [string first "=" $line] - 1]]
	set line [string range $line [expr [string length $pattern] + 1] end]
	switch $pattern {
	    "session_date"    {set tmp_session(date) $line}
	    "session_name"    {set tmp_session(name) $line}
	    "session_desc"    {
		set tmp_session(desc) $line
		regsub -all "%n" $tmp_session(desc) "\n" tmp_session(desc)
	    }
	    "session_max_ttl" {set tmp_session(max_ttl) $line}
	    default {
		if {$pattern == $nb_of_media} {
		    if {[parse_media $nb_of_media $line $filename] == 0} {
			return "0"
		    }
		} elseif {[expr $nb_of_media + 1] == $pattern} {
		    set nb_of_media [expr $nb_of_media + 1]
		    set tmp_session(media_list,$pattern,media_type)	"1"
		    set tmp_session(media_list,$pattern,media_name)	""
		    set tmp_session(media_list,$pattern,media_desc)	""
		    set tmp_session(media_list,$pattern,media_ip)	""
		    set tmp_session(media_list,$pattern,media_port)	""
		    set tmp_session(media_list,$pattern,media_ttl)	"1"
		    set tmp_session(media_list,$pattern,media_scope)	\
			    [ttl2scope $tmp_session(media_list,$pattern,media_ttl)]
		    set tmp_session(media_list,$pattern,media_cid)	"0"
		    set tmp_session(media_list,$pattern,media_cmd)	""
		    set tmp_session(media_list,$pattern,media_mute)	"0"

		    if {[parse_media $nb_of_media $line $filename] == 0} {
			return "0"
		    }
		} else {
		    if { $force == 0} {
			AlertBox "\n $VERSION error: \n\n File $filename has wrong format. \n\
				Unknown pattern: \"$pattern\". \n" "" "" "$VERSION" Ok
		    }
		    return "0"
		}
	    }
	}
	set line [get_next_line $sf]
    }
    set tmp_session(nb_of_media) $nb_of_media
    set index_media "0"
    set tmp_session(index_media) "0"
    for {set i 1} {$i <= $nb_of_media} {incr i} {
	set nb_of_indices($i) 0
	set index_list($i) {}
	set tmp_session(media_list,$i,media_scope) \
		[ttl2scope $tmp_session(media_list,$i,media_ttl)]
	if { $tmp_session(media_list,$i,media_type) == 2 } { set index_media $i	}
    }
    if { ($nb_of_media > 0) && ($index_media == "0") } { set index_media "1" }
    if { $session(indicate) == 1 } {
	set colors(led,normal) "gray60"
	set colors(led(blink) "limegreen"
	global leds
	for { set i 1 } { $i <= 10 } { incr i } {
	    set_led $i $leds($i)
	}
    }
    set tmp_session(saved) "1"
    fill_session
    switch_media $index_media
    set_goto_start_end
    
    if { $session(read_only) != "" } {
	.vcr.top.record.m entryconfigure 0 -state disabled
	.vcr.top.record.m entryconfigure 1 -state disabled
	.vcr.top.record.m entryconfigure 2 -state disabled
    }

    return "1"
}

proc parse_media {{media_nb} {line} {filename}} {
    global tmp_session VERSION

    set pattern [string range $line 0 [expr [string first "=" $line]-1]]
    set line [string range $line [expr [string length $pattern]+1] end]
    switch $pattern {
	"media_type"   {set tmp_session(media_list,$media_nb,media_type)   $line}
	"media_name"   {set tmp_session(media_list,$media_nb,media_name)   $line}
	"media_desc"   {set tmp_session(media_list,$media_nb,media_desc)   $line}
	"media_ip"     {set tmp_session(media_list,$media_nb,media_ip)     $line}
	"media_port"   {set tmp_session(media_list,$media_nb,media_port)   $line}
	"media_ttl"    {set tmp_session(media_list,$media_nb,media_ttl)    $line}
	"media_scope"  {set tmp_session(media_list,$media_nb,media_scope)  $line}
	"media_cid"    {set tmp_session(media_list,$media_nb,media_cid)    $line}
	"media_cmd"    {set tmp_session(media_list,$media_nb,media_cmd)    $line}
	"media_mute"   {set tmp_session(media_list,$media_nb,media_mute)   $line}
	default {
	    AlertBox "\n $VERSION error: \n\n File $filename has wrong format. \n\
		    Unknow pattern: \"$pattern\". \n" "" "" "$VERSION" Ok
	    return "0"
	}
    }
    return "1"
}

proc name2filename { name } {
    regsub -all "\[ \t\]" $name "_" new_name
    regsub -all "\[-\"`';:,<>{}[]()~!?#$%^&*+=|\\\]" $new_name "" new_name
    while { [regsub -all "__" $new_name "_" new_name] == 1 } {}
    return $new_name
}

# read next no-comment line from file sf
proc get_next_line {sf} {
    set length 0
    set line ""
    while {([string first "#" $line] == 0) || ($length==0)} {
	if {[gets $sf line] < 0} {return ""}
	set line [string trim $line]
	set length [string length $line]
    }
    return $line
}

proc wrong_format {{filename ""}} {
    global {VERSION}
    AlertBox "\n $VERSION error: \n\n File $filename has wrong format. \n" "" "" "$VERSION" Ok
    return "0"
}


proc ttl2scope {{ttl "1"}} {
    switch $ttl {
	"0"	{return "host"}
	"15"	{return "site"}
	"47"	{return "region"}
	"63"	{return "continent"}
	"127"	{return "world"}
	default	{return "other"}
    }
}

proc scope2ttl {{scope "host"} {other "0"}} {
    switch $scope {
	"host"		{return "0"}
	"site"		{return "15"}
	"region"	{return "47"}
	"continent"	{return "63"}
	"world"		{return "127"}
	"other"		{return $other}
	default		{return "0"}
    }
}

proc clear_session {} {
    global session init READ_ONLY
    set session(state)		"idle"
    set session(read_only)	$READ_ONLY
    set session(date)		"0"
    set session(loaded)		"0"
    set session(rec_enabled)	$init(rec_enabled)
    set session(max_ttl)	"[scope2ttl site]"
    set session(scope)		[ttl2scope $session(max_ttl)]
    set session(name)		[center "No session" 40]
    set session(desc)		""
    set session(filename)	""
    set session(data_dir)	""
    set session(nb_of_media)	"0"
    set session(saved)		"1"
    set session(speed)		"1"
    set session(loop_mode)	$init(loop_mode)
    set session(blink_mode)	"1"
    set session(start_blinking)	""
    set session(active_button)	""
    set session(index_ctrl)	"0"
    set session(index_media)	"0"
    set session(file_prefix)	""
    set session(start_recording) $init(start_recording)
    set session(ff_rew_factor)   $init(ff_rew_factor)
    set session(indicate)        $init(indicate)
    set session(auto_start)	$init(auto_start)
    set session(rec_timeout)	$init(rec_timeout)
    set session(tools_started)	""
}

proc clear_tmp_session {} {
    global tmp_session init READ_ONLY
    set tmp_session(state)		"idle"
    set tmp_session(read_only)		$READ_ONLY
    set tmp_session(date)		""
    set tmp_session(loaded)		"0"
    set tmp_session(rec_enabled)	$init(rec_enabled)
    set tmp_session(max_ttl)		"[scope2ttl site]"
    set tmp_session(scope)		[ttl2scope $tmp_session(max_ttl)]
    set tmp_session(name)		[center "No session" 40]
    set tmp_session(desc)		""
    set tmp_session(filename)		""
    set tmp_session(data_dir)		""
    set tmp_session(nb_of_media)	"0"
    set tmp_session(speed)		"1"
    set tmp_session(loop_mode)		$init(loop_mode)
    set tmp_session(blink_mode)		"1"
    set tmp_session(start_blinking)	""
    set tmp_session(active_button)	""
    set tmp_session(index_ctrl)		"0"
    set tmp_session(index_media)	"0"
    set tmp_session(file_prefix)	""
    set tmp_session(start_recording)    $init(start_recording)
    set tmp_session(ff_rew_factor)	$init(ff_rew_factor)
    set tmp_session(indicate)		$init(indicate)
    set tmp_session(auto_start)		$init(auto_start)
    set tmp_session(rec_timeout)	$init(rec_timeout)
    set tmp_session(tools_started)	""
}


proc clear_media {} {
    global media media_saved media_nb
    set media(media_type)	"1"
    set media(media_name)	""
    set media(media_desc)	""
    set media(media_ip)		""
    set media(media_port)	""
    set media(media_ttl)	"[scope2ttl site]"
    set media(media_scope)	[ttl2scope $media(media_ttl)]
    set media(media_cid)	"0"
    set media(media_cmd)	""
    set media(media_mute)	"0"
    set media_saved		"1"
    set media_nb		"0"
}

proc update_session {} {
    global media session tmp_session VERSION nb_of_indices index_list

    for {set i 1} {$i <= $session(nb_of_media)} {incr i} {
	if {$session(media_list,$i,media_ttl) > $session(max_ttl)} {
	    set session(media_list,$i,media_ttl) $session(max_ttl)
	    set session(media_list,$i,media_scope) [ttl2scope $session(max_ttl)]
	}
	set nb_of_indices($i) 0
    }

    if { ($session(loaded) == 1) } {
	if { ($session(saved) == 0) } {
	    set last_ts [vcr_ctrl get_ts]
	    set session(loaded) [vcr_ctrl delete]
	    set session(loaded) [vcr_ctrl read]
	    if { $session(loaded) == "0" } {
		set error [vcr_ctrl get_error]
		vcr_ctrl delete
		AlertBox "\n $VERSION error: \n\n Cannot open session \"$tmp_session(name)\" \
			  \nError: \n $error \n\n" "" "" "$VERSION" Ok
		fill_session
		update_main_win
		return "0"
	    }
	    vcr_ctrl goto ms $last_ts
	}
    } else {
	set session(loaded) [vcr_ctrl read]
	if { $session(loaded) == "0" } {
	    set error [vcr_ctrl get_error]
	    vcr_ctrl delete
	    AlertBox "\n $VERSION error: \n\n Cannot open session \"$session(name)\" \n\
		    Error: \n $error \n\n" "" "" "$VERSION" Ok
	    update_main_win
	    return "0"
	}
    }

    update_index_menu
    fill_tmp_session
    switch_media $session(index_media)

    return "1"
}

proc fill_tmp_session {} {
    global media session tmp_session

    set tmp_session(loaded) $session(loaded)
    set tmp_session(read_only) $session(read_only)
    set tmp_session(saved) $session(saved)
    set tmp_session(name) $session(name)
    set tmp_session(desc) $session(desc)
    set tmp_session(scope) $session(scope)
    set tmp_session(max_ttl) $session(max_ttl)
    set tmp_session(nb_of_media) $session(nb_of_media)
    set tmp_session(filename) $session(filename)
    set tmp_session(file_prefix) [file rootname [file tail $session(filename)]]
    set tmp_session(data_dir) [file dirname $session(filename)]
    set tmp_session(file_prefix) $session(file_prefix)
    set tmp_session(index_media) $session(index_media)
    set tmp_session(start_recording) $session(start_recording)
    set tmp_session(ff_rew_factor) $session(ff_rew_factor)

    for {set i 1} { $i <= $tmp_session(nb_of_media) } {incr i} {
	foreach x [array names media] {
	    set tmp_session(media_list,$i,$x) $session(media_list,$i,$x)
	}
    }
}

proc fill_session {} {
    global media session tmp_session

    set session(loaded) $tmp_session(loaded)
    set session(read_only) $tmp_session(read_only)
    set session(saved) $tmp_session(saved)
    set session(name) $tmp_session(name)
    set session(desc) $tmp_session(desc)
    set session(max_ttl) $tmp_session(max_ttl)
    set session(scope) $tmp_session(scope)
    set session(filename) $tmp_session(filename)
    set session(nb_of_media) $tmp_session(nb_of_media)
    set session(file_prefix) [file rootname [file tail $tmp_session(filename)]]
    set session(data_dir) [file dirname $tmp_session(filename)]
    set session(index_media) $tmp_session(index_media)
    set session(start_recording) $tmp_session(start_recording)
    set session(ff_rew_factor) $tmp_session(ff_rew_factor)

    for {set i 1} { $i <= $tmp_session(nb_of_media) } {incr i} {
	foreach x [array names media] {
	    set session(media_list,$i,$x) $tmp_session(media_list,$i,$x)
	}
    }
}

proc save_session {{current_session ""}} {
    global VERSION MAGIC session media protocols

    if { $session(read_only) != "" } {
	AlertBox "\n $VERSION error: \n\n Session \"$session(name)\" is read only. \n\n"\
		"" "" "$VERSION (Save Session)" Ok
	return "0"
    }

    if { [file exists $session(filename)] == "1" } {
	if { [AlertBox "\n File \"$session(filename)\" already exist. \n\n Overwrite? \n\n"\
		"" "" "$VERSION (Save Session)" Yes Cancel] == 1} {
	    return "0"
	}
	if { ([file writable $session(filename)] == "0") } {
	    AlertBox "\n $VERSION error: \n\n Cannot write to file \
		    \"$session(filename)\". \n\n\
		    Permission denied. \n\n" "" ""\
		    "$VERSION (Save Session)" Ok
	    return "0"
	}
	set files "$session(data_dir)/$session(file_prefix)_*_*.idx"
	catch "eval exec rm -f ${files}.idx"
	catch "eval exec rm -f ${files}.rec"
    }

    set session(data_dir) [file dirname $session(filename)]
    set session(file_prefix) [file rootname [file tail $session(filename)]]

    if {($current_session != "") && ($current_session != $session(filename))} {
	if { [file exist $current_session] == "1" } {
	    set file_size 0
	    set curr_name "[file dirname $current_session]/[file rootname [file tail $current_session]]"

	    for {set i 1} { $i <= $session(nb_of_media) } {incr i} {
		set filename "${curr_name}_$protocols($session(media_list,$i,media_type))_$i"
		set file_size [expr $file_size + [file size ${filename}.idx]]
		set file_size [expr $file_size + [file size ${filename}.rec]]
	    }

	    set ret [AlertBox "\n\n $VERSION: \n\n Do you want to copy the data files of \n\
		    this session to the new sessionname \n ($file_size Bytes) \n\
		    or start with an empty session? \n\n"\
		    "" "" "$VERSION (Save Session)" {Copy Data} {Empty Session} {Cancel}]

	    if { $ret == 0 } {
		set session(data_dir) [file dirname $session(filename)]

		for {set i 1} { $i <= $session(nb_of_media) } {incr i} {
		    set filename "${curr_name}_$protocols($session(media_list,$i,media_type))_$i"
		    set new_filename "$session(data_dir)/$session(file_prefix)_$protocols($session(media_list,$i,media_type))_$i"
		    set error ""
		    catch "eval exec cp ${filename}.idx ${new_filename}.idx" error
		    if { $error != "" } {
			AlertBox "\n $VERSION error: \n\n Could not copy file \
				\"${new_filename}.idx\".\n\n\
				($error) \n\n" "" "" "$VERSION (Save Session)" Ok
			catch "eval exec rm -f ${new_filename}.idx"
			return "0"
		    }

		    set error ""
		    catch "eval exec cp ${filename}.rec ${new_filename}.rec" error
		    if { $error != "" } {
			AlertBox "\n $VERSION error: \n\n Could not copy file \
				\"${new_filename}.rec\".\n\n\
				($error) \n\n" "" "" "$VERSION (Save Session)" Ok
			catch "eval exec rm -f ${new_filename}.idx
			catch "eval exec rm -f ${new_filename}.rec
			return "0"
		    }
		}
	    } elseif { $ret == 1 } {
		set session(loaded) 1
		set session(saved) 0
		enable_recording
	    } elseif { $ret == 2 } {
		return "0"
	    }
	}
    }

    global errorInfo
    set errorInfo ""
    catch "set sf [open $session(filename) w]"
    if { "$errorInfo" != "" } {
	AlertBox "\n $VERSION error: \n\n Cannot write to file \"$session(filename)\" \n\
		Permission denied. \n\n ($errorInfo)\n\n" "" "" "$VERSION" Ok
	return "0"
    }
    puts $sf $MAGIC
    set date [eval exec "date"]
    puts $sf "# This file was generated by $VERSION."
    puts $sf "session_date=$date"
    puts $sf "session_name=$session(name)"
    set x $session(desc)
    regsub -all "\n" $x "%n" x
    puts $sf "session_desc=$x"
    puts $sf "session_max_ttl=$session(max_ttl)"
    for {set i 1} { $i <= $session(nb_of_media) } {incr i} {
	foreach x [array names media] {
	    if { $x == "media_ttl" } {
		if { $session(media_list,$i,$x) > $session(max_ttl) } {
		    set session(media_list,$i,$x) $session(max_ttl)
		}
		puts $sf "$i=$x=$session(media_list,$i,$x)"
	    } elseif { $x == "media_scope" } {
		if { $session(media_list,$i,$x) != $session(scope) } {
		    set session(media_list,$i,$x) $session(scope)
		}
	    } else {
		puts $sf "$i=$x=$session(media_list,$i,$x)"
	    }
	}
    }
    set error ""
    catch "close $sf" error
    if { $error != "" } {
	AlertBox "\n $VERSION error: \n\n Could not save session \"$session(name)\".\
		\n\n ($error) \n\n" "" "" "$VERSION (Save Session)" Ok
	clear_session
	update_main_win
	return "0"
    }
    if { ($session(loaded) == 1) && ($session(saved) == 0) } {
	set last_ts [vcr_ctrl get_ts]
	set session(data_dir) [file dirname $session(filename)]
	set session(file_prefix) [file rootname [file tail $session(filename)]]
	set session(loaded) [vcr_ctrl delete]
	set session(loaded) [vcr_ctrl read]
	if { $session(loaded) == "0" } {
	    set error [vcr_ctrl get_error]
	    vcr_ctrl delete
	    update_main_win
	    AlertBox "\n $VERSION error: \n\n Cannot re-open session \"$session(name)\" \n\
		    Error: \n $error \n\n" "" "" "$VERSION" Ok
	    clear_session
	    update_main_win
	    return "0"
	}
	vcr_ctrl goto ms $last_ts
    }
    set session(saved) "1"
    set session(saved) "1"
    return "1"
}

proc session_need_save { name element op } {
    global tmp_session
    set tmp_session(saved) "0"
    return
}

proc trace_session {} {
    global session tmp_session
    if { $session(read_only) == "" } {
	catch {
	    trace variable tmp_session(max_ttl) w session_need_save
	    trace variable tmp_session(name) w session_need_save
	    trace variable tmp_session(desc) w session_need_save
	    trace variable tmp_session(filename) w session_need_save
	    trace variable tmp_session(nb_of_media) w session_need_save
	}
    }
}

proc untrace_session {} {
    global session tmp_session
    if { $session(read_only) == "" } {
	catch {
	    trace vdelete tmp_session(max_ttl) w session_need_save
	    trace vdelete tmp_session(name) w session_need_save
	    trace vdelete tmp_session(desc) w session_need_save
	    trace vdelete tmp_session(filename) w session_need_save
	    trace vdelete tmp_session(nb_of_media) w session_need_save
	}
    }
}

proc scope_warning {ttl} {
    global VERSION
    if { $ttl>16 } {
	return [AlertBox "\n $VERSION Warning:\n\n \
		Are you sure you want to set the scope to $ttl?\n\n \
		If you play the session, you are going to send \n \
		your packets beyond the scope of your site. \n" "" "" \
		"$VERSION (Scope Warning)" "Set ttl to $ttl" Cancel]
    } else {
	return 0
    }
}


# eof
#
