#/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#
# Copyright (c) 1995 International Computer Science Institute
# All Rights Reserved.
#
# Created by Wieland Holfelder (whd@icsi.berkeley.edu)
# Tenet Group
# International Computer Science Institute
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for educational, research and non-profit purposes,
# without fee, and without a written agreement is hereby granted,
# provided that the above copyright notice, this paragraph and the
# following five paragraphs appear in all copies.
#
# Permission to incorporate this software into commercial products may
# be obtained from the International Computer Science Institute,
# 1947 Center Street, Suite 600, Berkeley, CA 94704.
#
# IN NO EVENT SHALL THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE BE
# LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
# CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE
# USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE INTERNATIONAL
# COMPUTER SCIENCE INSTITUTE  HAS BEEN ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE SPECIFICALLY DISCLAIMS
# ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
# PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE INTERNATIONAL
# COMPUTER SCIENCE INSTITUTE  HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE,
# SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
#
# The International Computer Science Institute will appreciate
# receiving any enhancements and modifications of this software
# made by others, and reserves the right to include them in future
# releases. The Institute also recommends that this software not
# be redistributed by third parties, but be obtained by all directly
# from the ICSI server, so that all users of the software may
# receive bug notifications and other relevant announcements.
#
# The German Academic Exchange Service has supported the development
# of this software. Their support is gratefully acknowledged.
#
#* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#
# $Id: sd_import.tcl,v 1.2 1996/08/12 07:57:35 whd Exp $ (ICSI)
#
#* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
global new_session init

# procedure to show window .sd_import
proc sd_import {{new 0}} {

    global VERSION DEMO_VERSION session tmp_session media new_session

    if { [file exists "~/.sd_cache"] == 0 } {
	AlertBox "\n $VERSION: \n\n Cannot import sessions from SD. \n\
		File ~/.sd_cache not found. \n\n"\
		"" "" "$VERSION (Import Session)" Ok
	return 0
    }

    # build widget .edit_session
    toplevel .sd_import
    vcr_grab .sd_import

    # Window manager configurations
    global tk_version
    wm positionfrom .sd_import program
    wm sizefrom .sd_import program

    wm title .sd_import "$VERSION (Import Session from SD)"

    # build widget .sd_import.top
    frame .sd_import.top \
	    -borderwidth {2} \
	    -relief {ridge}

    # build widget .sd_import.top.title
    label .sd_import.top.title \
	    -borderwidth {0} \
	    -font {-*-helvetica-bold-r-normal--*-140-*} \
	    -height {2} \
	    -text {Import Session from SD}

    file stat "~/.sd_cache" file_info
    label .sd_import.top.update \
	    -borderwidth {0} \
	    -text "Last ~/.sd_cache update: [time2str $file_info(ctime)]"

    # pack widget .sd_import.top
    pack append .sd_import.top \
	    .sd_import.top.title {top frame center expand fill} \
	    .sd_import.top.update {top frame center expand fill}

    # build widget .sd_import.session
    frame .sd_import.session \
	    -borderwidth {2} \
	    -relief {ridge}

    label .sd_import.session.label \
	    -borderwidth {0} \
	    -text {Sessions:}

    # build widget .sd_import.session.frame
    frame .sd_import.session.frame

    # build widget .sd_import.session.frame.x_scroll
    scrollbar .sd_import.session.frame.x_scroll \
	    -command {.sd_import.session.frame.list xview} \
	    -orient {horizontal} \
	    -width {10} \
	    -relief {raised}

    # build widget .sd_import.session.frame.y_scroll
    scrollbar .sd_import.session.frame.y_scroll \
	    -command {.sd_import.session.frame.list yview} \
	    -width {10} \
	    -relief {raised}

    # build widget .sd_import.session.frame.list
    listbox .sd_import.session.frame.list \
	    -relief {ridge} \
	    -width {45} \
	    -height {15} \
	    -xscrollcommand {.sd_import.session.frame.x_scroll set} \
	    -yscrollcommand {.sd_import.session.frame.y_scroll set}

    bind .sd_import.session.frame.list <Button-1> {
	global sd_session_index
	set index [%W nearest %y]
	%W select anchor $index
	%W select set $index
	set sd_session_index [expr $index + 1]
	update_media_sd_list [expr $index + 1]
    }

    if { $new == 1 } {
	bind .sd_import.session.frame.list <Double-Button-1> {
	    add_new_sd_session
	}
    } else {
	bind .sd_import.session.frame.list <Double-Button-1> {
	    add_sd_session
	}
    }

    if {$new == 0} {
	bind .sd_import.session.frame.list <Double-Button-2> {
	    add_sd_title
	}
    }

    # pack widget .sd_import.session.frame
    pack append .sd_import.session.frame \
	    .sd_import.session.frame.y_scroll {right frame center filly} \
	    .sd_import.session.frame.list {top frame center expand fill} \
	    .sd_import.session.frame.x_scroll {bottom frame center fillx}

    # pack widget .sd_import.session
    pack append .sd_import.session \
	    .sd_import.session.label {top frame center expand fill} \
	    .sd_import.session.frame {top frame center expand fill}

    # build widget .media
    frame .sd_import.media \
	    -borderwidth {2} \
	    -relief {ridge}

    label .sd_import.media.label \
	    -borderwidth {0} \
	    -text {}

    # build widget .sd_import.media.frame
    frame .sd_import.media.frame

    # build widget .sd_import.media.frame.x_scroll
    scrollbar .sd_import.media.frame.x_scroll \
	    -command {.sd_import.media.frame.list xview} \
	    -orient {horizontal} \
	    -width {10} \
	    -relief {raised}

    # build widget .sd_import.media.frame.y_scroll
    scrollbar .sd_import.media.frame.y_scroll \
	    -command {.sd_import.media.frame.list yview} \
	    -width {10} \
	    -relief {raised}

    # build widget .sd_import.media.frame.list
    listbox .sd_import.media.frame.list \
	    -relief {raised} \
	    -width {45} \
	    -height {5} \
	    -xscrollcommand {.sd_import.media.frame.x_scroll set} \
	    -yscrollcommand {.sd_import.media.frame.y_scroll set}

    if {$new == 0} {
	bind .sd_import.media.frame.list <Double-Button-1> {
	    add_sd_media
	}
    }

    # pack widget .sd_import.media.frame
    pack append .sd_import.media.frame \
	    .sd_import.media.frame.y_scroll {right frame center filly} \
	    .sd_import.media.frame.list {top frame center expand fill} \
	    .sd_import.media.frame.x_scroll {bottom frame center fillx}

    # pack widget .sd_import.media
    pack append .sd_import.media \
	    .sd_import.media.label {top frame center fill}\
	    .sd_import.media.frame {top frame center fill}

    frame .sd_import.buttons

    frame .sd_import.buttons.get

    if {$new == 0} {
	# build widget .sd_import.buttons.get.session
	button .sd_import.buttons.get.session \
		-text {Get Session} \
		-width {15} \
		-command {
	    add_sd_session
	}

	# build widget .sd_import.buttons.get.title
	button .sd_import.buttons.get.title \
		-text {Get Title} \
		-width {15}  \
		-command {
	    add_sd_title
	}

	# build widget .sd_import.buttons.get.media
	button .sd_import.buttons.get.media \
		-text {Add Media} \
		-width {15} \
		-command {
	    add_sd_media
	}

	pack append .sd_import.buttons.get \
		.sd_import.buttons.get.session {left frame center expand padx 20 pady 20} \
		.sd_import.buttons.get.title {left frame center expand padx 20 pady 20} \
		.sd_import.buttons.get.media {left frame center expand padx 20 pady 20}
    } else {
	# build widget .sd_import.buttons.get.session
	button .sd_import.buttons.get.session \
		-text {Import Session} \
		-width {32} \
		-command {
	    add_new_sd_session
	}

	pack append .sd_import.buttons.get \
		.sd_import.buttons.get.session {left frame center expand padx 20 pady 20}
    }

    # build widget .dismiss
    frame .sd_import.buttons.dismiss \
	    -borderwidth {2} \
	    -relief {ridge}

    # build widget .sd_import.buttons.dismiss.dismiss
    button .sd_import.buttons.dismiss.dismiss \
	    -text {Dismiss} \
	    -width {32} \
	    -command {
	destroy .sd_import
    }

    # pack widget .sd_import.buttons.dismiss
    pack append .sd_import.buttons.dismiss \
	    .sd_import.buttons.dismiss.dismiss {top frame center expand}

    # pack widget .sd_import.buttons
    pack append .sd_import.buttons \
	    .sd_import.buttons.get {top frame center expand} \
	    .sd_import.buttons.dismiss {top frame center expand}

    # pack widget .sd_import.
    pack append .sd_import \
	    .sd_import.top {top frame center expand fill} \
	    .sd_import.session {top frame center expand fill} \
	    .sd_import.media {top frame center expand fill} \
	    .sd_import.buttons {top frame center pady 20 expand fill}

    read_sd_cache
}

proc read_sd_cache {} {
    global sd_list VERSION
    set sdf [open "~/.sd_cache" r]
    set line [get_next_line $sdf]
    set counter 0
    while {$line != ""} {
	set prefix [string index $line 0]
	set value [string range $line 2 end]
	case $prefix in {
	    {n} {
		if {$counter > 0 } {
		    set sd_list($counter,media) $media_counter
		}
		incr counter
		set sd_list($counter,name) ""
		set sd_list($counter,desc) ""
		set sd_list($counter,addr) ""
		set sd_list($counter,ttl) 0
		set sd_list($counter,media) 0
		set media_counter 0
	    }
	    {s} {
		set sd_list($counter,name) $value
	    }
	    {i} {
		set sd_list($counter,desc) $value
	    }
	    {o} {
		set sd_list($counter,desc) "$sd_list($counter,desc) ($value)"
	    }
	    {c} {
		set sd_list($counter,addr) [lindex [split $value] 0]
		set sd_list($counter,ttl) [lindex [split $value] 1]
	    }
	    {m} {
		incr media_counter
		set sd_list($counter,$media_counter,name) [lindex [split $value] 0]
		set sd_list($counter,$media_counter,addr) $sd_list($counter,addr)
		set sd_list($counter,$media_counter,port) [lindex [split $value] 1]
		set sd_list($counter,$media_counter,cid) [lindex [split $value] 2]
		set sd_list($counter,$media_counter,ttl) $sd_list($counter,ttl)
		set sd_list($counter,$media_counter,fmt) ""
		set sd_list($counter,$media_counter,desc) ""
	    }
	    {a} {
		set sd_list($counter,$media_counter,desc) [lindex [split $value ":"] 1]
		set sd_list($counter,$media_counter,fmt) [lindex [split $value ":"] 1]
	    }
	    {default} {
		global VERSION
		puts "$VERSION error in ~/.sd_cache \n Unknow prefix: $prefix\n"
	    }
	}
	set line [get_next_line $sdf]
    }
    catch {close $sdf}
    if {$counter > 0 } {
	set sd_list($counter,media) $media_counter
	for {set i 1} {$i <= $counter} {incr i} {
	    .sd_import.session.frame.list insert end $sd_list($i,name)
	}
	global sd_session_index
	set sd_session_index 1
	update_media_sd_list 1
	.sd_import.session.frame.list select anchor 0
	.sd_import.session.frame.list select set 0
    }
}

proc update_media_sd_list {index} {
    global sd_list
    .sd_import.media.frame.list delete 0 end
    for {set i 1} {$i <= $sd_list($index,media)} {incr i} {
	if {$sd_list($index,$i,desc) != ""} {
	    .sd_import.media.frame.list insert end "$sd_list($index,$i,name) \
		    ($sd_list($index,$i,desc)) \
		    $sd_list($index,$i,addr)/$sd_list($index,$i,port)/\
		    $sd_list($index,$i,cid)/$sd_list($index,$i,ttl)"
	} else {
	    .sd_import.media.frame.list insert end "$sd_list($index,$i,name) \
		    $sd_list($index,$i,addr)/$sd_list($index,$i,port)/\
		    $sd_list($index,$i,cid)/$sd_list($index,$i,ttl)"
	}
	.sd_import.media.label configure -text "Media list of session:\n\"$sd_list($index,name)\""
    }
}


proc import_sd_title {index} {
    global sd_list tmp_session
    set tmp_session(name) $sd_list($index,name)
    set tmp_session(desc) $sd_list($index,desc)
    set tmp_session(max_ttl) $sd_list($index,ttl)
}

proc import_sd_filename {index} {
    global sd_list tmp_session
    regsub -all "\[ \t\]" $sd_list($index,name) "_" new_name
    regsub -all "\[-\"`';:,.<>{}[]()~!?#$%^&*+=|\\\]" $new_name "" new_name
    while { [regsub -all "__" $new_name "_" new_name] == 1 } {}
    set tmp_session(filename) [name2filename $sd_list($index,name)]
    set tmp_session(filename) $tmp_session(filename).vcr
}

proc import_sd_session {index} {
    global tmp_session sd_list VERSION
    if { $tmp_session(nb_of_media) > 0 } {
	set ret [AlertBox "\n $VERSION Import SD Session: \n\n Overwrite current media \n\
		or append new media to the session? \n\n" "" "" "$VERSION (Import SD Session)"\
		Overwrite Append Cancel]
	case $ret in {
	    {0} {
		set tmp_session(nb_of_media) 0
		global new_session
		set new_session 1
	    }
	    {1} {
		# do nothing
	    }
	    {2} {
		return
	    }
	}
    } else {
	set tmp_session(nb_of_media) 0
    }
    for {set i 0} {$i < $sd_list($index,media) } { incr i } {
	import_sd_media $index $i
    }
    import_sd_title $index
    import_sd_filename $index
}

proc import_sd_media {index list} {
    global tmp_session sd_list VERSION
    foreach j $list {
	set i [expr $j + 1]
	if { $sd_list($index,$i,name) == "audio" || $sd_list($index,$i,name) == "video" } {
	    incr tmp_session(nb_of_media)
	    set tmp_session(media_list,$tmp_session(nb_of_media),media_name) $sd_list($index,$i,name)
	    set tmp_session(media_list,$tmp_session(nb_of_media),media_desc) $sd_list($index,$i,desc)
	    set tmp_session(media_list,$tmp_session(nb_of_media),media_ip) $sd_list($index,$i,addr)
	    set tmp_session(media_list,$tmp_session(nb_of_media),media_port) $sd_list($index,$i,port)
	    set tmp_session(media_list,$tmp_session(nb_of_media),media_ttl) $sd_list($index,$i,ttl)
	    set tmp_session(media_list,$tmp_session(nb_of_media),media_scope) [scope2ttl $sd_list($index,$i,ttl) ]
	    set tmp_session(media_list,$tmp_session(nb_of_media),media_cid) $sd_list($index,$i,cid)
	    set tmp_session(media_list,$tmp_session(nb_of_media),media_type) [sd_format2type $sd_list($index,$i,name) $sd_list($index,$i,fmt)]
	    set tmp_session(media_list,$tmp_session(nb_of_media),media_cmd) [sd_format2cmd $sd_list($index,$i,name) $sd_list($index,$i,fmt)]
	    set tmp_session(media_list,$tmp_session(nb_of_media),media_mute) 0
	} else {
	    AlertBox "\n $VERSION Error: \n\n Format \"$sd_list($index,$i,name)\" not supported. \n\n"\
		    "" "" "$VERSION (Error)" Ignore
	}
    }
}

proc sd_format2type {name format} {
#currently everything is RTP (v0, v1 or v2)  
    case $name in {
	{audio} {
	    # vat protocol format
	    return 1
	}
	{video} {
	    case $format in {
		{vic} {
		    # rtp protocol format
		    return 1
		}
		{nv} {
		    # rtp protocol format
		    return 1
		}
		{ivs} {
		    # rtp protocol format
		    return 1
		}
		{jpeg} {
		    # rtp protocol format
		    return 1
		}
		{default} {
		    # use rtp as default video protocol
		    return 1
		    return 1
		}
	    }
	}
	{default} {
	    return 1
	}
    }
}

proc sd_format2cmd {name format} {
    case $name in {
	{audio} {
	    # vat protocol format
	    return "vat"
	}
	{video} {
	    case $format in {
		{vic} {
		    # vic
		    return "vic"
		}
		{nv} {
		    # nv
		    return "nv"
		}
		{ivs} {
		    # ivs
		    return "ivs"
		}
		{jpeg} {
		    # vic
		    return "vic"
		}
		{default} {
		    return ""
		}
	    }
	}
	{default} {
	    return ""
	}
    }
}

proc add_sd_session {} {
    global tmp_session sd_session_index
    set tmp_session(saved) 0
    import_sd_session $sd_session_index
    destroy .sd_import
    if { [winfo exists .edit_session] == 0} {
	set session(saved) 0
	update_session
	update_main_win
    } else {
	set tmp_session(saved) 0
	update_session_win
    }
}

proc add_sd_media {} {
    global VERSION tmp_session sd_session_index
    set tmp_session(saved) 0
    set index [.sd_import.media.frame.list curselection]
    if { [llength $index] == 0 } {
	AlertBox "\n $VERSION Error: \n\n Please select a media. \n\n"\
		"" "" "$VERSION (Add Media)" Ok
	return
    }
    import_sd_media $sd_session_index $index
    destroy .sd_import
    if { [winfo exists .edit_session] == 0} {
	set session(saved) 0
	update_session
	update_main_win
    } else {
	set tmp_session(saved) 0
	update_session_win
    }
}

proc add_sd_title {} {
    global tmp_session sd_session_index
    set tmp_session(saved) 0
    import_sd_title $sd_session_index
    import_sd_filename $sd_session_index
    destroy .sd_import
    if { [winfo exists .edit_session] == 0} {
	set session(saved) 0
	update_session
	update_main_win
    } else {
	set tmp_session(saved) 0
	update_session_win
    }
}

proc add_new_sd_session {} {
    global sd_session_index tmp_session
    destroy .sd_import
    import_sd_session $sd_session_index
    set file [FSBox "Select a filename for this session:" "$tmp_session(filename)" "" "" "*.vcr" "MBone VCR v0.1 (Import Session)"]
#     set file [FSBox "Select a filename for this session:" "$tmp_session(filename)" "" ""]
    if { $file != "" } {
	set tmp_session(filename) $file
	set tmp_session(read_only) ""
	fill_session
	if { [save_session] == 1 } {
	    update_session
	    enable_recording
	}
	update_main_win
	update_index_menu
    }
}
