#/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
#  
# Copyright (c) 1995 International Computer Science Institute
# All Rights Reserved.
#
# Created by Wieland Holfelder (whd@icsi.berkeley.edu)
# Tenet Group
# International Computer Science Institute
#
# Permission to use, copy, modify, and distribute this software and its 
# documentation for educational, research and non-profit purposes, 
# without fee, and without a written agreement is hereby granted, 
# provided that the above copyright notice, this paragraph and the 
# following five paragraphs appear in all copies.  
#
# Permission to incorporate this software into commercial products may 
# be obtained from the International Computer Science Institute,
# 1947 Center Street, Suite 600, Berkeley, CA 94704. 
#
# IN NO EVENT SHALL THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE BE 
# LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR 
# CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE 
# USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE INTERNATIONAL
# COMPUTER SCIENCE INSTITUTE  HAS BEEN ADVISED OF THE POSSIBILITY OF 
# SUCH DAMAGE.
#
# THE INTERNATIONAL COMPUTER SCIENCE INSTITUTE SPECIFICALLY DISCLAIMS 
# ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE 
# PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE INTERNATIONAL
# COMPUTER SCIENCE INSTITUTE  HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, 
# SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
#
# The International Computer Science Institute will appreciate 
# receiving any enhancements and modifications of this software 
# made by others, and reserves the right to include them in future 
# releases. The Institute also recommends that this software not 
# be redistributed by third parties, but be obtained by all directly 
# from the ICSI server, so that all users of the software may 
# receive bug notifications and other relevant announcements.
#
# The German Academic Exchange Service has supported the development 
# of this software. Their support is gratefully acknowledged.
#
#* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#
# $Id: tkerror.tcl,v 1.1 1996/07/26 12:12:57 whd Exp $ (ICSI)
# 
#* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


proc error { msg } {
    tkerror $msg
}

proc tkerror { msg } {
    global errorInfo VERSION tcl_interactive
    global HOST_TARGET HOST_VENDOR HOST_CPU HOST_OS 
    if { [info exists errorInfo] == 1 } {
	set info $errorInfo
    } else {
	set info "tkerror: variable errorInfo doesn't exist!?"
    }
    if { [string match "no more colors*" $msg] } {
	puts "$VERSION: $msg"
	return
    }
    set date [eval exec "date"]
    set error_log "./vcr.errorlog" 
    catch {set fd [open $error_log a]}
    if { $fd == 0 } {
	set error_log "~/vcr.errorlog"
	catch {set fd [open $errorlog a]}
	if { $fd == 0 } {
	    set error_log "/tmp/vcr.errorlog"
	    catch {set fd [open $errorlog a]}
	}
    }
    if { $fd > 0 } {
	puts $fd "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
	puts $fd "$VERSION: tcl runtime error: \n$msg\n"
	puts $fd "Date  : $date"
	puts $fd "System: $HOST_VENDOR $HOST_CPU"
	puts $fd "OS    : $HOST_OS\n"
	puts $fd "Stack trace:\n"
	puts $fd $info
	puts $fd ""
	puts $fd "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
	puts $fd "This error was logged in file: \"$error_log\""
	puts $fd "Please mail this error message and, if possible, an explanation"
	puts $fd "of what user-action caused it to mbone-vcr@icsi.berkeley.edu."
	puts $fd "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *"
	puts $fd "\n"
    }
    set dialog "$VERSION Error:\n\n$msg\n\n"
    if { $fd > 0 } {
	set dialog "${dialog}This error was logged in file: \"$error_log\"\n\n" 
    }
    set dialog "${dialog}Please mail this error message and, if possible, \
	    an explanation of what user-action caused it to mbone-vcr@icsi.berkeley.edu."

    if {[tk_dialog .tkerrorDialog "$VERSION Error in Tcl Script" \
	    "$dialog" error 0 OK "See Stack Trace"] == 0} {
	return
    }

    set w .tkerrorTrace
    catch {destroy $w}
    toplevel $w -class ErrorTrace
    wm minsize $w 1 1
    wm title $w "$VERSION Stack Trace for Error"
    wm iconname $w "Stack Trace"
    button $w.ok -text OK -command "destroy $w"
    text $w.text -relief raised -bd 2 -yscrollcommand "$w.scroll set" \
	    -setgrid true -width 70 -height 20
    scrollbar $w.scroll -relief flat -command "$w.text yview"
    pack $w.ok -side bottom -padx 3m -pady 3m -ipadx 2m -ipady 1m
    pack $w.scroll -side right -fill y
    pack $w.text -side left -expand yes -fill both
     
    $w.text insert end "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\n"
    $w.text insert end "$VERSION: tcl runtime error: \n$msg\n\n"
    $w.text insert end "Date  : $date\n"
    $w.text insert end "System: $HOST_VENDOR $HOST_CPU\n"
    $w.text insert end "OS    : $HOST_OS\n\n"
    $w.text insert end "Stack trace:\n\n"
    $w.text insert end "$info\n"
    $w.text insert end "\n"
    $w.text insert end "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\n"
    if { $fd > 0 } {
	$w.text insert end "This error was logged in file: \"$error_log\"\n"
    }
    $w.text insert end "Please mail this error message and, if possible, an explanation\n"
    $w.text insert end "of what user-action caused it to mbone-vcr@icsi.berkeley.edu.\n"
    $w.text insert end  "* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\n"
    $w.text insert end  "\n"
    
    $w.text mark set insert 0.0
    
    # Center the window on the screen.
    
    wm withdraw $w
    update idletasks
    set x [expr [winfo screenwidth $w]/2 - [winfo reqwidth $w]/2 \
	    - [winfo vrootx [winfo parent $w]]]
    set y [expr [winfo screenheight $w]/2 - [winfo reqheight $w]/2 \
	    - [winfo vrooty [winfo parent $w]]]
    wm geom $w +$x+$y
    wm deiconify $w
}
